Require Import ZArith.
Require Import Ensembles.
Require Import List.
Require Import Bool.
Require Import util.
Require Import wmm.
Require Import basic.

Module Machine (A:Archi) (dp:Dp).
Import dp.
Inductive DEvent : Set :=
  | DR : Event*Event -> DEvent
  | DW : Event -> DEvent.
Lemma eqDEvt_dec : forall (x y:DEvent), {x=y}+{x<>y}.
Proof.
intros x y; decide_equality.
Defined.
Hint Resolve eqDEvt_dec : equality.
Inductive Label : Set :=
  | d : DEvent -> Label
  | f : DEvent -> Label.
Lemma eqLabel_dec : forall (x y:Label), {x=y}+{x<>y}.
Proof.
intros x y; decide_equality.
Defined.
Hint Resolve eqLabel_dec : equality.

Definition Memory := Address -> Event. 
Definition wf_mem (E:Event_struct) (M:Memory) := 
  (forall a, writes E (M a) /\ loc (M a) = a).

Parameter dB : Event.
Hypothesis dB_not_in_E :
  forall E e, events E e -> e <> dB.

Definition Buffer := Rln Event.
Definition wf_buff (E:Event_struct) (B:Buffer) := 
  (forall w, udr B w -> writes E w) /\
  (forall w1 w2, w1 <> w2 -> ran B w1 -> ran B w2 -> loc w1 = loc w2 -> B w1 w2 \/ B w2 w1) /\
  (forall e, acyclic (rrestrict B (fun w => loc w = loc e))) /\
  
  trans B /\
  (forall w1 w2, B w1 w2 -> loc w1 = loc w2).

Parameter dQ : Event.
Hypothesis dQ_not_in_E :
  forall E e, events E e -> dQ <> e.
Definition Queue := Rln Event. 

Definition evt_of_devt de : Event :=
  match de with
    | DW w => w
    | DR (w,r) => r
  end.

Definition wf_queue (E:Event_struct) (Q:Queue) :=
  forall e, ran Q e -> reads E e. 

Record State := mks {
  M : Memory;
  B : Buffer;
  Q : Queue
}.

Definition upd_mem (M:Memory) (e:Event): Memory :=
  fun a => if eqLoc_dec a (loc e) then e else (M a).

Parameter bin : set Event -> Event -> bool.

Hypothesis In_bin :
  forall (r : set Event) e,
  r e -> bin r e = true.
Hypothesis bin_In :
  forall (r : set Event) e,
  bin r e = true -> r e.
Hypothesis nbin_nIn :
  forall (r : set Event) e,
  bin r e = false -> ~r e.

Definition upd_buff (B:Buffer) (e:Event): Buffer :=
  if bin (udr B) e then B else
  rel_union B (fun e1 => fun e2 => ( ( udr B e1 /\ loc e1 = loc e)) /\ e2 = e).  

Definition upd_queue (Q:Queue) e: Queue :=
  rel_union Q (fun e1 => fun e2 => e1 = dQ /\ e2 = e).

Definition del_buff (B:Buffer) (e:Event): Buffer :=
  (fun e1 => fun e2 => B e1 e2 /\ e1 <> e /\ e2 <> e).

Definition del_queue (Q:Queue) (e:Event) : Queue :=
  (fun e1 => fun e2 => Q e1 e2 /\ ( e2 <> e)).

Parameter bemp : Rln Event -> bool.
Hypothesis bemp_axt : forall r s, bemp (rrestrict r s) = true -> ~(exists e, ran r e /\ s e).
Hypothesis bemp_axf : forall r s, bemp (rrestrict r s) = false -> (exists e, ran r e /\ s e).

Definition buff_write_step (s:State) (w:Event) : option State :=
  let M := M s in let B := B s in let Q := Q s in 
        let B' := upd_buff B w in Some (mks M B' Q).

Definition buff_read_step (s:State) (r:Event) : option State :=
  let M := M s in let B := B s in let Q := Q s in 
      let Q' := upd_queue Q r in Some (mks M B Q').

Module EPick.
Parameter pick : set Event -> Event.
Axiom pick_singleton :
  forall e, pick (Singleton _ e) = e.
Axiom pick_in :
  forall s, s (pick s).
End EPick.

Definition pick_last B :=
  EPick.pick (fun w => (~(exists w, B dB w) /\ w = dB) \/ 
                       ((exists w, B dB w) /\ B dB w /\ ~(exists w', B w' w))).

Definition p2ws (E:Event_struct) (p:Rln Label) : Rln Event :=
  fun w1 => fun w2 => writes E w1 /\ writes E w2 /\ loc w1 = loc w2 /\ 
                      exists l1, exists l2, l1 = f (DW w1) /\ l2 = f (DW w2) /\ p l1 l2.

Definition rrB r :=
  fun e => 
    loc e = loc r /\
    if A.intra then 
       if A.inter then (proc_of e = proc_of r \/ proc_of e <> proc_of r)
       else proc_of e = proc_of r
    else if A.inter then proc_of e <> proc_of r
         else False.

Definition wbef E B r := EPick.pick (fun w => ran B w /\ loc r = loc w /\ po_iico E w r /\ 
                                     ~(exists w', loc r = loc w' /\ po_iico E w w' /\ po_iico E w' r)).
Definition waft E B r := EPick.pick (fun w => ran B w /\ loc r = loc w /\ po_iico E r w /\ 
                                     ~(exists w', loc r = loc w' /\ po_iico E r w' /\ po_iico E w' w)).

Definition p2rf E L : Rln Event :=
  fun w => fun r => rfmaps (events E) w r /\ L (f (DR (w,r))).

Definition p2X (E:Event_struct) (p: Rln Label) (L:set Label) : Execution_witness := mkew (p2ws E p) (p2rf E L).

Definition flush_write_step E p L (s:State) (w:Event) : option State :=
  let M := M s in let B := B s in let Q := Q s in 
  if bemp (rrestrict (rel_union B Q) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w)) then
  if bemp (rrestrict Q (fun r' => loc r' = loc w /\ po_iico E r' w)) then 
    let lbe := pick_last (rrestrict B (fun e => loc e = loc w)) in 
        if eqEv_dec lbe w then
            let M' := upd_mem M w in let B' := del_buff B w in
            Some (mks M' B' Q)
        else None
  else None
  else None.

Definition buff_before s r :=
  fun w => if eqEv_dec w ((M s) (loc r)) then True 
           else if bin (ran (B s)) w then True
           else False.

Definition flush_resolved_read_step E p L (s:State) (w:Event) (r:Event) : option State :=
  let M := M s in let B := B s in let Q := Q s in
  if bin (ran Q) r then
  if bemp (rrestrict Q (fun e => (dp.dp E) e w)) then
  if bemp (rrestrict (rel_union B Q) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r)) then
  if eqEv_dec w (M (loc r)) then
     if bemp (rrestrict B (fun w => po_iico E w r /\ loc r = loc w)) then 
       let Q' := del_queue Q r in Some (mks M B Q')
     else None
  else if bin (Intersection _ (ran B) (rrB r)) w then None
       else if bin (fun w => B w (waft E B r) /\ loc r = loc w) w then
            if bin (fun w => (w = (wbef E B r) \/ B (wbef E B r) w) /\ loc r = loc w) w then
            let Q' := del_queue Q r in Some (mks M B Q')
  else None
  else None
  else None
  else None
  else None.  

Definition step E p L (s:option State) (l:Label) : option State :=
  match s with
  | None => None
  | Some s =>
  let M := M s in let B := B s in let Q := Q s in
  match l with
  | d de =>
    match de with
    | DW w => buff_write_step s w
    | DR (_,r) => buff_read_step s r
    end
  | f de => 
    match de with
    | DW w => flush_write_step E p L s w
    | DR (w,r) => flush_resolved_read_step E p L s w r
    end
  end
  end. 

Definition pred (r:Rln Label) (l:Label) :=
  fun l1 => fun l2 => r l1 l2 /\  r l2 l.
Module LPick.
Parameter pick : set Label -> Label.
Axiom pick_singleton :
  forall e, pick (Singleton _ e) = e.
Axiom pick_in :
  forall s, s (pick s).
End LPick.
Definition last r :=
  LPick.pick (fun l => ran r l /\ ~(exists l', r l l')).

Fixpoint machine (E:Event_struct) (r:Rln Label) (L:set Label) (lst:Label) (n:nat) : Rln (option State) :=
  match n with 
  | O => fun s1 => fun s2 => s1 = s2 /\ 
     match s1 with 
     | None => False 
     | Some s1 => (forall l w, w = M s1 l -> udr r (f (DW w))) /\
                  (forall w, udr (B s1) w <-> udr r (d (DW w))) /\
                  (forall dr, ran (Q s1) (evt_of_devt dr) <-> udr r (d dr)) /\
                  (forall dr, ~ran (Q s1) (evt_of_devt dr) <-> udr r (f dr)) /\
                  (forall w, ~(udr r (d (DW w)) /\ udr r (f (DW w))))
     end
  | S m => fun s0 => fun s2 =>
     exists l, exists s1, (last r = l ) /\ 
     machine E (pred r l) L l m s0 (Some s1) /\ step E r L (Some s1) l = s2 /\
     (forall s, s2 = Some s -> forall a, r (f (DW (M s1 a))) (f (DW (M s a)))) /\
     wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1)
  end.

Definition machine_not_stuck E p (L:set Label) :=
  forall l, L l -> 
  (exists n, exists s0, exists s1, exists s2, wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1) /\
     machine E (pred p l) L l n (Some s0) (Some s1) /\ step E p L (Some s1) l = Some s2).

Axiom excluded_middle : forall (A:Prop), A \/ ~A.
End Machine.

Module Soundness (A:Archi) (dp:Dp).
  Module AWmm := Wmm A dp.
  Module ABasic := Basic A dp.
  Module AMach := Machine A dp.
  Import AMach.

Definition devt_of_label l : DEvent :=
  match l with
    | d de | f de => de
  end.

Definition wf_labels (E:Event_struct) (L:set Label) :=
  ((forall w, L (d (DW w)) -> writes E w) /\
   (forall w, L (f (DW w)) -> writes E w)) /\
  ((forall w r, L (d (DR (w,r))) -> rfmaps (events E) w r) /\
   (forall w r, L (f (DR (w,r))) -> rfmaps (events E) w r)) /\
  (forall w, writes E w -> L (f (DW w))) /\
  (forall de, L (f de) -> L (d de)).

Definition evt_of_label l :=
  match l with 
    | d de | f de => evt_of_devt de 
  end.
  
Lemma wfl_w :
  forall E L,
  wf_labels E L ->
  (forall w, writes E w -> L (f (DW w))).
Proof.
intros E L [? [? [Hle ?]]] w Hw.
  apply Hle; auto.
Qed.

Lemma wfl_e :
  forall E L,
  wf_labels E L ->
  (forall l : Label, L l -> events E (evt_of_label l)).
Proof.
intros E L [[Hwd Hwf] [[Hrd Hrf] ?]] l HL.
case_eq l; intros de Hl; rewrite Hl in HL; case_eq de.

  intros [w r] Hde; rewrite Hde in HL.
    simpl; generalize (Hrd w r HL); intros [? [? ?]]; auto.

  intros w Hde; rewrite Hde in HL.
    simpl; generalize (Hwd w HL); intros [? ?]; auto.

  intros [w r] Hde; rewrite Hde in HL.
    simpl; generalize (Hrf w r HL); intros [? [? ?]]; auto.

  intros w Hde; rewrite Hde in HL.
    simpl; generalize (Hwf w HL); intros [? ?]; auto.
Qed.

Lemma path_ws_wf :
  forall E p (L:set Label),
  (forall w, writes E w -> L (f (DW w))) ->
  linear_strict_order p L ->
  write_serialization_well_formed (events E) (ws (p2X E p L)).
Proof.
intros E p L HLw Hlp; split.    

intros l; split; [|split; [|split]].

  intros ? [e1 [e2 [H12 [? ?]]] | e1 [e2 [H12 [? ?]]]]; auto.

  intros e1 e2 e3 
    [[[Hw1 [Hw2 [Hl12 [l1 [l2 [? [Hl2 H12]]]]]]] [? ?]] 
    [[Hw2' [Hw3 [Hl23 [l2' [l3 [Hl2' [? H23]]]]]]] [? ?]]]; split; [|split]; auto.
  split; [|split]; auto.
  split. 
    rewrite Hl12; auto.
    exists l1; exists l3; split; [|split]; auto.
      destruct_lin Hlp; apply Htrans with (f (DW e2)); split.
      rewrite <- Hl2; auto. rewrite <- Hl2'; auto. 
    
  intros x [[? [? [? [lx [lx' [Hlx [Hlx' Hx]]]]]]] ?].
  rewrite Hlx in Hx; rewrite Hlx' in Hx; destruct_lin Hlp;
  apply Hac with (f (DW x)); auto.

  intros e1 e2 Hd H1 H2. 
  destruct_lin Hlp.
  assert (f (DW e1) <> f (DW e2)) as Hdiff.
    intro Hef; inversion Hef as [Heq]; rewrite Heq in Hd; apply Hd; auto.

  assert (writes E e1) as Hw1.
    destruct H1 as [? [v ?]]; split; auto; exists l; exists v; auto.
  assert (writes E e2) as Hw2.
    destruct H2 as [? [v ?]]; split; auto; exists l; exists v; auto.

  assert (L (f (DW e1))) as Hl1.
    apply HLw; auto.
  assert (L (f (DW e2))) as Hl2.
    apply HLw; auto.
  assert (loc e1 = loc e2) as Heql.
    destruct H1; rewrite ABasic.write_to_implies_this_loc with e1 l; auto;
    destruct H2; apply sym_eq; apply ABasic.write_to_implies_this_loc; auto.
  generalize (Htot (f (DW e1)) (f (DW e2)) Hdiff Hl1 Hl2); intro Hor.

  inversion Hor as [H12 | H21]; [left | right]; split ; auto; split; auto; split; auto; split; auto.
    exists (f (DW e1)); exists (f (DW e2)); split; [|split]; auto.
    exists (f (DW e2)); exists (f (DW e1)); split; [|split]; auto.

  intros e1 e2 [[He1 [l1 [v1 Hw1]]] [[He2 [l2 [v2 Hw2]]] [Heql [lb1 [lb2 [H1 [H2 H12]]]]]]]; 
    exists (loc e1); split; split; auto.
      exists v1; unfold loc; rewrite Hw1; auto.
      exists v2; rewrite Heql; unfold loc; rewrite Hw2; auto.
Qed.

Definition wf_rf_labels E L :=
  (forall r, reads E r -> exists w, rfmaps (events E) w r /\ L (f (DR (w,r)))) /\
  ((forall x w1 w2, L (d (DR (w1, x))) -> L (d (DR (w2, x))) -> w1 = w2) /\
   (forall x w1 w2, L (f (DR (w1, x))) -> L (f (DR (w2, x))) -> w1 = w2)).

Lemma path_rf_wf :
  forall E p L,
  linear_strict_order p L ->
  wf_rf_labels E L ->
  rfmaps_well_formed E (events E) (rf (p2X E p L)).
Proof.
intros E p L Hlin [Hex Huni]; split; [|split].

  intros r Hr; generalize (Hex r Hr); intros [w [Hrfm ?]]; 
  exists w; split; auto.
    destruct Hrfm as [? ?]; auto.
    split; auto.

  intros e1 e2 [? ?]; auto.

  intros r w1 w2 [? ?] [? ?]; subst; auto.
  apply Huni with r; auto; destruct_lin Hlin.
Qed.

Lemma td_tpred :
  forall p lb,
  trans p -> trans (pred p lb).
Proof.
intros p lb Htd x y z [Hxy ?] [Hyz ?]; split; auto.
  apply Htd with y; auto.
Qed.

Lemma flush_read_m_contrad :
  forall E p L s0 s1 wr r l w,
  (M s0) l <> w -> (M s1) l = w ->
  step E p L (Some s0) (f (DR (wr,r))) = Some s1 ->
  f (DR (wr,r)) = f (DW w).  
Proof.
intros E p L s0 s1 wr r l' w' Hni0 Hi1 Hs; unfold step in Hs.
unfold flush_resolved_read_step in Hs.
destruct (bin (ran (Q s0)) r); [|inversion Hs].
destruct (bemp (rrestrict (Q s0) (fun e => (dp.dp E) e wr))); [|inversion Hs].
destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].
destruct (eqEv_dec wr (M s0 (loc r))) as [Him | Hnim].
  destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r /\ loc r = loc w))); [|inversion Hs].
    inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.
destruct (bin (Intersection _ (ran (B s0)) (rrB r)) wr); [inversion Hs|].
  destruct (bin (fun w => (B s0 w (waft E (B s0) r)) /\ loc r = loc w) wr); [|inversion Hs].
  destruct (bin (fun w : Event => (w = wbef E (B s0) r \/ B s0 (wbef E (B s0) r) w) /\ loc r = loc w)  wr); [|inversion Hs].
    inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.          
Qed.

Lemma m_step :
  forall E p L s0 s1 lb l w,
  (M s0) l <> w -> (M s1) l = w ->
  step E p L (Some s0) lb = Some s1 ->
  lb = f (DW w).
Proof.
intros E p L s0 s1 lb l' w' Hni0 Hi1 Hs.
case_eq lb; intros de Hde; rewrite Hde in Hs; case_eq de.

  
    
    intros [w r] Hr; rewrite Hr in Hs; unfold step in Hs.
      unfold buff_read_step in Hs.
        inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.

    
    intros w Hw; rewrite Hw in Hs; unfold step in Hs.
      unfold buff_write_step in Hs.
      inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.

  
    
    intros [w r] Hr; rewrite Hr in Hs; apply flush_read_m_contrad with E p L s0 s1 l'; auto.   

   
    intros w Hw; rewrite Hw in Hs; unfold step in Hs.
    unfold flush_write_step in Hs.
    destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs].
    destruct (bemp (rrestrict (Q s0) (fun r' => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
    destruct (eqEv_dec
           (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc w))) w); [|inversion Hs].
    inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1.     
        unfold upd_mem in Hi1; destruct (eqLoc_dec l' (loc w)) as [Heql | Hneql].
          rewrite Hi1; auto.
          contradiction.
Qed.

Lemma last_in_ran :
  forall r l, last r = l -> ran r l.
Proof.
unfold last; intros r l Hl.
generalize (LPick.pick_in (fun l : Label => ran r l /\ ~ (exists l' : Label, r l l')));
rewrite Hl; intros [? ?]; auto.
Qed.

Lemma in_mem_implies_in_order :
  forall E p L lst n s0 s1 lb w l,
  trans p ->
  (exists d, trans d /\ rel_incl d (pred p lb) /\ machine E d L lst n (Some s0) (Some s1)) ->
  w = M s1 l ->
  p (f (DW w)) lb .
Proof.
intros E p L lst n s0 s2 lb w l Htp Hm Hw; generalize p L lst lb s2 Htp Hm Hw; 
clear p L lst lb s2 Htp Hm Hw; induction n;
intros p L lst lb s2 Htp [d [Htd [Hid Hm]]] Hw.

  inversion Hm as [Heqs [Him ?]];
  inversion Heqs as [Heq]; rewrite <- Heq in Hw. 
  generalize (Him l w Hw); intros [x [y Hd] | y [x Hd]];
    generalize (Hid x y Hd); intros [? ?]; auto.
    apply Htp with y; auto.

 destruct Hm as [lb' [s1 [Hd [Hm [Hs12 ?]]]]]; fold (machine E (pred d lb') L lb' n (Some s0) (Some s1)) in Hm. 
  generalize (excluded_middle (w = M s1 l)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.

    generalize (td_tpred d lb' Htd); intro Htpp.
    assert (exists d0 : Rln Label, trans d0 /\
      rel_incl d0 (pred d lb') /\ machine E d0 L lb' n (Some s0) (Some s1)) as Hed.
      exists (pred d lb'); split; [|split]; auto. intros x y Hxy; auto.
    generalize (IHn d L lb' lb' s1 Htd Hed Hi); intro Hdwlb'.
    
    generalize (Hid (f (DW w)) lb' Hdwlb'); intros [Hpwlb' Hdlb'].
      apply Htp with lb'; auto. 
      
    assert (lb' = f (DW w)) as Hlb'.
      apply m_step with E d L s1 s2 l; auto.

      
      assert (d (last (pred d lb')) lb') as Hdlb''.
        assert ((last (pred d lb')) = (last (pred d lb'))) as Ht.
          auto.
        generalize (last_in_ran (pred d lb') (last (pred d lb')) Ht); intros [x [Hxplb' ?]]; auto.

      rewrite <- Hlb'; generalize (Hid  (last (pred d lb')) lb' Hdlb''); intros [? Hor]; auto.
Qed. 

Lemma upd_queue_keeps :
  forall Q r r',
  ran Q r ->
  ran (upd_queue Q r') r.
Proof.
intros q r r' Hin.
unfold upd_queue.
destruct Hin as [e' Hd]; exists e'; left; auto.
Qed.

Lemma del_queue_keeps :
  forall Q r r',
  r <> r' ->
  ran Q r ->
  ran (del_queue Q r') r.
Proof.
intros q r r' Hdiff Hin;
unfold del_queue.
destruct Hin as [e' Hd]; exists e'; split; auto.
Qed.

Lemma flush_read_same_queue :
  forall E (p:Rln Label) L s0 s1 wr r wr' r',
  events E r ->
  step E p L (Some s0) (f (DR (wr',r'))) = Some s1 ->
  wf_rf_labels E L ->
  L (f (DR (wr',r'))) -> L (f (DR (wr,r))) ->
  f (DR (wr',r')) <> f (DR (wr,r)) ->
  ran (Q s0) r ->
  ran (Q s1) r.   
Proof.
intros E p L s0 s1 wr r wr' r' Her Hs Hwfr Hlwr' Hlwr Hd Hi.
unfold step in Hs; unfold flush_resolved_read_step in Hs.
destruct (bin (ran (Q s0)) r'); [|inversion Hs];
destruct (bemp (rrestrict (Q s0) (fun e => (dp.dp E) e wr'))); [|inversion Hs];
destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r'))); [|inversion Hs].

  destruct (eqEv_dec wr' (M s0 (loc r'))) as [Him | Hnim].
    destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r' /\ loc r' = loc w))); [|inversion Hs].
      inversion Hs; apply del_queue_keeps; auto.
        intro Heq; rewrite Heq in Hd; rewrite <- Heq in Hlwr'.
        
        destruct Hwfr as [? [? Huni]]; generalize (Huni r wr wr' Hlwr Hlwr');
        intro Heqw; rewrite <- Heq in Hd; rewrite Heqw in Hd; apply Hd; auto.

    destruct (bin (Intersection _ (ran (B s0)) (rrB r')) wr'); [inversion Hs|].
    destruct (bin (fun w : Event => (B s0 w (waft E (B s0) r')) /\ loc r' = loc w) wr'); [|inversion Hs].
      destruct (bin (fun w : Event => (w = wbef E (B s0) r' \/ B s0 (wbef E (B s0) r') w) /\ loc r' = loc w) wr'); [|inversion Hs].
      inversion Hs; apply del_queue_keeps; auto.
        intro Heq; rewrite Heq in Hd; rewrite <- Heq in Hlwr'.

        destruct Hwfr as [? [? Huni]]; generalize (Huni r wr wr' Hlwr Hlwr');
        intro Heqw; rewrite <- Heq in Hd; rewrite Heqw in Hd; apply Hd; auto.
Qed.

Lemma not_flush_read_step_implies_same_queue :
  forall E (p:Rln Label) L s0 s1 lb wr r,
  events E r ->
  step E p L (Some s0) lb = Some s1 ->
  wf_rf_labels E L ->
  lb <> f (DR (wr,r)) ->
  L lb -> L (f (DR (wr,r))) ->
  ran (Q s0) r ->
  ran (Q s1) r.  
Proof.
intros E p L s0 s1 lb wr r Her Hs Hwfr Hd Hllb Hlwr Hi; case_eq lb; 
intros de Hlb; rewrite Hlb in Hs; case_eq de; 
intros c Hde; rewrite Hde in Hs; simpl in Hs.



  
  destruct c as [w' r']; unfold buff_read_step in Hs;
  inversion Hs; auto;
  apply upd_queue_keeps; auto.

  
  unfold buff_write_step in Hs;
  inversion Hs; auto.


  
  rewrite Hde in Hlb; rewrite Hlb in Hd;
  destruct c as [w' r']; apply flush_read_same_queue with E p L s0 wr w' r'; auto.
    rewrite Hlb in Hllb; auto.

  
  unfold flush_write_step in Hs;
  destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e c))); [|inversion Hs];
  destruct (bemp (rrestrict (Q s0) (fun r' => loc r' = loc c /\ po_iico E r' c))); [|inversion Hs];
  destruct (eqEv_dec
           (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc c))) c); [|inversion Hs];
  inversion Hs; auto.  
Qed.

Lemma no_interv_last :
  forall d p l,
  rel_incl d p ->
  ~(exists l', tc p (last (pred d l)) l' /\ tc p l' l).
Proof.
intros d p l Hi [l' [Hdll' Hl'l]].
Admitted. 
Lemma before_last_implies_in :
  forall p r l,
  rel_incl r p ->
  p l (last r) ->
  udr r l.
Proof.
intros p r l Hi Hp.
Admitted. 

Lemma in_order_implies_in_queue :
  forall E (p:Rln Label) L lst n s0 s1 lb wr r,
  Included _ (Union _ (dom p) (ran p)) L ->  
  trans p ->
  acyclic p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->  
  wf_rf_labels E L ->
  L lb -> 
  (forall de, L (f de) -> p (d de) (f de)) ->
  p (d (DR (wr,r))) lb -> p (d (DR (wr,r))) lst ->
  p lb (f (DR (wr,r))) \/ lb = (f (DR (wr,r))) ->
  (~ (exists l' : Label, tc p lst l' /\ tc p l' lb) \/ lst = lb) ->
  (exists d, trans d /\ rel_incl d (pred p lb) /\ 
     ~(exists l', tc p (last d) l' /\ tc p l' lst) /\
     machine E d L lst n (Some s0) (Some s1)) ->
  ran (Q s1) r.
Proof.
intros E p L lst n s0 s2 lb'' wr r Hudrp Htp Hacp Htotp Hwfr Hllb Hdf  
  Hprlb Hprlst H21 Hni Hm; 
generalize s2 lst lb'' Hwfr Hm Hllb  Hprlb Hprlst H21 Hni; 
clear s2 lst lb'' Hwfr Hm Hllb  Hprlb Hprlst H21 Hni; 
induction n; 
intros s2 lst lb'' Hwfr [pd [Htd [Hi [Hniw Hm]]]] Hllb  Hprlb Hprlst H21 Hni.

  destruct Hm as [Hseq [? [? [Hiq ?]]]].
    assert (p (d (DR (wr,r))) (f (DR (wr,r)))) as Hpdf.
      apply Hdf; auto.
        inversion H21 as [Hip | Heq].
          apply Hudrp.
            right; exists lb''; auto.
            rewrite <- Heq; auto.
    destruct (eqLabel_dec (last pd) (d (DR (wr,r)))) as [Heql | Hneql].
      generalize (last_in_ran pd (d (DR (wr,r))) Heql).
      generalize (Hiq (DR (wr,r))); intros [dir bak].
        intro Hr; assert (udr pd (d (DR (wr,r)))) as Hudr.
          right; auto.
        generalize (bak Hudr); inversion Hseq as [Heqs]; simpl; auto.
      
      assert (p (last pd) (d (DR (wr,r))) \/ p (d (DR (wr,r))) (last pd)) as Hor.
        apply Htotp; auto; apply Hudrp; auto.
          assert (last pd = last pd) as Ht.
            auto.
          generalize (last_in_ran pd (last pd) Ht); intros [x Hx].
          generalize (Hi x (last pd) Hx); intros [? ?]; left; exists lb''; auto.
          left; exists (f (DR (wr,r))); auto.          

      inversion Hor as [Hlr | Hrl]; clear Hor.
          inversion Hni as [Hniw1 | Heq]. 
            assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
              exists (d (DR (wr,r))); split; apply trc_step; auto.
            contradiction.
            assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
              exists (d (DR (wr,r))); split; apply trc_step; auto.
            contradiction.

      generalize (Hiq (DR (wr,r))); intros [dir bak].
        assert (udr pd (d (DR (wr,r)))) as Hudr.
          apply before_last_implies_in with p; auto.
            intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto.
        generalize (bak Hudr); inversion Hseq as [Heqs]; simpl; auto.

  destruct Hm as [lb [s1 [Hdlb [Hm [Hs [? [? [? Hwfq]]]]]]]];
  fold (machine E (pred pd lb) L lb n (Some s0) (Some s1)) in Hm.
  destruct (eqLabel_dec lb (d (DR (wr,r)))) as [Heqlbr | Hneqlbr].
    rewrite Heqlbr in Hs; unfold step in Hs; unfold buff_read_step in Hs;
    inversion Hs as [Heqs]; simpl; exists dQ; right; auto.

  assert (exists d, trans d /\ rel_incl d (pred p lb) /\
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' lb) /\
          machine E d L lb n (Some s0) (Some s1)) as Hex.
    exists (pred pd lb); split; [|split; [|split]]; auto. 
      apply (td_tpred pd lb Htd).
      intros x y [Hxy Hylb]; 
      generalize (Hi x y Hxy); intros [? ?];
      generalize (Hi y lb Hylb); intros [? ?]; split; auto.
      apply no_interv_last; auto. 
        intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto.
  assert (L lb) as Hlb.
    apply Hudrp; auto.
    generalize (last_in_ran pd lb Hdlb); intros [x Hx];
    generalize (Hi x lb Hx); intros [? ?]; right; exists x; auto.
  assert (p lb (f (DR (wr, r))) \/ lb = f (DR (wr, r))) as Hor.
    generalize (last_in_ran pd lb Hdlb); intros [x Hd].
    generalize (Hi x lb Hd); intros [? Hplblb''].
    inversion H21 as [Hplb''r | Heqlb''r].
      left; apply Htp with lb''; auto.
      left; rewrite Heqlb''r in Hplblb''; auto.
  assert (~ (exists l' : Label, tc p lb l' /\ tc p l' lb) \/ lb = lb) as Hnieq.
    auto.
  assert (p (d (DR (wr,r))) lb) as Hrlb.
    assert (L (d (DR (wr,r)))) as Hlr.
      apply Hudrp; left; exists lb''; auto.
    generalize (Htotp lb (d (DR (wr,r))) Hneqlbr Hlb Hlr); intros [Hlbr | ?]; auto.
    assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
      rewrite Hdlb; exists (d (DR (wr,r))); split; apply trc_step; auto.
    contradiction.
  generalize (IHn s1 lb lb Hwfr Hex Hlb Hrlb Hrlb Hor Hnieq); intro Hi1.

  apply (not_flush_read_step_implies_same_queue E pd L s1 s2 lb wr); auto.
    generalize (Hwfq r Hi1); intros [? ?]; auto. 

    intro Heq; rewrite Heq in Hdlb; 
    assert (exists x, pd x (f (DR (wr,r)))) as Hdlbb.
      generalize (last_in_ran pd (f (DR (wr,r))) Hdlb); intros [x ?]; exists x; auto.
    destruct Hdlbb as [x Hdlbb];
    generalize (Hi x (f (DR (wr,r))) Hdlbb); intros [? Hbef].
    inversion H21 as [Haft | Heqlb'']; apply Hacp with lb''; auto. 
      apply trc_ind with (f (DR (wr,r))); apply trc_step; auto.
      subst; apply trc_step; auto.      

  apply Hudrp.
    assert (exists l, pd l lb) as Hdlbb.
      generalize (last_in_ran pd lb Hdlb); intros [l ?]; exists l; auto.
    destruct Hdlbb as [l Hdlbb]; generalize (Hi l lb Hdlbb); intros [? ?]; right; exists l; auto.

  inversion Hor as [Hp | Heq].
    apply Htp with lb; auto.
    rewrite Heq in H2; auto.
Qed.

Lemma upd_buff_keeps :
  forall (B:Buffer) w1 w2 w,
  B w1 w2 ->
 (upd_buff B w) w1 w2.
Proof.
intros b w1 w2 w Hin.
unfold upd_buff;
destruct (bin (udr b) w); auto.
left; auto.
Qed.

Lemma del_buff_keeps :
  forall (B:Buffer) w1 w2 w,
  w1 <> w -> w2 <> w ->
  B w1 w2 ->
  (del_buff B w) w1 w2.
Proof.
intros b w1 w2 w Hd1 Hd2 Hin;
unfold del_buff; auto.
Qed.

Lemma flush_read_same_buff :
  forall E p L s0 s1 w1 w2 wr r,
  step E p L (Some s0) (f (DR (wr,r))) = Some s1 ->
  f (DR (wr,r)) <> f (DW w1) ->
  (B s0) w1 w2 ->
  (B s1) w1 w2.  
Proof.
intros E p L s0 s1 w1 w2 wr' r' Hs Hd H12.
unfold step in Hs; unfold flush_resolved_read_step in Hs;
destruct (bin (ran (Q s0)) r'); [|inversion Hs];
destruct (bemp (rrestrict (Q s0) (fun e => (dp.dp E) e wr'))); [|inversion Hs];
destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r'))); [|inversion Hs].

  destruct (eqEv_dec wr' (M s0 (loc r'))) as [Him | Hnim].
    destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r' /\ loc r' = loc w))); [|inversion Hs].
      inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.
  destruct (bin (Intersection _ (ran (B s0)) (rrB r')) wr'); [inversion Hs|].
  destruct (bin (fun w : Event => (B s0 w (waft E (B s0) r')) /\ loc r' = loc w) wr'); [|inversion Hs].
  destruct (bin (fun w : Event => (w = wbef E (B s0) r' \/ B s0 (wbef E (B s0) r') w) /\ loc r' = loc w) wr'); [|inversion Hs].
    inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.           
Qed.

Lemma not_flush_write_step_implies_same_buff :
  forall E p L s0 s1 lb w1 w2,
  step E p L (Some s0) lb = Some s1 ->
  lb <> f (DW w1) -> lb <> f (DW w2) ->
  (B s0) w1 w2 ->
  (B s1) w1 w2.
Proof.
intros E p L s0 s1 lb w1 w2 Hs Hd1 Hd2 H12.
case_eq lb; 
intros de Hlb; rewrite Hlb in Hs; case_eq de; 
intros c Hde; rewrite Hde in Hs; simpl in Hs.



  
  destruct c as [w' r']; unfold buff_read_step in Hs;
  inversion Hs; auto.

  
  unfold buff_write_step in Hs;
  inversion Hs; apply upd_buff_keeps; auto.



  
  rewrite Hde in Hlb; rewrite Hlb in Hd1;
  destruct c as [w' r'];
  apply flush_read_same_buff with E p L s0 w' r'; auto.

  
  unfold flush_write_step in Hs;
  destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e c))); [|inversion Hs];
    destruct (bemp (rrestrict (Q s0) (fun r' => loc r' = loc c /\ po_iico E r' c))); [|inversion Hs];
  destruct (eqEv_dec
           (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc c))) c); [|inversion Hs];
  inversion Hs; apply del_buff_keeps; auto.
    rewrite Hde in Hlb; rewrite Hlb in Hd1; intros Heq; rewrite Heq in Hd1; apply Hd1; auto.
    rewrite Hde in Hlb; rewrite Hlb in Hd2; intros Heq; rewrite Heq in Hd2; apply Hd2; auto.
Qed.

Lemma in_order_implies_in_buff : 
  forall E p L lst n s0 s1 w lb,
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->  
  (forall l, L l -> events E (evt_of_label l) ) ->
  p (d (DW w)) lb -> p (d (DW w)) lst ->
  p lb (f (DW w)) ->
  (forall de, L (f de) -> p (d de) (f de)) ->
  (~ (exists l' : Label, tc p lst l' /\ tc p l' lb) \/ lst = lb) ->
  (exists d, trans d /\ rel_incl d (pred p lb) /\ 
     ~(exists l', tc p (last d) l' /\ tc p l' lst) /\
     machine E d L lst n (Some s0) (Some s1)) ->
  
  (B s1) dB w.
Proof.
intros E p L lst n s0 s2 w lb'' Hudrp Htp Hacp Htotp Hle Hpwlb Hpwlst Hib Hdf Hni Hm; 
generalize s2 lst lb'' Hpwlb Hpwlst Hib Hni Hm; clear s2 lst lb'' Hpwlb Hpwlst Hib Hni Hm; 
induction n; intros s2 lst lb'' Hpwlb Hpwlst Hib Hni [pd [Htd [Hi [Hniw Hm]]]].

  destruct Hm as [Hseq [? [Hibs ?]]].
    assert (p (d (DW w)) (f (DW w))) as Hpdf.
      apply Hdf; auto.
        apply Hudrp; right; exists lb''; auto.
    destruct (eqLabel_dec (last pd) (d (DW w))) as [Heql | Hneql].
      generalize (last_in_ran pd (d (DW w)) Heql).
      generalize (Hibs w); intros [dir bak].
        intro Hw; assert (udr pd (d (DW w))) as Hudr.
          right; auto.
        generalize (bak Hudr); inversion Hseq as [Heqs]; simpl; auto.
        admit. 
      
      assert (p (last pd) (d (DW w)) \/ p (d (DW w)) (last pd)) as Hor.
        apply Htotp; auto; apply Hudrp; auto.
          assert (last pd = last pd) as Ht.
            auto.
          generalize (last_in_ran pd (last pd) Ht); intros [x Hx].
          generalize (Hi x (last pd) Hx); intros [? ?]; left; exists lb''; auto.
          left; exists (f (DW w)); auto.          

      inversion Hor as [Hlr | Hrl]; clear Hor.
          inversion Hni as [Hniw1 | Heq]. 
            assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
              exists (d (DW w)); split; apply trc_step; auto.
            contradiction.
            assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
              exists (d (DW w)); split; apply trc_step; auto.
            contradiction.

      generalize (Hibs w); intros [dir bak].
        assert (udr pd (d (DW w))) as Hudr.
          apply before_last_implies_in with p; auto.
            intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto.
        generalize (bak Hudr); inversion Hseq as [Heqs]; simpl; auto.
        admit. 

  destruct Hm as [lb [s1 [Hdlb [Hm [Hs ?]]]]];
  fold (machine E (pred pd lb) L lb n (Some s0) (Some s1)) in Hm.
  destruct (eqLabel_dec lb (d (DW w))) as [Heqlbw | Hneqlbw].
    rewrite Heqlbw in Hs; unfold step in Hs; unfold buff_write_step in Hs;
    inversion Hs as [Heqs]; simpl. unfold upd_buff.
      case_eq (bin (udr (B s1)) w); intros Hb. 
        apply bin_In; auto. admit. 
        right; split; auto. admit. 

  assert (exists d, trans d /\ rel_incl d (pred p lb) /\
          ~(exists l', tc p (last d) l' /\ tc p l' lb) /\
         machine E d L lb n (Some s0) (Some s1)) as Hex.
    exists (pred pd lb); split; [|split; [|split]]; auto. 
      apply td_tpred; auto.
        intros x y [Hxy Hylb]; split; auto.
        generalize (Hi x y Hxy); intros [? ?]; auto.
        generalize (Hi y lb Hylb); intros [? ?]; auto.
      apply no_interv_last; auto.
        intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto. 
  assert (lb <> (f (DW w))) as Hdiff.
    intro Heq; rewrite Heq in Hdlb.
    generalize (last_in_ran pd (f (DW w)) Hdlb); intros [x Hx].
    generalize (Hi x (f (DW w)) Hx); intros [? ?].
    assert (p lb'' lb'') as Hc.
      apply Htp with (f (DW w)); auto.
    apply Hacp with lb''; auto; apply trc_step; auto.
  assert (L lb) as Hllb.
    generalize (last_in_ran pd lb Hdlb); intros [x Hx]; generalize (Hi x lb Hx); intros [? ?].
    apply Hudrp; left; exists lb''; auto.
  assert (L (f (DW w))) as Hlfw.
    apply Hudrp; right; exists lb''; auto.
  assert (p lb (f (DW w))) as Hplbw.
    generalize (Htotp lb (f (DW w)) Hdiff Hllb Hlfw); intros [|Hwlb]; auto.
    rewrite Hdlb in Hniw.
    assert (p (f (DW w)) (f (DW w))) as Hc.
      apply Htp with lb; auto.
        apply Htp with lb''; auto.
        generalize (last_in_ran pd lb Hdlb); intros [x Hx].
        generalize (Hi x lb Hx); intros [? ?]; auto.
    assert False as Ht.
      apply Hacp with (f (DW w)); auto; apply trc_step; auto.     
    inversion Ht.
  assert (~ (exists l' : Label, tc p lb l' /\ tc p l' lb) \/ lb = lb) as Hnilb.
    auto.
  assert (p (d (DW w)) lb) as Hpdwlb.
    assert (L (d (DW w))) as Hlw.
      apply Hudrp; left; exists lb''; auto.
    generalize (Htotp lb (d (DW w)) Hneqlbw Hllb Hlw); intros [Hlbw | ?]; auto.
    assert (exists l' : Label, tc p (last pd) l' /\ tc p l' lst) as Hc.
      rewrite Hdlb; exists (d (DW w)); split; apply trc_step; auto.
    contradiction.    
  generalize (IHn s1 lb lb Hpdwlb Hpdwlb Hplbw Hnilb Hex); intro His1.

   assert (ran pd lb) as Hran.
     destruct Hdlb as [? ?]; apply last_in_ran; auto.

  apply not_flush_write_step_implies_same_buff with E pd L s1 lb; auto.
   assert (L lb) as Helb.
      apply Hudrp; right. 
        destruct Hran as [lb' Hran].
        generalize (Hi lb' lb Hran); intros [? ?]; exists lb'; auto.

        intro Heq; rewrite Heq in Helb. 
        generalize (Hle (f (DW dB)) Helb); simpl; intro.

            assert (dB = dB) as Ht.
              trivial.
            generalize Ht; fold (dB <> dB). 
            apply dB_not_in_E with E; auto.   
Qed.

Lemma rq_implies_step_none :
  forall E p L s e1 e2,
  pio_llh E e1 e2 ->
  bin (ran (Q s)) e1 = true ->
  step E p L (Some s) (f (DW e2)) = None.
Proof.
intros E p L s e1 e2 [? [? ?]] Hq1.
generalize (excluded_middle (step E p L (Some s) (f (DW e2)) = None)); intro Hor;
inversion Hor as [|Hs]; auto.
simpl in Hs; unfold flush_write_step in Hs.
destruct (bemp (rrestrict (rel_union (B s) (Q s)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2))). 
case_eq (bemp (rrestrict (Q s) (fun r' => loc r' = loc e2 /\ po_iico E r' e2))); intro Hiq; [| rewrite Hiq in Hs].
generalize (bemp_axt (Q s) (fun r' => loc r' = loc e2 /\ po_iico E r' e2) Hiq); intro Hn.
assert (exists e, ran (Q s) e /\ loc e = loc e2 /\ po_iico E e e2) as Hc.
  exists e1; split; [|split]; auto.
    apply bin_In; auto. 
contradiction.

assert False as Ht.
  apply Hs; auto.
inversion Ht.

assert False as Ht.
  apply Hs; auto.
inversion Ht.
Qed.

Definition wf_l_p E p (L:set Label) :=
  forall l, L l ->
  ~(exists s1, exists s2, exists s1', exists s2', s1 <> s1' /\ 
    step E p L (Some s1) l = s2 /\ step E p L (Some s1') l = s2').  

Lemma mns_implies_not_none :
  forall E p (L:set Label) s l, 
  L l ->
  
  wf_l_p E p L ->
  machine_not_stuck E p L ->
  ~(step E p L (Some s) l = None).
Proof.
intros E p L s l Hd  Hus Hmns Hn.
destruct (Hmns l Hd) as [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]]; auto.
generalize (excluded_middle (s = s1)); intros [Heqs | Hneqs].
  rewrite Heqs in Hn; rewrite Hn in Hs; inversion Hs.
  generalize (Hus l Hd); intro.
  assert (exists s1, exists s2, exists s1', exists s2', s1 <> s1' /\ 
    step E p L (Some s1) l = s2 /\ step E p L (Some s1') l = s2') as Hc.
    exists s1; exists (Some s2); exists s; exists None; split; [|split]; auto.
  contradiction.
Qed.

Lemma rq_implies_contrad :
  forall E p (L:set Label) s e1 e2,
  wf_l_p E p L ->
  (forall w, writes E w -> L (f (DW w))) ->
  writes E e2 ->
  well_formed_event_structure E ->
  pio_llh E e1 e2 -> 
  bin (ran (Q s)) e1 = true ->
  ~(machine_not_stuck E p L).
Proof.
intros E p L s e1 e2 Hwflp Hlw Hw Hwf H12 Hiq Hmns.

assert (L (f (DW e2))) as Hd.
  apply Hlw; auto.
destruct (Hmns (f (DW e2)) Hd) as [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
generalize (rq_implies_step_none E p L s e1 e2 H12 Hiq);
apply mns_implies_not_none; auto.
Qed.

Lemma path_unirf : 
  forall E p L e1 e2,
  well_formed_event_structure E ->
  wf_labels E L ->
  wf_l_p E p L ->
  trans p -> acyclic p ->
  machine_not_stuck E p L ->  
  pio_llh E e1 e2 -> ~ rf (p2X E p L) e2 e1.
Proof.
intros E p L e1 e2 Hwf Hwfl Hwflp Htp Hacp Hmns H12 Hrf'.

assert (L (f (DR (e2,e1)))) as Hdl.
  simpl in Hrf'; unfold p2rf in Hrf'; destruct Hrf'; auto.
destruct (Hmns (f (DR (e2,e1))) Hdl) as [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
simpl in Hs; unfold flush_resolved_read_step in Hs.

case_eq (bin (ran (Q s1)) e1); intro Hiq; rewrite Hiq in Hs; [| inversion Hs].

destruct (bemp (rrestrict ((Q s1)) (fun e => (dp.dp E) e e2))); [|inversion Hs].

destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e1))); [|inversion Hs].

destruct (eqEv_dec e2 (M s1 (loc e1))) as [Him | Hnim].
  case_eq (bemp (rrestrict (B s1) (fun w : Event => po_iico E w e1 /\ loc e1 = loc w))); [intros Hemp | intros Hnemp; rewrite Hnemp in Hs; inversion Hs].
    assert (forall w, writes E w -> L (f (DW w))) as Hlw.
      apply wfl_w; auto.
    assert (writes E e2) as Hw.
      rewrite Him; generalize (Hwfm (loc e1)); intros [? ?]; auto.
    generalize (rq_implies_contrad E p L s1 e1 e2 Hwflp Hlw Hw Hwf H12 Hiq); intro; contradiction.

destruct (bin (Intersection _ (ran (B s1)) (rrB e1))); [inversion Hs|].  
case_eq (bin (fun w => (B s1 w (waft E (B s1) e1)) /\ loc e1 = loc w) e2); intros Haft; [|rewrite Haft in Hs; inversion Hs].

generalize (bin_In (fun w => (B s1 w (waft E (B s1) e1)) /\ loc e1 = loc w) e2 Haft);
intros [Hb Heql].

generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc e1 = loc w /\
                                    po_iico E e1 w /\
                                    ~ (exists w', (loc e1 = loc w' /\ po_iico E e1 w'))));
fold (waft E (B s1) e1); intros [Hr [Hl [Hpo Hn]]].

assert (exists w', loc e1 = loc w' /\ po_iico E e1 w') as Hc.
  destruct H12 as [? [? ?]]; exists e2; split; auto.
contradiction.
Qed.

Lemma pred_incl :
  forall r l,
  rel_incl (pred r l) r.
Proof.
intros r l x y [Hxy ?]; auto.
Qed.

Lemma pred_incl2 :
  forall r1 r2 l,
  rel_incl r1 r2 ->
  rel_incl (pred r1 l) (pred r2 l).
Proof.
intros r1 r2 l Hi12 x y [Hxy Hor]; split; auto.

Qed.

Lemma flush_before_implies_before_in_mem :
  forall E (p:Rln Label) L lst d n s0 s2 lb w,
  p (f (DW w)) lb ->
  rel_incl d (pred p lb) -> machine E d L lst n (Some s0) (Some s2) ->
  exists s1, machine E d L lst n (Some s1) (Some s2) /\
             w = M s1 (loc w).
Proof.
intros E p L lst pd n s0 s2 lb w Hp Hm; 
generalize s2 lst pd Hm; clear s2 lst pd Hm; induction n; 
intros s2 lst pd Hipd Hm.

  inversion Hm.
    admit. 

  destruct Hm as [lb' [s1 [Hdlb' [Hm Hs]]]];
  fold (machine E (pred pd lb') L lb' n (Some s0) (Some s1)) in Hm.
  assert (rel_incl (pred pd lb') (pred p lb)) as Hipd'.
    intros x y [Hxy ?]; auto; apply Hipd; auto.

  generalize (IHn s1 lb' (pred pd lb') Hipd' Hm); intros [s  [Hss1 Hinm]].
  exists s; split; auto.
    exists lb'; exists s1; split; auto.
Qed. 

Lemma in_order_implies_in_obuff : 
  forall E p L lst n s0 s1 w1 w2 lb,
  trans p -> acyclic p ->
  machine_not_stuck E p L -> 
  loc w1 = loc w2 ->
  p (d (DW w1)) (d (DW w2)) ->
  tc p lb (f (DW w1)) ->
  (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  (exists d, rel_incl d (pred p lb) /\ machine E d L lst n (Some s0) (Some s1)) ->
  B s1 w1 w2.
Proof.
intros E p L lst n s0 s2 w1 w2 lb'' Htp Hacp Hmns Hl Hd12 Hdl1 Hfifo Hm; 
generalize s2 lst Hm; clear s2 lst Hm; induction n; intros s2 lst [pd [Hi Hm]].

  inversion Hm.
    admit. 

  destruct Hm as [lb [s1 [Hdlb [Hm [Hs ?]]]]];
  fold (machine E (pred pd lb) L lb n (Some s0) (Some s1)) in Hm.
  assert (exists d, rel_incl d (pred p lb'') /\
         machine E d L lb n (Some s0) (Some s1)) as Hex.
    exists (pred pd lb); split; auto; intros x y [Hxy ?]; apply Hi; auto.
  generalize (IHn s1 lb Hex); intro His1.

   assert (ran pd lb) as Hran.
     destruct Hdlb; apply last_in_ran; auto.

  apply not_flush_write_step_implies_same_buff with E pd L s1 lb; auto.
  intro Heq; rewrite <- Heq in Hdl1.
  destruct Hran as [x Hran]; generalize (Hi x lb Hran); intros [? Hbef].
  apply Hacp with lb''; apply trc_ind with lb; auto. 
    apply trc_step; auto.

  intro Heq.
  assert (tc p (f (DW w2)) (f (DW w1))) as Hf21.
    apply trc_ind with lb''; auto.
      rewrite Heq in Hran; destruct Hran as [x Hran].
  generalize (Hi x (f (DW w2)) Hran); intros [? Hbef].
      apply trc_step; auto.
  apply Hacp with (f (DW w1)); apply trc_ind with (f (DW w2)); auto.
    rewrite trans_tc; auto.
  
Qed.

Lemma flush_read_b_step :
  forall E p L s0 s1 w wr r,
  ~(udr (B s0) w) -> (udr (B s1) w) ->
  step E p L (Some s0) (f (DR (wr,r))) = Some s1 ->
  f (DR (wr,r)) = d (DW w).
Proof.
intros E p L s0 s1 w' w r Hni0 Hi1 Hs.
unfold step in Hs; unfold flush_resolved_read_step in Hs.
destruct (bin (ran (Q s0)) r); [|inversion Hs];
destruct (bemp (rrestrict ((Q s0)) (fun e => (dp.dp E) e w))); [|inversion Hs];
destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].

      destruct (eqEv_dec w (M s0 (loc r))) as [Him | Hnim].
        destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r /\ loc r = loc w))); [|inversion Hs].
          inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.
      destruct (bin (Intersection _ (ran (B s0)) (rrB r)) w); [inversion Hs|].
        destruct (bin (fun w : Event => (B s0 w (waft E (B s0) r)) /\ loc r = loc w) w); [|inversion Hs].
          destruct (bin (fun w : Event => (w = wbef E (B s0) r \/ B s0 (wbef E (B s0) r) w) /\ loc r = loc w) w); [|inversion Hs].
          inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.           
Qed.

Lemma b_step :
  forall E p L s0 s1 lb w,
  events E w ->
  ~(udr (B s0) w) -> (udr (B s1) w) ->
  step E p L (Some s0) lb = Some s1 ->
  lb = d (DW w).
Proof.
intros E p L s0 s1 lb w' Hew' Hni0 Hi1 Hs.
case_eq lb; intros de Hde; rewrite Hde in Hs; case_eq de.

  
    
    intros [w r] Hr; rewrite Hr in Hs; unfold step in Hs; unfold buff_read_step in Hs. 
      inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1; contradiction.

    
    intros w Hw; rewrite Hw in Hs; unfold step in Hs; unfold buff_write_step in Hs.
      inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1.
      destruct Hi1 as [? Hi1 | ? Hi1]; unfold upd_buff in Hi1; destruct (bin (udr (B s0)) w).
        assert (udr (B s0) x) as Hc.
          left; auto.
        contradiction.

      inversion Hi1 as [w' [His0 | [Hor Heqww']]]. 
        assert (udr (B s0) x) as Hc.
          left; exists w'; auto.
        contradiction.
          destruct Hor; assert (udr (B s0) x) as Hc.
            auto.
        contradiction.

        assert (udr (B s0) x) as Hc.
          right; auto.
        contradiction.

      inversion Hi1 as [w' [His0 | [Hor Heqww']]]. 
        assert (udr (B s0) x) as Hc.
          right; exists w'; auto.
        contradiction.

        subst; auto.

  
    
    intros [w r] Hr; rewrite Hr in Hs; apply flush_read_b_step with E p L s0 s1; auto.

    
    intros w Hw; rewrite Hw in Hs; unfold step in Hs; unfold flush_write_step in Hs.
      destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs].
      destruct (bemp (rrestrict (Q s0) (fun r' => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
      destruct (eqEv_dec (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc w))) w); [|inversion Hs].
      inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1. 
      destruct (eqEv_dec w w') as [Heqw | Hneqw].
        subst; auto. 
          inversion Hi1 as [? [x [? [Hd ?]]] | ? [x [? [? Hd]]]].
            assert False as Ht.
              apply Hd; auto.
            inversion Ht.
            
            assert False as Ht.
              apply Hd; auto.
            inversion Ht.

      assert (udr (B s0) w') as Hc.
        inversion Hi1 as [? [x [Hi ?]] | ? [x [Hi ?]]].
          left; exists x; auto. right; exists x; auto.
      contradiction.
Qed.
  
Lemma in_buff_implies_in_order :
  forall E p L lst n s0 s1 lb w,
  events E w ->
  machine E (pred p lb) L lst n (Some s0) (Some s1) ->
  
  udr (B s1) w ->
  tc p (d (DW w)) lb \/ (d (DW w)) = lb.
Proof.
intros E p L lst n s0 s2 lb w Hew Hm  Hw; generalize p lb s2 lst Hw Hm ; 
clear p lb s2 lst Hw Hm ; induction n; intros p lb s2 lst Hw Hm .
  inversion Hm as [Heqs [? [Hib ?]]]; inversion Heqs as [Heq]; rewrite <- Heq in Hw.
    generalize (Hib w); intros [dir ?];
    generalize (dir Hw); intros [x [y [? ?]] | x [y [? ?]]]. 
      left; apply trc_ind with y; auto; apply trc_step; auto.
      left; apply trc_step; auto.

  destruct Hm as [lb' [s1 [Hd [Hm [Hs12 ?]]]]]; fold (machine E (pred (pred p lb) lb') L lb' n (Some s0) (Some s1)) in Hm.
  generalize (excluded_middle (udr (B s1) w)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.
    
    generalize (IHn (pred p lb) lb' s1 lb' Hi Hm ).
    assert (rel_incl (pred p lb) p) as Hip.
      intros x y Hxy; inversion Hxy as [Hdwlb' Hdlb' ]; auto.
    generalize (tc_incl Hip); intro Hitc.
    assert (ran (pred p lb) lb') as Hran.
      destruct Hd; apply last_in_ran; auto.
    destruct Hran as [x [? ?]]; intros [Hpwlb' | Heqwlb']; left.
        
          apply trc_ind with lb'; auto.
          apply trc_step; auto.
        rewrite Heqwlb';  apply trc_step; auto.

    assert (lb' = d (DW w)) as Hlb'.
      apply b_step with E (pred p lb) L s1 s2; auto.
    assert (ran (pred p lb) lb') as Hran.
      destruct Hd; apply last_in_ran; auto.      
    destruct Hran as [lb'' [Hdlb'' Hdlb']].
      rewrite <- Hlb'; auto; left;  apply trc_step; auto.       
      
Qed. 

Lemma pick_last_in_r :
  forall r s e,
  e <> dB ->
  pick_last (rrestrict r s) = e ->
  r dB e.
Proof.
intros r s e Hdiff Hp.
unfold pick_last in Hp.
generalize (EPick.pick_in (fun w : Event =>
        (~ (exists w0 : Event, rrestrict r s dB w0) /\ w = dB) \/
        ((exists w0 : Event, rrestrict r s dB w0) /\
         rrestrict r s dB w /\ ~ (exists w' : Event, rrestrict r s w' w))));
rewrite Hp; intros [[? ?] | [? [[? ?] ?]]]; [contradiction| auto].
Qed.

Lemma d_before_f :
  forall E p (L:set Label) w,
  (forall w, writes E w -> L (f (DW w))) ->
  (forall de, L (f de) -> L (d de)) ->
  trans p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->  
  machine_not_stuck E p L ->
  writes E w ->
  tc p (d (DW w)) (f (DW w)).
Proof.
intros E p L w Hlw Hlfd Htp Htotp Hmns Hew.
assert (d (DW w) <> f (DW w)) as Hdiff.
  intro Heq; inversion Heq.
assert (L (d (DW w))) as Hld.
  apply Hlfd; apply Hlw; auto.
assert (L (f (DW w))) as Hlf.
  apply Hlw; auto.
generalize (Htotp (d (DW w)) (f (DW w)) Hdiff Hld Hlf); intro Hor;
  inversion Hor as [|Hfd]; clear Hor; auto.
  apply trc_step; auto.
generalize (Hmns (f (DW w)) Hlf); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
unfold step in Hs; unfold flush_write_step in Hs.
destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs];
destruct (bemp (rrestrict (Q s1) (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs];
destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w))) w) as [Hl|]; [|inversion Hs].
assert (B s1 dB w) as Hi.
  apply pick_last_in_r with (fun e => loc e = loc w); auto.
    apply dB_not_in_E with E; auto. destruct Hew; auto.
assert (udr (B s1) w) as Hudr.
  right; exists dB; auto.
destruct Hew as [Hew ?];
generalize (in_buff_implies_in_order E p L (f (DW w)) n s0 s1 (f (DW w)) w Hew Hm Hudr);
intros [?|Heq]; auto; inversion Heq.
Qed.

Lemma in_buff_bef_aft_implies_not_flush :
  forall E r L s1 s2 l w,
  step E r L (Some s1) l = Some s2 ->
  udr (B s1) w -> udr (B s2) w ->
  l <> f (DW w).
Proof.
intros E r L s1 s2 l w Hs H1 H2 Heq; rewrite Heq in Hs;
unfold step in Hs; unfold flush_write_step in Hs.

  destruct (bemp (rrestrict (rel_union (B s1) (Q s1))
                (fun e : Event =>
                 rel_union (A.ppo E) (A.abc E (p2X E r L)) e w))); [|inversion Hs].
  destruct (bemp
           (rrestrict (Q s1)
              (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
  destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w))) w); [|inversion Hs].
  inversion Hs as [Heqs]; rewrite <- Heqs in H2; simpl in H2.
  
  inversion H2 as [? [x Hi] | ? [x Hi]]; unfold del_buff in Hi;
    destruct Hi as [? [Hc Hc']]; [apply Hc | apply Hc']; auto.
Qed.

Lemma in_buff_implies_f_after :
  forall E (L:set Label) lst p n s0 s1 lb w,
  Included _ (Union _ (dom p) (ran p)) L ->
  (forall w0 : Event, writes E w0 -> L (f (DW w0))) ->
  (forall de : DEvent, L (f de) -> L (d de)) ->
  writes E w ->
  lb <> f (DW w) ->
  L lb -> 
  acyclic p ->
  trans p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  machine_not_stuck E p L -> 
  lst <> f (DW w) ->
  L lst ->
  (~(exists l', tc p lst l' /\ tc p l' lb) \/ lst = lb) ->
  
  (exists r, rel_incl r (pred p lb) /\ 
     (~(exists l', tc p (last r) l' /\ tc p l' lst) ) /\
     machine E r L lst n (Some s0) (Some s1)) ->
  udr (B s1) w ->
  tc p lb (f (DW w)).
Proof.
intros E L lst p n s0 s2 lb w Hudr Hlw Hlfd Hew Hdiff Hlb Hacp Htp Htotp Hmns Hnw Hllst Hlst Hm Hw; 
generalize p lb s2 lst Hdiff Hlb Hacp Htp Htotp Hmns Hnw Hllst Hlst Hm Hw Hudr; 
clear p lb s2 lst Hdiff Hlb Hacp Htp Htotp Hmns Hnw Hllst Hlst Hm Hw Hudr; 
induction n;
intros p lb s2 lst Hdiff Hlb Hacp Htp Htotp Hmns Hnw Hllst Hlst Hm Hw Hudr.

  assert (L (f (DW w))) as Hlf.
    apply Hlw; auto.
  destruct Hm as [r [Hid [Hlr Hm]]]. 
    destruct Hm as [Heqs [? [Hib [? [? Hcw]]]]].
    generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intros [|Hwlb]; [apply trc_step|]; auto.
    inversion Heqs as [Heq]; rewrite <- Heq in Hw; 
    generalize (Hib w); intros [dir ?]; generalize (dir Hw); intro Hidw.
    destruct (eqLabel_dec (last r) (f (DW w))) as [Heql | Hneql].
      generalize (last_in_ran r (f (DW w)) Heql); intro Hifw.
      assert False as Ht.
        apply (Hcw w); split; auto; right; auto.
      inversion Ht.
    assert (p (f (DW w)) (last r)) as Hfwr.
      assert (L (last r)) as Hlastr.
        apply Hudr.
        assert (last r = last r) as Heqt.
          auto.
        generalize (last_in_ran r (last r) Heqt); intros [x Hx];
        generalize (Hid x (last r) Hx); intros [? ?]; right; exists x; auto.
      generalize (Htotp (last r) (f (DW w)) Hneql Hlastr Hlf); intros [Hlrw|]; auto.
      inversion Hlst as [Hni | Heqlst].
        generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw|Hwlst].
        assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
          exists (f (DW w)); split; apply trc_step; auto.
        contradiction.
        assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hc. 
          exists (f (DW w)); split; apply trc_step; auto.
        contradiction.

        assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hc. 
          exists (f (DW w)); split; apply trc_step; auto; rewrite Heqlst; auto.
        contradiction.

    assert (udr r (f (DW w))) as Hifw.
      apply before_last_implies_in with p; auto.
        intros e1 e2 H12; generalize (Hid e1 e2 H12); intros [? ?]; auto.
    assert False as Ht.
      apply (Hcw w); split; auto.
    inversion Ht.

  destruct Hm as [r [Hid [Hlr Hm]]].
  destruct Hm as [lb' [s1 [Hd [Hm [Hs12 ?]]]]]; fold (machine E (pred r lb') L lb' n (Some s0) (Some s1)) in Hm. 
  generalize (excluded_middle (udr (B s1) w)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.

  assert (lb' <> f (DW w)) as Hdiff'.
    apply in_buff_bef_aft_implies_not_flush with E r L s1 s2; auto.
  assert (L lb') as Hlb'.
    apply Hudr.
      generalize (last_in_ran r lb' Hd); intros [x Hr];
      generalize (Hid x lb' Hr); intros [? ?]; left; exists lb; auto.
  assert (~ (exists l' : Label, tc p lb' l' /\ tc p l' lb') \/ lb' = lb') as Hllb'.
    auto.
  assert (exists r0 : Rln Label,
    rel_incl r0 (pred p lb') /\ 
    (~ (exists l' : Label, tc p (last r0) l' /\ tc p l' lb') )/\
    machine E r0 L lb' n (Some s0) (Some s1)) as Hed.
    exists (pred r lb'); split; [|split]; auto.
       intros x y [Hxy Hylb']; 
         generalize (Hid x y Hxy); intros [? ?];
         generalize (Hid y lb' Hylb'); intros [? ?];
         split; auto.
       apply no_interv_last; auto.
         intros x y Hxy; generalize (Hid x y Hxy); intros [? ?]; auto.

  generalize (IHn p lb' s1 lb' Hdiff' Hlb' Hacp Htp Htotp Hmns Hdiff' Hlb' Hllb' Hed Hi Hudr); auto.
  intro Hlb'w.

  assert (L (f (DW w))) as Hlf.
    apply Hlw; auto.
  generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intros [|Hwlb]; [apply trc_step; auto|].
  generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw | Hwlst].
  inversion Hlst as [Hn | Heq].
    assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
      exists (f (DW w)); split; apply trc_step; auto.
    contradiction.
    assert (tc p lb lb) as Hc.
      apply trc_ind with (f (DW w)); apply trc_step; auto; rewrite <- Heq; auto.
    generalize (Hacp lb Hc); intro Ht; inversion Ht.
    assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hco.
      exists (f (DW w)); split; auto.
        rewrite Hd; auto.
        apply trc_step; auto.
    contradiction.

    assert (lb' = d (DW w)) as Hlb'.
      apply b_step with E r L s1 s2; auto.
      destruct Hew; auto.
    assert (tc p lb' (f (DW w))) as Hpdf.
      rewrite Hlb'. apply d_before_f with E L; auto.

  assert (L (f (DW w))) as Hlf.
    apply Hlw; auto.

    generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intro Hor; 
      inversion Hor as [? | Hflb]; clear Hor; apply trc_step; auto.

      rewrite Hd in Hlr.
      generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw | Hwlst].
        inversion Hlst as [Hniw | Heqlstlb].
          assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
            exists (f (DW w)); split; apply trc_step; auto.
          contradiction.
          assert (tc p lb lb) as Hc.
            apply trc_ind with (f (DW w)); rewrite Heqlstlb in Hlstw; apply trc_step; auto.
          generalize (Hacp lb Hc); intro Ht; inversion Ht.

        assert (exists l' : Label, tc p lb' l' /\ tc p l' lst) as Hc.
          exists (f (DW w)); split; auto; apply trc_step; auto.
        contradiction.
Qed.

Lemma in_mem_implies_in_order2 :
  forall E p L lst n s0 s1 lb w l,
  trans p ->
  (exists d, trans d /\ rel_incl d (pred p lb) /\ machine E d L lst n (Some s0) (Some s1)) ->
  w = M s0 l ->
  p (f (DW w)) lb.
Proof.
intros E p L lst n s0 s2 lb w l Htp Hm Hw; 
generalize p L lst lb s2 Htp Hm Hw; clear p L lst lb s2 Htp Hm Hw; induction n;
intros p L lst lb s2 Htp [d [Htd [Hid Hm]]] Hw.
  inversion Hm as [Heq [Him ?]]; inversion Heq as [Heqs].
  generalize (Him l w Hw); intros [x [y Hd] | y [x Hd]];
  generalize (Hid x y Hd); intros [? ?]; auto.
    apply Htp with y; auto.

  destruct Hm as [lb' [s1 [Hd [Hm [Hs12 ?]]]]]; fold (machine E (pred  d lb') L lb' n (Some s0) (Some s1)) in Hm. 
  generalize (excluded_middle (w = M s0 l)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.

    generalize (td_tpred d lb' Htd); intro Htpp.
    assert (exists d0 : Rln Label, trans d0 /\
      rel_incl d0 (pred d lb') /\ machine E d0 L lb' n (Some s0) (Some s1)) as Hed.
      exists (pred d lb'); split; [|split]; auto. intros x y Hxy; auto.
    generalize (IHn d L lb' lb' s1 Htd Hed Hi); intro Hdwlb'.
    generalize (Hid (f (DW w)) lb' Hdwlb'); intros [Hpwlb' Hdlb'].
      apply Htp with lb'; auto. 
      subst; auto.

    assert False as Ht.
      apply Hni; auto.
    inversion Ht.
Qed. 

Lemma in_mem_after_implies_fafter :
  forall E p L lst n s2 s3 w' w l,
  w' = M s2 l -> w = M s3 l -> w <> w' ->
  (exists d, rel_incl d p /\ machine E d L lst n (Some s2) (Some s3)) ->
  trans p ->
  p (f (DW w')) (f (DW w)).
Proof.
intros E p L lst n s2 s3 w' w l Hw' Hw Hdiff Hed Htp;
generalize w w' lst Hdiff s2 s3 Hw' Hw Hed; 
clear w w' lst Hdiff s2 s3 Hw' Hw Hed;
induction n; intros w w' lst Hdiff s2 s3 Hw' Hw [d [Hid Hm]].

   inversion Hm as [Heq]; inversion Heq.
    
    subst; assert False as Ht. 
      apply Hdiff; auto.
    inversion Ht.

  destruct Hm as [lb [s [Hlb [Hm [Hs [Hds Hwfs]]]]]];
  fold (machine E (pred d lb) L lb n (Some s2) (Some s)) in Hm.
  destruct (eqEv_dec w' (M s l)) as [Heq | Hd].
    assert (Some s3 = Some s3) as Ht.
      auto.
    generalize (Hds s3 Ht l); rewrite Heq; rewrite Hw; auto.

    apply Htp with (f (DW (M s l))).
    apply IHn with lb s2 s; auto.
      exists (pred d lb); split; auto.
        intros e1 e2 [? ?]; auto.

    assert (Some s3 = Some s3) as Ht.
      auto.
    generalize (Hds s3 Ht l); rewrite Hw; auto.
Qed.

Lemma in_mem_implies_no_interv_write :
  forall E p L lst n s0 s1 lb w l,
  acyclic p ->
  trans p ->
  machine E (pred p lb) L lst n (Some s0) (Some s1) ->
  wf_mem E (M s1) -> 
  w = M s1 l ->
  ~(exists w', loc w' = l /\ p (f (DW w)) (f (DW w')) /\ p (f (DW w')) lb).
Proof.
intros E p L lst n s0 s3 lb w l Hacp Htp Hm Hwfm Hw [w' [Heql [Hpww' Hpw'lb]]].
assert (rel_incl (pred p lb) (pred p lb)) as Hit.
  intros x y Hxy; auto.
assert (loc w = loc w') as Heqlw.
  rewrite Heql; rewrite Hw.
  destruct Hwfm with l; auto.
generalize (flush_before_implies_before_in_mem E p L lst (pred p lb) n s0 s3 lb w' Hpw'lb Hit Hm);
intros [s2 [Hs2s3 Himw's2]].
assert (p (f (DW w)) lb) as Hpwlb.
  apply Htp with (f (DW w')); auto.
generalize (flush_before_implies_before_in_mem E p L lst (pred p lb) n s2 s3 lb w Hpwlb Hit Hs2s3);
intros [s1 [Hs1s3 Himws1]]; clear Hit.

assert (w <> w') as Hdiff.
  intro Heq; rewrite Heq in Hpww'; apply Hacp with (f (DW w')); apply trc_step; auto.
assert ((exists d, rel_incl d p /\ machine E d L lst n (Some s2) (Some s3))) as Hed.  
  exists (pred p lb); split; auto. 
    intros x y [? ?]; auto. 

rewrite Heql in Himw's2;
generalize (in_mem_after_implies_fafter E p L lst n s2 s3 w' w l Himw's2 Hw Hdiff Hed Htp);
intros Hw'w.

apply Hacp with (f (DW w')); apply trc_ind with (f (DW w)); apply trc_step; auto.
Qed.

Lemma path_uniws : 
  forall E p L e1 e2,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  machine_not_stuck E p L ->  
  (forall e1 e2, po_iico E e1 e2 -> p (d (DW e1)) (d (DW e2))) ->
  (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  pio_llh E e1 e2 -> ~ ws (p2X E p L) e2 e1.
Proof.
intros E p L e1 e2 Hwf Hudr Htp Hacp Hmns Hpod Hfifo H12 Hws.
simpl in Hws; unfold p2ws in Hws; destruct Hws as [Hw1 [Hw2 [Heql [l2 [l1 [Hf2 [Hf1 Hp]]]]]]].

assert (dom p l2) as Hudrp.
  exists l1; auto.
assert (L l2) as Hl2.
  apply Hudr; left; auto.
generalize (Hmns l2 Hl2); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]]; rewrite Hf2 in Hs;
  simpl in Hs; unfold flush_write_step in Hs.
destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2))); [|inversion Hs].
destruct (bemp (rrestrict (Q s1) (fun r' : Event => loc r' = loc e2 /\ po_iico E r' e2))); [|inversion Hs].
destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc e2))) e2) as [Heq | Hneq]; [|inversion Hs].
  unfold pick_last in Heq.
  generalize (EPick.pick_in (fun w : Event =>
         (~(exists w, rrestrict (B s1) (fun e => loc e = loc e2) dB w) /\ w = dB) \/
         ((exists w, rrestrict (B s1) (fun e => loc e = loc e2) dB w) /\ rrestrict (B s1)
           (fun e => loc e = loc e2) dB w /\
         ~(exists w', rrestrict (B s1)
              (fun e : Event => loc e = loc e2) w' w)))); rewrite Heq; intro Hor;
  inversion Hor as [HdB | HndB]; clear Hor.
  assert (events E e2) as He2.
    destruct H12 as [? [? ?]]; apply ABasic.po_iico_range_in_events with e1; auto.
  destruct HdB as [? HdB]; generalize HdB; fold (e2 <> dB); 
  apply dB_not_in_E with E; auto.  

  destruct HndB as [Hi [? Hn]].
  assert (exists w' : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc e2) w' e2) as Hc.
    exists e1; split; [|split; unfold Ensembles.In]; auto.
    apply in_order_implies_in_obuff with E p L l2 n s0 l2; auto.
      apply Hpod; destruct H12 as [? [? ?]]; auto.
      rewrite <- Hf1; auto. apply trc_step; auto.
      exists (pred p l2); split; auto.
        intros x y Hxy; auto.
   contradiction.
Qed.

Lemma path_uniwsrf : 
  forall E p L e1 e2,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p -> 
  (forall w, writes E w -> L (f (DW w))) ->
  (forall de, p (d de) (f de)) ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  machine_not_stuck E p L -> wf_labels E L -> wf_rf_labels E L ->
  pio_llh E e1 e2 -> ~ (rel_seq (ws (p2X E p L)) (rf (p2X E p L))) e2 e1.
Proof.
intros E p L e1 e2 Hwf Hudrp Htp Hacp Hlw Hdf Hpop Htotp Hmns Hwfl Hwfrfl H12 [e [Hws Hrf']].
generalize H12; intros [Heql12 [Hpo12 ?]].
generalize Hrf'; intro Hrf.
simpl in Hws; unfold p2ws in Hws; destruct Hws as [Hw1 [Hw2 [Heqlw [l1 [l2 [Hf1 [Hf2 Hp]]]]]]].
simpl in Hrf; unfold p2rf in Hrf; destruct Hrf as [? Hdl].

assert (L (f (DW e2))) as Hdl2.
  rewrite <- Hf1; apply Hudrp; left; exists l2; auto.  

assert (f (DR (e,e1)) <> f (DW e2)) as Hdiff.
  intros Heq; inversion Heq.

generalize (Htotp (f (DR (e,e1))) (f (DW e2)) Hdiff Hdl Hdl2); intros [Hp12 | Hp21].

destruct (Hmns (f (DR (e,e1))) Hdl) as [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
simpl in Hs; unfold flush_resolved_read_step in Hs.

case_eq (bin (ran (Q s1)) e1); intro Hiq; rewrite Hiq in Hs; [|inversion Hs].

destruct (bemp (rrestrict ((Q s1)) (fun e0 => (dp.dp E) e0 e))); [|inversion Hs].

destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e1))); [|inversion Hs].

destruct (eqEv_dec e (M s1 (loc e1))) as [Him | Hnim].

  assert (exists d, trans d /\ rel_incl d (pred p (f (DR (e,e1)))) /\ machine E d L (f (DR (e,e1))) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (DR (e,e1)))); split; [|split]; auto.
      apply td_tpred; auto.
      intros x y Hxy; auto.
  generalize (in_mem_implies_in_order E p L (f (DR (e,e1))) n s0 s1 (f (DR (e,e1))) e (loc e1) Htp Hed Him);
  intros Hpee1.
    apply Hacp with (f (DR (e,e1))); auto.
      apply trc_ind with (f (DW e)); [|apply trc_step]; auto.
      apply trc_ind with (f (DW e2)); apply trc_step; auto.
      rewrite <- Hf1; rewrite <- Hf2; auto.
  
destruct (bin (Intersection _ (ran (B s1)) (rrB e1)) e); [inversion Hs|].
case_eq (bin (fun w => (B s1 w (waft E (B s1) e1)) /\ loc e1 = loc w) e); intros Haft; [|rewrite Haft in Hs; inversion Hs].

generalize (bin_In (fun w => (B s1 w (waft E (B s1) e1)) /\ loc e1 = loc w) e Haft);
intros [Hb Heql].

generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc e1 = loc w /\
                                    po_iico E e1 w /\
                                    ~ (exists w', loc e1 = loc w' /\ po_iico E e1 w')));
fold (waft E (B s1) e1); intros [Hr [Hl [Hpo Hn]]].

assert (exists w', (loc e1 = loc w' /\ po_iico E e1 w')) as Hc.
  exists e2; split; auto.
contradiction.

destruct (Hmns (f (DW e2)) Hdl2) as [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
simpl in Hs; unfold flush_write_step in Hs.

  destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event =>
               rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2))); [|inversion Hs].
  case_eq (bemp
           (rrestrict (Q s1)
              (fun r' : Event => loc r' = loc e2 /\ po_iico E r' e2))); intros Hiq; 
  [|rewrite Hiq in Hs; inversion Hs].

  generalize (bemp_axt (Q s1) (fun r' : Event => loc r' = loc e2 /\ po_iico E r' e2) Hiq); intro Hn.
  assert (exists e : Event, ran (Q s1) e /\ loc e = loc e2 /\ po_iico E e e2) as Hc.
    exists e1; split; [|split]; auto.
      assert (p (d (DR (e, e1))) (f (DW e2))) as Hp12.
        apply Htp with (d (DW e2)).
         apply Hpop; auto.
         apply Hdf; auto.
      apply in_order_implies_in_queue with E p L (f (DW e2)) n s0 (f (DW e2)) e; auto.       

       exists (pred p (f (DW e2))); split; [|split; [|split]]; auto.
      apply td_tpred; auto. intros x y Hxy; auto.
      apply no_interv_last; auto. intros x y Hxy; auto.

  contradiction.
Qed.

Lemma rf_dp_irr :
  forall E p L x,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  machine_not_stuck E p L ->  
  wf_labels E L -> wf_rf_labels E L ->
  ~ rel_seq (dp.dp E) (rf (p2X E p L)) x x.
Proof.
intros E p L x Hwf Hudrp Htp Hacp Htotp Hmns Hwfl Hwfrfl  Hxx.

assert (linear_strict_order p L) as Hlp.
  split; [|split; [|split]]; auto.
    intros x1 x2 x3 [H12 H23]; apply Htp with x2; auto.
    intros e He; apply Hacp with e; apply trc_step; auto.

generalize (path_rf_wf E p L); intros [Hrf_ex [Hrf_cands Hrf_uni]]; auto.

assert (reads E x) as Hrx.
  destruct Hxx as [w [Hdp Hrf]]; split.
    apply ABasic.po_iico_domain_in_events with w; auto.
      apply dp.dp_valid; auto.
    apply ABasic.ran_rf_is_read with E (p2X E p L) w; auto.

generalize (Hrf_ex x Hrx); intros [wx [Hewx Hrfwx]].

assert (L (f (DR (wx,x)))) as Hlwx.
  simpl in Hrfwx; destruct Hrfwx as [? ?]; auto.

generalize (Hmns (f (DR (wx,x))) Hlwx); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
  simpl in Hs; unfold flush_resolved_read_step in Hs.

case_eq (bin (ran (Q s1)) x); intro Hbinx; rewrite Hbinx in Hs; [|inversion Hs].
case_eq (bemp (rrestrict (Q s1) (fun e0 => (dp.dp E) e0 wx))); 
  intro Hbemp; rewrite Hbemp in Hs; [|inversion Hs]; clear Hs.
generalize (bemp_axt ((Q s1)) (fun e0 => (dp.dp E) e0 wx) Hbemp); intro Hn.
assert (exists e0, ran (Q s1) e0 /\ ((dp.dp E) e0 wx)) as Hy.
  exists x; split; auto.
    apply bin_In; auto.
    destruct Hxx as [w [Hdp Hrf]].
    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [? Huni]].
    generalize (Huni x wx w Hrfwx Hrf); intro Heq; rewrite Heq; auto.

contradiction.
Qed.

Lemma rf_dp_implies_in_p :
  forall E p L w1 w2 r1 r2,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L -> wf_labels E L -> wf_rf_labels E L ->
  L (f (DR (w1,r1))) -> L (f (DR (w2,r2))) ->
  (rel_seq (dp.dp E) (rf (p2X E p L))) r1 r2 -> 
  p (f (DR (w1,r1))) (f (DR (w2,r2))).
Proof.
intros E p L w1 w2 r1 r2 Hwf Hudrp Htp Hacp Htotp Hdf Hmns Hwfl Hwfrfl Hll1 Hll2 H12.
generalize H12; intros [w [Hdp1 Hrf2]].
assert (linear_strict_order p L) as Hlp.
  split; [|split; [|split]]; auto.
    intros e1 e2 e3 [He12 He23]; apply Htp with e2; auto.
    intros e He; apply Hacp with e; apply trc_step; auto.

assert ((f (DR (w1,r1))) <> (f (DR (w2,r2)))) as Hdiff.
  intro Heq; inversion Heq; subst. 
    generalize H12; apply rf_dp_irr; auto.
generalize (Htotp (f (DR (w1,r1))) (f (DR (w2,r2))) Hdiff Hll1 Hll2); intro Hor; 
  inversion Hor as [|H21]; clear Hor; auto.

generalize (Hmns (f (DR (w2,r2))) Hll2).
intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

  simpl in Hs; unfold flush_resolved_read_step in Hs.

destruct (bin (ran (Q s1)) r2); [|inversion Hs].
case_eq (bemp (rrestrict ((Q s1)) (fun e => (dp.dp E) e w2))); 
  intro Hbemp; rewrite Hbemp in Hs; [|inversion Hs]; clear Hs.
generalize (bemp_axt ((Q s1)) (fun e => (dp.dp E) e w2) Hbemp); intro Hn.
assert (exists e, ran ((Q s1)) e /\ ((dp.dp E) e w2)) as Hy.
  exists r1; split; auto.
    assert (p (d (DR (w1,r1))) (f (DR (w2,r2)))) as Hp12.
      admit. 
    apply in_order_implies_in_queue with E p L (f (DR (w2,r2))) n s0 (f (DR (w2,r2))) w1; auto.
      exists (pred p (f (DR (w2,r2)))); split; [apply td_tpred | 
                                          split; [intros x y Hxy |]]; auto.
        split; auto. apply no_interv_last; auto; intros x y Hxy; auto.

  assert (L (f (DR (w,r2)))) as Hllw.
    destruct Hrf2 as [? ?]; auto.
  destruct Hwfrfl as [? [? Huni]];
  rewrite (Huni r2 w2 w Hll2 Hllw); auto.
 
contradiction.
Qed.

Lemma tc_rf_dp_implies_in_tc_p :
  forall E p L w1 w2 r1 r2,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L -> wf_labels E L ->
  wf_rf_labels E L -> L (f (DR (w1,r1))) -> L (f (DR (w2,r2))) ->
  (evt_of_devt (DR (w1,r1))) = r1 -> 
  (evt_of_devt (DR (w2,r2))) = r2 ->
  tc(rel_seq (dp.dp E) (rf (p2X E p L))) r1 r2 -> 
  tc p (f (DR (w1,r1))) (f (DR (w2,r2))).
Proof.
intros E p L w1 w2 r1 r2 Hwf Hudrp Htp Hacp Htotp Hdf Hmns Hwfl Hwfrfl Hl1 Hl2 H1 H2 H12.
assert (linear_strict_order p L) as Hlp.
  split; [|split; [|split]]; auto.
    intros e1 e2 e3 [He12 He23]; apply Htp with e2; auto.
    intros x Hx; apply Hacp with x; apply trc_step; auto.
generalize w1 w2 H1 H2 Hl1 Hl2; clear w1 w2 H1 H2 Hl1 Hl2; 
induction H12 as [x y Hs |]. 
  intros wx wy Hx Hxy Hlx Hly; apply trc_step.
     apply rf_dp_implies_in_p with E L; auto.

  intros wx wy Hx Hy Hlx Hly.
  assert (reads E z) as Hrz.
    clear Hly Hlx y H12_0 IHtc1 IHtc2 Hx Hy.
    induction H12_; auto.
      destruct H as [w [Hdp Hrf]].
        split.
          apply ABasic.ran_rf_in_events with (p2X E p L) w; auto.
            apply path_rf_wf; auto; apply wfl_implies_wfr; auto.
          apply ABasic.ran_rf_is_read with E (p2X E p L) w; auto.
            generalize (path_rf_wf E p L); intros [? [? ?]]; auto.

  generalize (path_rf_wf E p L); auto; intros [Hrf_ex ?]; auto.
  generalize (Hrf_ex z Hrz); intros [wz [Hlwz [? ?]]].
  assert (evt_of_devt (DR (wz,z)) = z) as Hz.
    simpl; auto.
  apply trc_ind with (f (DR (wz,z))); [apply IHtc1 | apply IHtc2]; auto.
Qed.

Lemma path_thin :
  forall E p L,
  well_formed_event_structure E ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L -> wf_labels E L -> wf_rf_labels E L ->
  (forall e1 e2, 
     rel_seq (dp.dp E) (rf (p2X E p L)) e1 e2 -> ~ (tc (rel_seq (dp.dp E) (rf (p2X E p L))) e2 e1)).
Proof.
intros E p L Hwf Hudrp Htp Hacp Htotp Hdf Hmns Hwfl Hwfrfl e1 e2  H12 H21.
assert (linear_strict_order p L) as Hlp.
  split; [|split; [|split]]; auto.
    intros x1 x2 x3 [Hx12 Hx23]; apply Htp with x2; auto.
    intros x Hx; apply Hacp with x; apply trc_step; auto.
generalize H12; intros [w' [Hrf Hdp]].
assert (reads E e1) as Hr1.
  generalize (dp.dp_valid E); intros [? [? Hdpr]]; apply Hdpr with w'; auto.
assert (exists w1, L (f (DR (w1,e1)))) as Hw1.
  generalize (path_rf_wf E p L); intros [Hrf_ex ?]; auto.
  generalize (Hrf_ex e1 Hr1); intros [w1 [? [? ?]]]; exists w1; auto.
assert (reads E e2) as Hr2.
  split.
    apply ABasic.ran_rf_in_events with (p2X E p L) w'; auto.
      apply path_rf_wf; auto; apply wfl_implies_wfr; auto.
    apply ABasic.ran_rf_is_read with E (p2X E p L) w'; auto.
      generalize (path_rf_wf E p L); intros [? [? ?]]; auto; 
        apply wfl_implies_wfr; auto.
assert (exists w2, L (f (DR (w2,e2)))) as Hw2.
  exists w'; destruct Hdp; auto.
destruct Hw1 as [w1 Hl1]; destruct Hw2 as [w2 Hl2]. 
generalize H12; intros [w [Hdp1 Hrf2]].
assert (events E e1) as He1.
  apply ABasic.po_iico_domain_in_events with w; auto.
    apply dp.dp_valid; auto.
assert (events E e2) as He2.
  apply ABasic.ran_rf_in_events with (p2X E p L) w; auto. 
    apply (path_rf_wf E p L); auto; apply wfl_implies_wfr; auto.
generalize (rf_dp_implies_in_p E p L w1 w2 e1 e2 Hwf Hudrp Htp Hacp Htotp Hdf Hmns Hwfl Hwfrfl Hl1 Hl2 H12); intro Hp12.
assert (evt_of_devt (DR (w1,e1)) = e1) as Hde1.
  simpl; auto.
assert (evt_of_devt (DR (w2,e2)) = e2) as Hde2.
  simpl; auto.
generalize (tc_rf_dp_implies_in_tc_p E p L w2 w1 e2 e1 Hwf Hudrp Htp Hacp Htotp Hdf Hmns Hwfl Hwfrfl Hl2 Hl1 Hde2 Hde1 H21); intro Hp21.
assert (tc p (f (DR (w1,e1))) (f (DR (w1,e1)))) as Htcp1.
  apply trc_ind with (f (DR (w2,e2))); auto; apply trc_step; auto.
generalize Htcp1; fold (~(tc p (f (DR (w1,e1))) (f (DR (w1,e1))))); 
  apply (Hacp (f (DR (w1,e1)))).
Qed.

Lemma dom_tc_ghb_in_evts :
  forall E X x y,
  well_formed_event_structure E ->
  write_serialization_well_formed (events E) (ws X) ->
  rfmaps_well_formed E (events E) (rf X) ->
  tc (AWmm.ghb E X) x y ->
  events E x.
Proof.
intros E X x y Hwf Hwswf Hrfwf Hxy; induction Hxy as [x y Hghb|]; auto.
apply ABasic.ghb_domain_in_events with X y; auto.  
Qed.

Lemma ran_tc_ghb_in_evts :
  forall E X x y,
  well_formed_event_structure E ->
  write_serialization_well_formed (events E) (ws X) ->
  rfmaps_well_formed E (events E) (rf X) ->
  tc (AWmm.ghb E X) x y ->
  events E y.
Proof.
intros E X x y Hwf Hwswf Hrfwf Hxy; induction Hxy as [x y Hghb|]; auto.
apply ABasic.ghb_range_in_events with X x; auto.  
Qed.

Lemma ghb_implies_diff :
  forall E p L x y,
  well_formed_event_structure E ->
  linear_strict_order p L ->
  wf_labels E L -> wf_rf_labels E L ->
  AWmm.ghb E (p2X E p L) x y ->
  x <> y.
Proof.
intros E p L x y Hwf Hlp Hwfr Hwfrfl Hxy.
assert (write_serialization_well_formed (events E) (ws (p2X E p L)) /\
        rfmaps_well_formed E (events E) (rf (p2X E p L))) as Hand.
  split; [apply path_ws_wf | apply path_rf_wf]; auto.
    apply wfl_w; auto.
generalize (ABasic.ghb_range_in_events (p2X E p L) x y Hwf Hand Hxy); intro Hey.
unfold AWmm.ghb in Hxy; case_eq A.inter; case_eq A.intra; intros intra inter; 
rewrite intra in Hxy; rewrite inter in Hxy.

inversion Hxy as [Hrfe | Hr]; clear Hxy.
  destruct Hrfe as [Hrf ?].

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]]; generalize (Hrfc x y Hrf); intro Hrfm.
    destruct Hrfm as [? [? [a [Hwx [Hry ?]]]]].
      intro Heq; rewrite Heq in Hwx; destruct Hwx as [? Hax];
      destruct Hry as [? Hay]; rewrite Hax in Hay; inversion Hay.

  inversion Hr as [Hrfi | Hre]; clear Hr.
  destruct Hrfi as [Hrf ?].
    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]]; generalize (Hrfc x y Hrf); intro Hrfm.
    destruct Hrfm as [? [? [a [Hwx [Hry ?]]]]].  
      intro Heq; rewrite Heq in Hwx; destruct Hwx as [? Hax];
      destruct Hry as [? Hay]; rewrite Hax in Hay; inversion Hay.

    inversion Hre as [Hab | Hres]; clear Hre.
      intro Heq; rewrite Heq in Hab; generalize Hab; apply A.ab_irr.
      inversion Hres as [Hwsfr | Hppo]; clear Hres.
        inversion Hwsfr as [Hws | Hfr]; clear Hwsfr.
          destruct Hws as [Hwx [Hwy [? [lx [ly [Hx [Hy Hp]]]]]]].
            intro Heq; rewrite Heq in Hx; rewrite Hx in Hp; rewrite Hy in Hp.
            destruct_lin Hlp; apply Hac with (f (DW y)); auto.
          destruct Hfr as [? [? [w [Hrf Hws]]]].
            destruct Hand as [[? Hwtl] [? [Hrfc ?]]].
            generalize (Hwtl w y Hws); intros [? [? [? [? Hay]]]].
            generalize (Hrfc w x Hrf); intros [? [? [? [? [[? Hax] ?]]]]].
            intro Heq; rewrite Heq in Hax; rewrite Hax in Hay; inversion Hay.
        intro Heq; rewrite Heq in Hppo; generalize (A.ppo_valid Hppo); apply ABasic.po_ac; auto.

  inversion Hxy as [Hrfi | Hre]; clear Hxy.
  destruct Hrfi as [Hrf ?].

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]]; generalize (Hrfc x y Hrf); intro Hrfm.
    destruct Hrfm as [? [? [a [Hwx [Hry ?]]]]].
      intro Heq; rewrite Heq in Hwx; destruct Hwx as [? Hax];
      destruct Hry as [? Hay]; rewrite Hax in Hay; inversion Hay.

    inversion Hre as [Hab | Hres]; clear Hre.
      intro Heq; rewrite Heq in Hab; generalize Hab; apply A.ab_irr.
      inversion Hres as [Hwsfr | Hppo]; clear Hres.
        inversion Hwsfr as [Hws | Hfr]; clear Hwsfr.
          destruct Hws as [Hwx [Hwy [? [lx [ly [Hx [Hy Hp]]]]]]].
            intro Heq; rewrite Heq in Hx; rewrite Hx in Hp; rewrite Hy in Hp.
            destruct_lin Hlp; apply Hac with (f (DW y)); auto.
          destruct Hfr as [? [? [w [Hrf Hws]]]].
            destruct Hand as [[? Hwtl] [? [Hrfc ?]]].
            generalize (Hwtl w y Hws); intros [? [? [? [? Hay]]]].
            generalize (Hrfc w x Hrf); intros [? [? [? [? [[? Hax] ?]]]]].
            intro Heq; rewrite Heq in Hax; rewrite Hax in Hay; inversion Hay.
        intro Heq; rewrite Heq in Hppo; generalize (A.ppo_valid Hppo); apply ABasic.po_ac; auto.

  inversion Hxy as [Hrfe | Hre]; clear Hxy.
  destruct Hrfe as [Hrf ?].

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]]; generalize (Hrfc x y Hrf); intro Hrfm.
    destruct Hrfm as [? [? [a [Hwx [Hry ?]]]]].
      intro Heq; rewrite Heq in Hwx; destruct Hwx as [? Hax];
      destruct Hry as [? Hay]; rewrite Hax in Hay; inversion Hay.

    inversion Hre as [Hab | Hres]; clear Hre.
      intro Heq; rewrite Heq in Hab; generalize Hab; apply A.ab_irr.
      inversion Hres as [Hwsfr | Hppo]; clear Hres.
        inversion Hwsfr as [Hws | Hfr]; clear Hwsfr.
          destruct Hws as [Hwx [Hwy [? [lx [ly [Hx [Hy Hp]]]]]]].
            intro Heq; rewrite Heq in Hx; rewrite Hx in Hp; rewrite Hy in Hp.
            destruct_lin Hlp; apply Hac with (f (DW y)); auto.
          destruct Hfr as [? [? [w [Hrf Hws]]]].
            destruct Hand as [[? Hwtl] [? [Hrfc ?]]].
            generalize (Hwtl w y Hws); intros [? [? [? [? Hay]]]].
            generalize (Hrfc w x Hrf); intros [? [? [? [? [[? Hax] ?]]]]].
            intro Heq; rewrite Heq in Hax; rewrite Hax in Hay; inversion Hay.
        intro Heq; rewrite Heq in Hppo; generalize (A.ppo_valid Hppo); apply ABasic.po_ac; auto.

    inversion Hxy as [Hab | Hres]; clear Hxy.
      intro Heq; rewrite Heq in Hab; generalize Hab; apply A.ab_irr.
      inversion Hres as [Hwsfr | Hppo]; clear Hres.
        inversion Hwsfr as [Hws | Hfr]; clear Hwsfr.
          destruct Hws as [Hwx [Hwy [? [lx [ly [Hx [Hy Hp]]]]]]].
            intro Heq; rewrite Heq in Hx; rewrite Hx in Hp; rewrite Hy in Hp.
            destruct_lin Hlp; apply Hac with (f (DW y)); auto.
          destruct Hfr as [? [? [w [Hrf Hws]]]].
            destruct Hand as [[? Hwtl] [? [Hrfc ?]]].
            generalize (Hwtl w y Hws); intros [? [? [? [? Hay]]]].
            generalize (Hrfc w x Hrf); intros [? [? [? [? [[? Hax] ?]]]]].
            intro Heq; rewrite Heq in Hax; rewrite Hax in Hay; inversion Hay.
        intro Heq; rewrite Heq in Hppo; generalize (A.ppo_valid Hppo); apply ABasic.po_ac; auto.
Qed.

Lemma ghb_irr :
  forall E p L x,
  well_formed_event_structure E ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->  
  wf_labels E L -> wf_rf_labels E L ->
  ~ AWmm.ghb E (p2X E p L) x x.
Proof.
intros E p L x Hwf Hl Hmns Hwfl Hwfrfl Hx.
generalize (ghb_implies_diff E p L x x Hwf Hl  Hwfl Hwfrfl Hx); auto.
Qed.

Lemma ppo_implies_p_flush_read :
  forall E p L  n s0 s1 s2 dx dy x wy ry,
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  evt_of_devt dx = x -> evt_of_devt dy = ry ->
  dy = DR (wy,ry) ->
  A.ppo E x ry ->
  p (f dy) (f dx) ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall de, L (f de) -> p (d de) (f de)) ->
  machine E (pred p (f dy)) L  (f dy) n (Some s0) (Some s1) ->
  flush_resolved_read_step E p L s1 wy ry = Some s2 ->
  False.
Proof.
intros E p L  n s0 s1 s2 dx dy x wy ry Hwfl Hwfrfl Hlp Hdx Hdy Heqdy Hxy Hyx Hpop Hdf Hm Hs.
rewrite <- Hdy in Hxy; unfold flush_resolved_read_step in Hs.
    destruct (bin (ran (Q s1)) ry); [|inversion Hs];
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e wy))); [|inversion Hs].
    case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry))); 
      [intros Hbemp | intros Hbnemp; rewrite Hbnemp in Hs; inversion Hs].
      generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry) Hbemp); intro Hn.
      assert (exists e, ran (rel_union (B s1) (Q s1)) e /\ rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry) as Hy.
        exists x; split; [|left]; auto.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ 
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
          machine E d L  (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto.

    assert (exists d : Rln Label,
    trans d /\
    rel_incl d (pred p (f dy)) /\
    ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
    machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
      exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto.
      intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto. 
  case_eq dx.
    intros [w r] Hdr.
    assert (reads E r) as Her.
      destruct Hwfl as [? [[Hrd Hrf] ?]]; generalize (Hrf w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.

    rewrite Hdr in Hyx.
    assert (p (f dy) (f (DR (w,r))) \/ f dy = f (DR (w,r))) as Horyx.
      auto.

    assert (L (f dy)) as Hldy.
      apply Hudrp; left; exists (f (DR (w,r))); auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.
    assert (p (d (DR (w,r))) (f dy)) as Hpxy.
      apply Htp with (d dy).
        rewrite <- Hdr. 
          rewrite <- Hdx in Hxy; apply Hpop; apply A.ppo_valid; auto.
        apply Hdf; auto.
    assert (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      clear Hdr; destruct_lin Hlp; auto.
    generalize (in_order_implies_in_queue E p L (f dy) n s0 s1 (f dy) w r Hudrp Htp Hacp Htotp Hwfrfl Hldy Hdf Hpxy Hpxy Horyx Hni Hed);
    intros [e Hrr]; exists e; right; rewrite <- Hdx; rewrite Hdr; auto.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l : Label, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      rewrite <- Hdw; apply Htp with (d dy); auto.
        apply Hpop; auto; rewrite <- Hdx in Hxy; apply A.ppo_valid; auto.
        apply Hdf; auto; apply Hudrp; left; exists (f (DW w)); auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.
    generalize (in_order_implies_in_buff E p L  (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed');
    intros Hrr; exists dB; left; rewrite <- Hdx; rewrite Hdw; auto.
      rewrite Heqdy in Hxy; simpl in Hxy; auto.

      contradiction.
Qed.

Lemma ppo_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall de, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  L (f dy) ->
  A.ppo E x y ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hpop Hdf Hmns Hdx Hdy Hly Hxy Hyx.
generalize (Hmns (f dy) Hly); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

case_eq dy; intros cy Hcy; rewrite Hcy in Hs; simpl in Hs; rewrite Hcy in Hdy; simpl in Hdy.

  destruct cy as [wy ry];
  apply ppo_implies_p_flush_read with E p L  n s0 s1 s2 dx dy x wy ry; auto.
    rewrite Hcy; simpl; auto. rewrite Hdy; auto.

  rewrite <- Hdy in Hxy; unfold flush_write_step in Hs.
    case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy))); 
      [intros Hbemp | intros Hbnemp; rewrite Hbnemp in Hs; inversion Hs].
      generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy) Hbemp); intro Hn.
      assert (exists e, ran (rel_union (B s1) (Q s1)) e /\ rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy) as Hy.
        exists x; split; [|left]; auto.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ 
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
          machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto.
  assert (exists d, rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; auto.
      intros e1 e2 H12; auto.
  case_eq dx.
    intros [w r] Hdr.
    assert (reads E r) as Her.
      destruct Hwfl as [? [[Hrd Hrf] ?]]; generalize (Hrf w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.

    rewrite Hdr in Hyx.
    assert (p (f dy) (f (DR (w,r))) \/ f dy = f (DR (w,r))) as Horyx.
      auto.
    assert (L (f dy)) as Hldy.
      apply Hudrp; left; exists (f (DR (w,r))); auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.
    assert (p (d (DR (w,r))) (f dy)) as Hpxy.
      apply Htp with (d dy).
        apply Hpop; auto.
          rewrite <- Hdr; rewrite Hdx; rewrite Hcy; simpl; auto.
          apply A.ppo_valid; auto.
        apply Hdf; auto.
    assert (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      clear Hdr; destruct_lin Hlp; auto.
    generalize (in_order_implies_in_queue E p L (f dy) n s0 s1 (f dy) w r Hudrp Htp Hacp Htotp Hwfrfl Hldy Hdf Hpxy Hpxy Horyx Hni Hed);
      intros [e Hrr]; exists e; right; rewrite <- Hdx; rewrite Hdr; auto.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l : Label, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      rewrite <- Hdw; apply Htp with (d dy); auto.
        apply Hpop; auto; rewrite <- Hdx in Hxy; rewrite Hcy; apply A.ppo_valid; auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.

    generalize (in_order_implies_in_buff E p L (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed);
    intros Hrr; exists dB; left; rewrite <- Hdx; rewrite Hdw; auto. 
      contradiction.
Qed.  

Lemma in_buff_implies_f_after2 :
  forall E p (L:set Label) lst n s0 s1 lb w,
  Included _ (Union _ (dom p) (ran p)) L ->
  
  
  writes E w ->
  lb <> f (DW w) ->
  L lb -> 
  lst <> f (DW w) ->
  L lst ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall w, writes E w -> tc p (d (DW w)) (f (DW w))) ->
  (~(exists l', tc p lst l' /\ tc p l' lb) \/ lst = lb) ->
  (exists r, rel_incl r (pred p lb) /\ 
     (~(exists l', tc p (last r) l' /\ tc p l' lst) \/ last r = lst) /\
     machine E r L lst n (Some s0) (Some s1)) ->
  udr (B s1) w ->
  tc p lb (f (DW w)).
Proof.
intros E p L lst n s0 s2 lb w Hudr   Hew Hdiff Hlb Hnw Hllst Htp Hacp Htotp Hdf Hlst Hm Hw; 
generalize p lst lb s2 Hdiff Hlb Hnw Hllst Htp Hacp Htotp Hdf Hlst Hm Hw Hudr ; 
clear p lst lb s2 Hdiff Hlb Hnw Hllst Htp Hacp Htotp Hdf Hlst Hm Hw Hudr ; 
induction n;
intros p lst lb s2 Hdiff Hlb Hnw Hllst Htp Hacp Htotp Hdf Hlst Hm Hw Hudr .

  assert (L (f (DW w))) as Hlf.
    apply Hudr; rewrite dom_ran_tc; auto; right; exists (d (DW w)); apply Hdf; auto.
  destruct Hm as [r [Hid [Hlr Hm]]]. 
    destruct Hm as [Heqs [? [Hib [? [? Hcw]]]]].
    generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intros [|Hwlb]; [apply trc_step|]; auto.
    inversion Heqs as [Heq]; rewrite <- Heq in Hw; 
    generalize (Hib w); intros [dir bak]; generalize (dir Hw); intro Hidw.
    destruct (eqLabel_dec (last r) (f (DW w))) as [Heql | Hneql].
      generalize (last_in_ran r (f (DW w)) Heql); intro Hifw.
      assert False as Ht.
        apply (Hcw w); split; auto; right; auto.
      inversion Ht.
    assert (p (f (DW w)) (last r)) as Hfwr.
      assert (L (last r)) as Hlastr.
        apply Hudr.
        assert (last r = last r) as Heqt.
          auto.
        generalize (last_in_ran r (last r) Heqt); intros [x Hx];
        generalize (Hid x (last r) Hx); intros [? ?]; right; exists x; auto.
      generalize (Htotp (last r) (f (DW w)) Hneql Hlastr Hlf); intros [Hlrw|]; auto.
      inversion Hlst as [Hni | Heqlst]. 
        inversion Hlr as [Hnir | Heqlr].
        generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw|Hwlst].
        assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
          exists (f (DW w)); split; apply trc_step; auto.
        contradiction.
        assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hc. 
          exists (f (DW w)); split; apply trc_step; auto.
        contradiction.
        assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.  
          rewrite <- Heqlr; exists (f (DW w)); split; apply trc_step; auto.
        contradiction.

        inversion Hlr as [Hnir | Heqlr].
        assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hc. 
          exists (f (DW w)); split; apply trc_step; auto; rewrite Heqlst; auto.
        contradiction.
        rewrite Heqlst in Heqlr.
        assert (tc p lb lb) as Hc.
          apply trc_ind with (f (DW w)); [rewrite Heqlr in Hlrw|]; apply trc_step; auto.
        generalize (Hacp lb Hc); intro Ht; inversion Ht.

    assert (udr r (f (DW w))) as Hifw.
      apply before_last_implies_in with p; auto.
        intros e1 e2 H12; generalize (Hid e1 e2 H12); intros [? ?]; auto.
    assert False as Ht.
      apply (Hcw w); split; auto.
    inversion Ht.

  destruct Hm as [r [Hid [Hlr Hm]]].
  destruct Hm as [lb' [s1 [Hd [Hm [Hs12 ?]]]]]; fold (machine E (pred r lb') L lb' n (Some s0) (Some s1)) in Hm. 
  generalize (excluded_middle (udr (B s1) w)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.

  assert (lb' <> f (DW w)) as Hdiff'.
    apply in_buff_bef_aft_implies_not_flush with E r L s1 s2; auto.
  assert (L lb') as Hlb'.
    apply Hudr.
      generalize (last_in_ran r lb' Hd); intros [x Hr];
      generalize (Hid x lb' Hr); intros [? ?]; left; exists lb; auto.
  assert (~ (exists l' : Label, tc p lb' l' /\ tc p l' lb') \/ lb' = lb') as Hllb'.
    auto.
  assert (exists r0 : Rln Label,
    rel_incl r0 (pred p lb') /\ 
    (~ (exists l' : Label, tc p (last r0) l' /\ tc p l' lb') \/ last r0 = lb') /\
    machine E r0 L lb' n (Some s0) (Some s1)) as Hed.
    exists (pred r lb'); split; [|split]; auto.
       intros x y [Hxy Hylb'];  
       generalize (Hid x y Hxy); intros [? ?]; 
       generalize (Hid y lb' Hylb'); intros [? ?]; split; auto.
         left; apply no_interv_last; auto.
           intros x y Hxy; generalize (Hid x y Hxy); intros [? ?]; auto.

  generalize (IHn p lb' lb' s1 Hdiff' Hlb' Hdiff' Hlb' Htp Hacp Htotp Hdf Hllb' Hed Hi Hudr); auto.
  intro Hlb'w.

  assert (L (f (DW w))) as Hlf.
    apply Hudr; rewrite dom_ran_tc; right; exists lb'; auto.
  generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intros [|Hwlb]; [apply trc_step; auto|].
  generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw | Hwlst].
  inversion Hlst as [Hn | Heq].
    assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
      exists (f (DW w)); split; apply trc_step; auto.
    contradiction.
    assert (tc p lb lb) as Hc.
      apply trc_ind with (f (DW w)); apply trc_step; auto; rewrite <- Heq; auto.
    generalize (Hacp lb Hc); intro Ht; inversion Ht.
    inversion Hlr as [Hllr | Heqlr].
    assert (exists l' : Label, tc p (last r) l' /\ tc p l' lst) as Hco.
      exists (f (DW w)); split; auto.
        rewrite Hd; auto.
        apply trc_step; auto.
    contradiction.
    rewrite Heqlr in Hd; rewrite Hd in Hwlst.
    assert (tc p lb' lb') as Hc.
      apply trc_ind with (f (DW w)); auto; apply trc_step; auto.
    generalize (Hacp lb' Hc); intro Ht; inversion Ht.    

    assert (lb' = d (DW w)) as Hlb'.
      apply b_step with E r L s1 s2; auto.
      destruct Hew; auto.
    assert (tc p lb' (f (DW w))) as Hpdf.
      rewrite Hlb'. apply Hdf; auto.

  assert (L (f (DW w))) as Hlf.
    apply Hudr; rewrite dom_ran_tc; right; exists lb'; auto.

    generalize (Htotp lb (f (DW w)) Hdiff Hlb Hlf); intro Hor; 
      inversion Hor as [? | Hflb]; clear Hor; apply trc_step; auto.

      rewrite Hd in Hlr.
      generalize (Htotp lst (f (DW w)) Hnw Hllst Hlf); intros [Hlstw | Hwlst].
        inversion Hlst as [Hniw | Heqlstlb].
          assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
            exists (f (DW w)); split; apply trc_step; auto.
          contradiction.
          assert (tc p lb lb) as Hc.
            apply trc_ind with (f (DW w)); rewrite Heqlstlb in Hlstw; apply trc_step; auto.
          generalize (Hacp lb Hc); intro Ht; inversion Ht.

        inversion Hlr as [Hllr | Heqlr].
        assert (exists l' : Label, tc p lb' l' /\ tc p l' lst) as Hc.
          exists (f (DW w)); split; auto; apply trc_step; auto.
        contradiction.

        rewrite Heqlr in Hpdf.
        assert (tc p lst lst) as Hc.
          apply trc_ind with (f (DW w)); auto; apply trc_step; auto.
          generalize (Hacp lst Hc); intro Ht; inversion Ht.
Qed.

Lemma grfi_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de : DEvent, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  A.intra = true ->
  L (f dy) ->
  rf_intra (p2X E p L) x y ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hdf Hmns Hdx Hdy Hgrfi Hly Hxy Hyx.

assert (dom p (f dy)) as Hudr.
  exists (f dx); auto. 
generalize (Hmns (f dy) Hly); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

generalize Hlp; intro Hlp'.

case_eq dy; intros cy Hcy; rewrite Hcy in Hs; simpl in Hs; rewrite Hcy in Hdy; simpl in Hdy.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ 
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
          machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto; intros e1 e2 H12; auto.
      intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto.

  destruct cy as [wy ry]; rewrite <- Hdy in Hxy; unfold flush_resolved_read_step in Hs.
  assert (rf (p2X E p L) x ry) as Hrfxry.
    destruct Hxy; auto.

  assert (rf (p2X E p L) wy ry) as Hrfy.
    simpl; unfold p2rf.
      destruct Hrfxry; split; auto.
      destruct Hwfl as [? [[Hrd Hrf] ?]]; apply Hrf; rewrite <- Hcy; auto.
      rewrite <- Hcy; auto.

  assert (x = wy) as Heqxy.
    destruct Hxy as [Hrf Hintra]; generalize (path_rf_wf E p L); auto; intros [? [? Hrfu]]; auto.

    apply Hrfu with ry; auto.

  destruct (bin (ran (Q s1)) ry); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e wy))); [|inversion Hs];
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry))); [|inversion Hs].

  destruct (eqEv_dec wy (M s1 (loc ry))) as [Him|Hnim].
    rewrite Hcy in Hed;
    generalize (in_mem_implies_in_order E p L (f (DR (wy,ry))) n s0 s1 (f (DR (wy,ry))) wy (loc ry) Htp Hed Him);
    intros Hpwr.

  case_eq dx.
    intros [w r] Hdr.

    assert (reads E r) as Her.
      destruct Hwfl as [? [[Hrd Hrf] ?]]; generalize (Hrf w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]];
    destruct Hxy as [Hrf ?]; generalize (Hrfc x ry Hrf); intros [? [? [? [[? Haw] ?]]]].
    destruct Her as [? [? [? Har]]].
    rewrite Hdr in Hdx; simpl in Hdx; subst. 
    rewrite Haw in Har; inversion Har.

    intros w Hdw.
    rewrite Hdw in Hdx.
    simpl in Hdx. 
    rewrite Hdw in Hyx; rewrite Hdx in Hyx; rewrite Heqxy in Hyx.
    rewrite Hcy in Hyx. 
      destruct_lin Hlp; apply Hac with (f (DW wy)). 
        apply Htrans with (f (DR (wy,ry))); auto.

  case_eq (bin (Intersection _ (ran (B s1)) (rrB ry)) wy); intro Hbemp; rewrite Hbemp in Hs; [inversion Hs|].
    assert (bin (Intersection _ (ran (B s1)) (rrB ry)) wy = true) as Hc.
      apply In_bin.

  case_eq dx.
    intros [w r] Hdr.

    assert (reads E r) as Her.
      destruct Hwfl as [? [[Hrd Hrf] ?]]; generalize (Hrf w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]];
    destruct Hxy as [Hrf ?]; generalize (Hrfc x ry Hrf); intros [? [? [? [[? Haw] ?]]]].
    destruct Her as [? [? [? Har]]].
    rewrite Hdr in Hdx; simpl in Hdx; subst. 
    rewrite Haw in Har; inversion Har.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l : Label, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    split.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      admit. 
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.

    generalize (in_order_implies_in_buff E p L (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed');
    intros Hrr. 
      exists dB. 
        rewrite <- Heqxy; rewrite <- Hdx; rewrite Hdw; auto.

        split; auto.
          apply ABasic.rf_implies_same_loc2 with E (p2X E p L); auto.
            split; [apply path_ws_wf|apply path_rf_wf]; auto.
              apply wfl_w; auto.
          rewrite Hgrfi; auto.
          case_eq A.inter; intro; auto. 
            destruct Hxy; rewrite <- Heqxy; auto. 
            destruct Hxy; rewrite <- Heqxy; auto.             
    rewrite Hbemp in Hc; inversion Hc. 

  rewrite <- Hdy in Hxy.
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
  destruct Hwfl as [[Hwd Hw_f] [[Hrd Hrf] ?]].
    rewrite Hcy in Hudr; destruct_lin Hlp.
    assert ((Union Label (dom p) (ran p)) (f (DW cy))) as Hi.
      left; auto.
    generalize (Hdr (f (DW cy)) Hi); intro Hl.
    generalize (Hw_f cy Hl); intros [? [? [? Haw]]].
    assert (L (f (DR (x,cy)))) as Hlcy.
      destruct Hxy as [[? ?] ?]; auto.
    generalize (Hrf x cy Hlcy); intros [? [? [? [? [[? Har] ?]]]]].  
    rewrite Haw in Har; inversion Har.
Qed.

Lemma grfe_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de : DEvent, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  A.inter = true ->
  L (f dy) ->
  rf_inter (p2X E p L) x y ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hdf Hmns Hdx Hdy Hgrfe Hly Hxy Hyx.
assert (dom p (f dy)) as Hudr.
  exists (f dx); auto. 
generalize (Hmns (f dy) Hly); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

case_eq dy; intros cy Hcy; rewrite Hcy in Hs; simpl in Hs; rewrite Hcy in Hdy; simpl in Hdy.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.

    assert (exists d, trans d /\ rel_incl d (pred p (f dy)) /\
    ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
    machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto; intros e1 e2 H12; auto.
      intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto.

  destruct cy as [wy ry]; rewrite <- Hdy in Hxy; unfold flush_resolved_read_step in Hs.
  assert (rf (p2X E p L) x ry) as Hrfxry.
    destruct Hxy; auto.

  assert (x = wy) as Heqxy.
    destruct Hxy as [Hrf Hintra]; generalize (path_rf_wf E p L); auto; intros [? [? Hrfu]]; auto.
    apply Hrfu with ry; auto.
      split. 
        destruct Hwfl as [? [[Hrd Hr_f] ?]]; apply Hr_f; rewrite <- Hcy; auto. 
        rewrite <- Hcy; auto.
  
  destruct (bin (ran (Q s1)) ry); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e wy))); [|inversion Hs];
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry))); [|inversion Hs].


  destruct (eqEv_dec wy (M s1 (loc ry))) as [Him|Hnim].
    rewrite Hcy in Hed;
    generalize (in_mem_implies_in_order E p L (f (DR (wy,ry))) n s0 s1 (f (DR (wy,ry))) wy (loc ry) Htp Hed Him);
    intros Hpwr.

  case_eq dx.
    intros [w r] Hdr.
    assert (reads E r) as Her.
      destruct Hwfl as [? [[? Hr_f] ?]]; generalize (Hr_f w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.

    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]];
    destruct Hxy as [Hrf ?]; generalize (Hrfc x ry Hrf); intros [? [? [? [[? Haw] ?]]]].
    destruct Her as [? [? [? Har]]].
    rewrite Hdr in Hdx; simpl in Hdx; subst. 
    rewrite Haw in Har; inversion Har.

    intros w Hdw.
    rewrite Hdw in Hdx.
    simpl in Hdx. 
    rewrite Hdw in Hyx; rewrite Hdx in Hyx; rewrite Heqxy in Hyx.
    rewrite Hcy in Hyx. 
      destruct_lin Hlp; apply Hac with (f (DW wy)). 
        apply Htrans with (f (DR (wy,ry))); auto.

  case_eq (bin (Intersection _ (ran (B s1)) (rrB ry)) wy); intro Hbemp; rewrite Hbemp in Hs; [inversion Hs|].
    assert (bin (Intersection _ (ran (B s1)) (rrB ry)) wy = true) as Hc.
      apply In_bin.

  case_eq dx.
    intros [w r] Hdr.
    assert (reads E r) as Her.
      destruct Hwfl as [? [[? Hr_f] ?]]; generalize (Hr_f w r); intros [? [? [l [? [[v Hrlv] ?]]]]].
        rewrite <- Hdr; auto; apply Hudrp; right; exists (f dy); auto.
        split; auto. exists l; exists v; auto.
    generalize (path_rf_wf E p L Hlp Hwfrfl); intros [? [Hrfc ?]];
    destruct Hxy as [Hrf ?]; generalize (Hrfc x ry Hrf); intros [? [? [? [[? Haw] ?]]]].
    destruct Her as [? [? [? Har]]].
    rewrite Hdr in Hdx; simpl in Hdx; subst. 
    rewrite Haw in Har; inversion Har.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l : Label, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      admit. 
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.

    generalize (in_order_implies_in_buff E p L (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed');
    intros Hrr; split. 
      exists dB. 
        rewrite <- Heqxy; rewrite <- Hdx; rewrite Hdw; auto.

        split; auto.
          apply ABasic.rf_implies_same_loc2 with E (p2X E p L); auto.
            split; [apply path_ws_wf|apply path_rf_wf]; auto.
              apply wfl_w; auto.
            split. 
              destruct Hwfl as [? [[? Hr_f] ?]]; apply Hr_f; auto;
                rewrite <- Hcy; auto. rewrite <- Hcy; auto.
          rewrite Hgrfe; auto.
          case_eq A.intra; intro; auto. 
            destruct Hxy; rewrite <- Heqxy; auto. 
            destruct Hxy; rewrite <- Heqxy; auto.             
    rewrite Hbemp in Hc; inversion Hc. 

  rewrite <- Hdy in Hxy.
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
  destruct Hwfl as [[? Hw_f] [[? Hr_f] ?]].
    rewrite Hcy in Hudr; destruct_lin Hlp.
    assert ((Union Label (dom p) (ran p)) (f (DW cy))) as Hi.
      left; auto.
    generalize (Hdr (f (DW cy)) Hi); intro Hl.
    generalize (Hw_f cy Hl); intros [? [? [? Haw]]].

    assert (L (f (DR (x,cy)))) as Hlxcy.
      destruct Hxy as [[? ?] ?]; auto.
    generalize (Hr_f x cy Hlxcy); intros [? [? [? [? [[? Har] ?]]]]]. 
    rewrite Haw in Har; inversion Har.
Qed.

Lemma ws_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  ws (p2X E p L) x y ->
  L (f dy) ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlfy Hyx.
assert (dom p (f dy)) as Hudr.
  exists (f dx); auto.
generalize (Hmns (f dy) Hlfy); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

case_eq dy; intros cy Hcy; rewrite Hcy in Hs; simpl in Hs; rewrite Hcy in Hdy; simpl in Hdy.

  destruct cy as [wy ry]; rewrite <- Hdy in Hxy.
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
      destruct Hwfl as [[? Hw_f] [[? Hr_f] ?]].
    rewrite Hcy in Hudr; destruct_lin Hlp.
    assert ((Union Label (dom p) (ran p)) (f (DR (wy,ry)))) as Hi'.
      left; auto.
    generalize (Hdr (f (DR (wy,ry))) Hi'); intro Hl'.
    generalize (Hr_f wy ry Hl'); intros [? [? [? [? [[? Har] ?]]]]]. 

    destruct Hxy as [Hwx [Hwry [? [? [lw [? [Hrlw Hdlw]]]]]]].
    rewrite Hrlw in Hdlw.    
    assert ((Union Label (dom p) (ran p)) (f (DW ry))) as Hi.
      right; auto; exists x2; auto.
    generalize (Hdr (f (DW ry)) Hi); intro Hl.
    generalize (Hw_f ry Hl); intros [? [? [? Haw]]].
    rewrite Haw in Har; inversion Har.

  rewrite <- Hdy in Hxy; unfold flush_write_step in Hs.
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy))); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun r' : Event => loc r' = loc cy /\ po_iico E r' cy))); [|inversion Hs];
  destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc cy))) cy) as [Heq|]; [|inversion Hs].
  unfold pick_last in Heq.
  generalize (EPick.pick_in (fun w : Event =>
         (~
         (exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc cy) dB w0) /\ w = dB) \/
         ((exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc cy) dB w0) /\
         rrestrict (B s1) (fun e : Event => loc e = loc cy) dB w /\
         ~
         (exists w' : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc cy) w' w))));
  rewrite Heq; intros [HdB | Hi].

    destruct HdB as [? HdB]; rewrite HdB in Hxy.
    generalize (wfl_w E L Hwfl); intro Hlw.
    generalize (ABasic.ran_ws_in_events (p2X E p L) x dB Hwf (path_ws_wf E p L Hlw Hlp) Hxy);
    intro HedB; apply (dB_not_in_E E dB HedB); auto.

  destruct Hi as [Hi [? Hn]].
  assert (exists w', rrestrict (B s1) (fun e => loc e = loc cy) w' cy) as Hy.
    exists x; split; auto.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
  assert (exists d, rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; auto.
      intros e1 e2 H12; auto.

  case_eq dx.
    intros [w r] Hdr.
    assert (reads E r) as Her.
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
      assert (L (f (DR (w,r)))) as Hlfwr.
        rewrite <- Hdr; apply Hudrp; right; exists (f dy); subst; auto.
      destruct Hwfl as [? [[? Hr_f] ?]]; generalize (Hr_f w r Hlfwr); intros [? [? [l [? [[v Hisr] ?]]]]]; 
        split; auto; exists l; exists v; auto.
    assert (forall w : Event, writes E w -> L (f (DW w))) as Hlw.
      apply wfl_w; auto.  
    generalize (path_ws_wf E p L Hlw Hlp); intros [? Hwtl]; generalize (Hwtl x cy Hxy);
    intros [? [[? [? Haw]] ?]].
    destruct Her as [? [? [? Har]]].
    rewrite Hdr in Hdx; simpl in Hdx; subst. 
    rewrite Haw in Har; inversion Har.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (loc w = loc cy) as Heql.
      assert (forall w : Event, writes E w -> L (f (DW w))) as Hlw.
        apply wfl_w; auto.
      generalize (path_ws_wf E p L Hlw Hlp); intros [? Hwtl]; generalize (Hwtl x cy Hxy).
      rewrite Hdw in Hdx; simpl in Hdx; subst.
      intros [? [[? [? Hax]] [? [? Hacy]]]]; unfold loc; rewrite Hax; rewrite Hacy; auto.
    assert (tc p (f dy) (f (DW w))) as Htcp.
      apply trc_step; auto.
    rewrite Hdy in Hxy; rewrite Hcy in Hyx; rewrite Hdy in Hyx; 
    rewrite <- Hdw in Hyx.
    destruct Hxy as [? [? [? [lx [ly [Hlx [Hly Hxy]]]]]]].
      rewrite Hdw in Hdx; simpl in Hdx. 
      rewrite Hdw in Hyx; rewrite Hdx in Hyx.
      rewrite Hlx in Hxy; rewrite Hly in Hxy.
      assert (tc p (f (DW x)) (f (DW x))) as Hc.
        apply trc_ind with (f (DW y)); apply trc_step; auto.
      generalize (Hacp (f (DW x)) Hc); intro Ht; inversion Ht.

      split; auto.
        apply ABasic.ws_implies_same_loc with E (p2X E p L); auto.
          split; [apply path_ws_wf | apply path_rf_wf]; auto.
            apply wfl_w; auto.
          unfold Ensembles.In; simpl; auto.
  contradiction.
Qed.

Lemma ws_implies_p' :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  ws (p2X E p L) x y ->
  L (f dx) -> L (f dy) ->
  p (f dx) (f dy).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlfx Hlfy. 

assert (f dx <> f dy) as Hdiff.
  intro Heq; inversion Heq as [Heqd]; rewrite <- Heqd in Hdy;
  rewrite Hdy in Hdx; rewrite Hdx in Hxy.
  destruct Hwfl as [? [? [Hwl ?]]]; generalize (path_ws_wf E p L Hwl Hlp);
  intros [Hlin ?].
    destruct_lin (Hlin (loc x)).  
    assert (rrestrict (ws (p2X E p L)) (writes_to_same_loc_l (events E) (loc x)) x x) as Hc.
      split; auto. 
        destruct Hxy as [[Hex [l [v Hax]]] ?]; split; split; auto; unfold loc; rewrite Hax; exists v; auto.
      apply (Hac x Hc).


generalize Hlp; intro Hlp'; destruct_lin Hlp'.
generalize (Htot (f dx) (f dy) Hdiff Hlfx Hlfy); intros [| Hyx]; auto.
generalize (ws_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlfy);
intro; contradiction.
Qed.

Lemma fr_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  fr E (p2X E p L) x y ->
  L (f dx) -> L (f dy) ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlx Hly Hyx.

generalize (Hmns (f dx) Hlx); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].
case_eq dx; intros cx Hcx; rewrite Hcx in Hs; simpl in Hs; rewrite Hcx in Hdx; simpl in Hdx.
case_eq dy; intros cy Hcy.
  destruct cy as [wy ry].
  destruct Hxy as [Hex [Hey [w [Hrf Hws]]]].
  assert (forall w : Event, writes E w -> L (f (DW w))) as Hlw.
    apply wfl_w; auto.
  generalize (ABasic.ran_ws_is_write E (p2X E p L) w y (path_ws_wf E p L Hlw Hlp) Hws); intros [v [l Hwy]].
  assert (ry = y) as Heqy.
    rewrite <- Hdy; rewrite Hcy; auto.
  assert (L(f (DR (wy,y)))) as Hlcy.
    rewrite <- Heqy; rewrite <- Hcy; auto.
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
      destruct Hwfl as [? [[? Hr_f] ?]];
  generalize (Hr_f wy y Hlcy); intros [? [? [? [? [[? Har] ?]]]]];
  rewrite Har in Hwy; inversion Hwy.
  
  rewrite Hcy in Hdy; simpl in Hdy; rewrite Hdy in Hcy.

        assert (trans p) as Htp.
          destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
        assert (acyclic p) as Hacp.
          destruct_lin Hlp; intros e He.
            apply (Hac e); rewrite trans_tc in He; auto.
       assert (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
          destruct_lin Hlp; auto. 

  destruct cx as [w r]; rewrite <- Hdx in Hxy; unfold flush_resolved_read_step in Hs.
    destruct (bin (ran (Q s1)) r); [|inversion Hs].
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); [|inversion Hs].
    destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].

      assert (ws (p2X E p L) w y) as Hwswy.
        destruct Hxy as [? [? [wx [Hrf Hws]]]].
        generalize (path_rf_wf E p L); intros [? [? Huni]]; auto.
        assert (L (f (DR (w,r)))) as Hlcy.
          rewrite <- Hcx; auto.
        assert (rf (p2X E p L) w r) as Hrfwr. 
          split; auto.
          destruct Hwfl as [? [[? Hr_f] ?]]; apply Hr_f; auto.        

        generalize (Huni r w wx Hrfwr Hrf); intro Heq; rewrite Heq; auto.
      assert (p (f (DW w)) (f dy)) as Hpwy.
        assert (f (DW w) <> f dy) as Hdwy.
          rewrite Hcy; intro Heq; inversion Heq as [Heqw].
      assert (forall w : Event, writes E w -> L (f (DW w))) as Hlw.
        apply wfl_w; auto.
          rewrite Heqw in Hwswy; generalize (path_ws_wf E p L Hlw Hlp); intros [Hlws ?].
          generalize (Hlws (loc r)); intros [? [? [Hac ?]]].
            assert (Ensembles.In Event (writes_to_same_loc_l (events E) (loc r)) y) as Hiy.
              split; auto.
                apply ABasic.dom_ws_in_events with (p2X E p L) y; auto; apply path_ws_wf; auto.
                generalize (ABasic.dom_ws_is_write E (p2X E p L) y y (path_ws_wf E p L Hlw Hlp) Hwswy); intros [l [v Haw]].
                apply ABasic.write_implies_write_to with v; auto.
                  rewrite ABasic.fr_implies_same_loc with E (p2X E p L) r y; auto; unfold loc. 
                    rewrite Haw; auto.
                    split; [apply path_ws_wf | apply path_rf_wf]; auto.

            apply (Hac y); split; auto.

      assert (L (f (DW w))) as Hlw.
        destruct Hwswy as [? [? [? [l1 [l2 [Hl1 [Hl2 Hp12]]]]]]].
        destruct_lin Hlp; apply Hdr; left; exists l2; rewrite <- Hl1; auto.

        generalize (Htotp (f (DW w)) (f dy) Hdwy Hlw Hly); intro Horwy; inversion Horwy as [|Hyw]; auto.
        assert (evt_of_devt (DW w) = w) as Hew.
          simpl; auto.
        assert (evt_of_devt dy = y) as Hey.
          rewrite Hcy; auto.
        generalize (ws_implies_p E p L (DW w) dy w y Hwf Hwfl Hwfrfl Hlp  Hmns Hew Hey Hwswy Hly) ; intro; contradiction.

    destruct (eqEv_dec w (M s1 (loc r))) as [Him | Hnim].

        generalize (in_mem_implies_no_interv_write E p L (f dx) n s0 s1 (f dx) w (loc r) Hacp Htp Hm Hwfm Him); intro Hn. 
        assert (exists w', loc w' = loc r /\ p (f (DW w)) (f (DW w')) /\ p (f (DW w')) (f dx)) as Hc.
          exists y; split; [|split]; auto.
            apply sym_eq; apply ABasic.fr_implies_same_loc with E (p2X E p L); auto.
              split; [apply path_ws_wf | apply path_rf_wf]; auto.
                apply wfl_w; auto.
            rewrite <- Hcy; auto. 
            rewrite <- Hcy; auto.
        contradiction.

    destruct (bin (Intersection _ (ran (B s1)) (rrB r))); [inversion Hs|].

    case_eq (bin (fun w : Event => (B s1 w (waft E  (B s1) r)) /\ loc r = loc w) w); intro Hi; [|rewrite Hi in Hs; inversion Hs].
      generalize (bin_In (fun w => (B s1 w (waft E  (B s1) r)) /\ loc r = loc w) w Hi); intros [Hor Hl].

       assert (events E w) as Hew.
         apply ABasic.dom_ws_in_events with (p2X E p L) y; auto.
           apply path_ws_wf; auto.
           apply wfl_w; auto.
       assert (f dx <> f (DW w)) as Hdiff.
         rewrite Hcx; intro Heq; inversion Heq.
       assert (L (f dx)) as Hlb.
         destruct_lin Hlp; apply Hdr; right; exists (f dy); auto.
       
       assert (exists r, rel_incl r (pred p (f dx)) /\ 
         
          ~ (exists l' : Label, tc p (last r) l' /\ tc p l' (f dx)) /\
         machine E r L (f dx) n (Some s0) (Some s1)) as Hed.
         exists (pred p (f dx)); split; [|split]; auto. intros e1 e2 H12; auto.
         apply no_interv_last; auto; intros e1 e2 H12; auto.
       assert (udr (B s1) w) as Hran.
         left; exists (waft E (B s1) r); auto.
      assert (forall w : Event, writes E w -> L (f (DW w))) as Hlw.
        apply wfl_w; auto.
    assert (forall de : DEvent,  L (f de) -> L (d de)) as Hlfd.
      destruct Hwfl as [? [? [? ?]]]; auto.
       assert (writes E w) as Hww.
         split. 
           apply ABasic.dom_ws_in_events with (p2X E p L) y; auto.
             apply path_ws_wf; auto. 
           apply ABasic.dom_ws_is_write with E (p2X E p L) y; auto.
             apply path_ws_wf; auto.
       assert ( ~ (exists l' : Label, tc p (f dx) l' /\ tc p l' (f dx)) \/ f dx = f dx) as Hfdx.
         auto.

     destruct_lin Hlp;
     generalize (in_buff_implies_f_after E L (f dx) p n s0 s1 (f dx) w Hdr Hlw Hlfd Hww Hdiff Hlb Hacp Htp Htotp Hmns Hdiff Hlb Hfdx Hed Hran); intro Hpxw.
       assert (tc p (f dx) (f dy)) as Hpxy.
         apply trc_ind with (f (DW w)); auto; apply trc_step; auto.
       apply (Hacp (f dx)); apply trc_ind with (f dy); auto; apply trc_step; auto.

  rewrite <- Hdx in Hxy; 
  destruct Hxy as [? [? [wr [Hrf Hws]]]].
      generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hwfrf.
  destruct Hwfl as [[? Hw_f] [[? Hr_f] ?]].
    rewrite Hcx in Hlx; generalize (Hw_f cx Hlx); intros [? [? [? Haw]]].

    assert (L (f (DR (wr,cx)))) as Hlfx.
      destruct Hrf as [? ?]; auto.
 
    generalize (Hr_f wr cx Hlfx); intros [? [? [? [? [[? Har] ?]]]]].
    rewrite Haw in Har; inversion Har.
Qed.

Lemma fr_implies_p' :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  fr E (p2X E p L) x y ->
  L (f dx) -> L (f dy) ->
  p (f dx) (f dy).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlx Hly.
generalize Hlp; intro Hlp'; destruct_lin Hlp'.
assert (f dx <> f dy) as Hdiff.
  intro Heq; inversion Heq as [Heqd]; rewrite Heqd in Hdx; rewrite Hdx in Hdy;
  rewrite Hdy in Hxy.
  generalize (path_rf_wf E p L Hlp Hwfrfl); intro Hrfwf.
  generalize (ABasic.dom_fr_is_read Hrfwf Hxy); intros [? [? Har]].
  destruct Hwfl as [? [? [Hwl ?]]].
  generalize (path_ws_wf E p L Hwl Hlp); intro Hwswf.
  generalize (ABasic.ran_fr_is_write Hwswf Hxy); intros [? [? Haw]].
  rewrite Har in Haw; inversion Haw.
generalize (Htot (f dx) (f dy) Hdiff Hlx Hly); intros [| Hyx]; auto.
generalize (fr_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlp  Hmns Hdx Hdy Hxy Hlx Hly);
intro; contradiction.
Qed.

Lemma ab_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  A.abc E (p2X E p L) x y ->
  L (f dy) ->
  ~p (f dy) (f dx).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hdf Hmns Hdx Hdy Hxy Hly Hyx.
assert (dom p (f dy)) as Hudr.
  exists (f dx); auto.
generalize (Hmns (f dy) Hly); intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]].

case_eq dy; intros cy Hcy; rewrite Hcy in Hs; simpl in Hs; rewrite Hcy in Hdy; simpl in Hdy.

  destruct cy as [wy ry]; rewrite <- Hdy in Hxy; unfold flush_resolved_read_step in Hs.

    destruct (bin (ran (Q s1)) ry); [|inversion Hs];
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e wy))); [|inversion Hs].
    case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry))); 
      [intros Hbemp | intros Hbnemp; rewrite Hbnemp in Hs; inversion Hs].
      generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry) Hbemp); intro Hn.
      assert (exists e, ran (rel_union (B s1) (Q s1)) e /\ rel_union (A.ppo E) (A.abc E (p2X E p L)) e ry) as Hy.
        exists x; split; [|right]; auto.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ 
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
          machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto. 
  assert (exists d, rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; auto.
      intros e1 e2 H12; auto.
  case_eq dx.
    intros [w r] Hdwr.
    assert (reads E r) as Her.
      assert (L (f (DR (w,r)))) as Hlwr.
        rewrite Hdwr in Hyx; destruct_lin Hlp; apply Hdr; right; exists (f dy); auto.
      destruct Hwfl as [? [[? Hr_f] ?]]; generalize (Hr_f w r Hlwr); intros [? [? [l [? [[v Har] ?]]]]]; 
      split; auto; exists l; exists v; auto.

    rewrite Hdwr in Hyx.
    assert (p (f dy) (f (DR (w, r))) \/ f dy = f (DR (w, r))) as Horyx.
      auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.
    assert (p (d (DR (w,r))) (f dy)) as Hpxy.
      admit. 
    assert (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    generalize (in_order_implies_in_queue E p L (f dy) n s0 s1 (f dy) w r  Hudrp Htp Hacp   Htotp Hwfrfl Hly Hdf Hpxy Hpxy Horyx Hni Hed);
    intros [e Hrr]; exists e; right; rewrite <- Hdx; rewrite Hdwr; auto.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      admit. 
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.

    generalize (in_order_implies_in_buff E p L (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed);
    intros Hrr; exists dB; left; rewrite <- Hdx; rewrite Hdw; auto.
      contradiction.

  rewrite <- Hdy in Hxy; unfold flush_write_step in Hs.
    case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy))); 
      [intros Hbemp | intros Hbnemp; rewrite Hbnemp in Hs; inversion Hs].
      generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy) Hbemp); intro Hn.
      assert (exists e, ran (rel_union (B s1) (Q s1)) e /\ rel_union (A.ppo E) (A.abc E (p2X E p L)) e cy) as Hy.
        exists x; split; [|right]; auto.

  assert (Included _ (Union _ (dom p) (ran p)) L) as Hudrp.
    destruct_lin Hlp; auto.
  assert (trans p) as Htp.
    destruct_lin Hlp; intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
  assert (acyclic p) as Hacp.
    destruct_lin Hlp; intros e He; rewrite trans_tc with Label p in He; auto. 
    apply (Hac e He).
  assert (exists d, trans d /\ rel_incl d (pred p (f (dy))) /\ 
          ~ (exists l' : Label, tc p (last d) l' /\ tc p l' (f dy)) /\
          machine E d L (f dy) n (Some s0) (Some s1)) as Hed.
    exists (pred p (f (dy))); split; [|split; [|split]]; auto.
      apply td_tpred; auto. intros e1 e2 H12; auto.
      apply no_interv_last; auto; intros e1 e2 H12; auto.
  assert (exists d, rel_incl d (pred p (f (dy))) /\ machine E d L (f dy) n (Some s0) (Some s1)) as Hed'.
    exists (pred p (f (dy))); split; auto.
      intros e1 e2 H12; auto.
  case_eq dx.
    intros [w r] Hdwr.
    assert (reads E r) as Her.
      assert (L (f (DR (w,r)))) as Hlwr.
        rewrite Hdwr in Hyx; destruct_lin Hlp; apply Hdr; right; exists (f dy); auto.
      destruct Hwfl as [? [[? Hr_f] ?]]; generalize (Hr_f w r Hlwr); intros [? [? [l [? [[v Har] ?]]]]]; 
      split; auto; exists l; exists v; auto.

    rewrite Hdwr in Hyx.
    assert (p (f dy) (f (DR (w, r))) \/ f dy = f (DR (w, r))) as Horyx.
      auto.
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.    
    assert (p (d (DR (w, r))) (f dy)) as Hpxy.
      admit. 
    assert (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    generalize (in_order_implies_in_queue E p L (f dy) n s0 s1 (f dy) w r  Hudrp Htp Hacp   Htotp Hwfrfl Hly Hdf Hpxy Hpxy Horyx Hni Hed);
    intros [e Hrr]; exists e; right; rewrite <- Hdx; rewrite Hdwr; auto.

    intros w Hdw.
    rewrite Hdw in Hyx.
    assert (forall l, L l -> events E (evt_of_label l)) as Hle.
      apply wfl_e; auto.
    assert (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) as Htotp.
      destruct_lin Hlp; auto.
    assert (p (d (DW w)) (f dy)) as Hwy.
      admit. 
    assert (~ (exists l' : Label, tc p (f dy) l' /\ tc p l' (f dy)) \/ f dy = f dy) as Hni.
      auto.
    generalize (in_order_implies_in_buff E p L (f dy) n s0 s1 w (f dy) Hudrp Htp Hacp Htotp Hle Hwy Hwy Hyx Hdf Hni Hed);
    intros Hrr; exists dB; left; rewrite <- Hdx; rewrite Hdw; auto. 
      contradiction.
Qed.   

Lemma ghb_implies_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall de, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  AWmm.ghb E (p2X E p L) x y ->
  L (f dx) -> L (f dy) ->
  p (f dx) (f dy).  
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hpop Hdf Hmns Hdx Hdy Hxy Hlx Hly.

assert (events E x) as Hex.
  apply ABasic.ghb_domain_in_events with (p2X E p L) y; auto.
    split; [apply path_ws_wf | apply path_rf_wf]; auto.
      apply wfl_w; auto.

assert (events E y) as Hey.
  apply ABasic.ghb_range_in_events with (p2X E p L) x; auto.
    split; [apply path_ws_wf | apply path_rf_wf]; auto.
      apply wfl_w; auto.

destruct (eqEv_dec x y) as [Heq | Hneq].
rewrite <- Heq in Hxy; generalize (ghb_irr E p L x Hwf Hlp  Hmns Hwfl Hwfrfl);
intro; contradiction.

assert (f dx <> f dy) as Hdiff.
  intro Heq; inversion Heq as [Hdeq].
  rewrite Hdeq in Hdx; rewrite <- Hdx in Hneq; rewrite <- Hdy in Hneq;
  apply Hneq; auto.
generalize Hlp; intro Hlin; destruct_lin Hlp; clear Hdr Htrans Hac;
generalize (Htot (f dx) (f dy) Hdiff Hlx Hly);
intro Hor; inversion Hor as [|Hyx]; clear Htot Hor; auto.

case_eq A.inter; case_eq A.intra; intros intra inter; 
unfold AWmm.ghb in Hxy; rewrite intra in Hxy; rewrite inter in Hxy.


  inversion Hxy as [Hrfe | Hr]; clear Hxy;
  [|inversion Hr as [Hrfi | Hre]; clear Hr;
  [| inversion Hre as [Hab | Hres]; clear Hre;
  [|inversion Hres as [Hwsfr | Hppo]; clear Hres;
  [inversion Hwsfr as [Hws | Hfr]; clear Hwsfr|]]]].

    generalize (grfe_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy inter Hly Hrfe);
    intro; contradiction.   
    generalize (grfi_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy intra Hly Hrfi);
    intro; contradiction.
    generalize (ab_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy Hab Hly);
    intro; contradiction.
    generalize (ws_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hws Hly);
    intro; contradiction.
    generalize (fr_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hfr Hlx Hly);
    intro; contradiction.
    generalize (ppo_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hpop Hdf Hmns Hdx Hdy Hly Hppo);
    intro; contradiction.

  inversion Hxy as [Hrfe | Hre]; clear Hxy;
  [| inversion Hre as [Hab | Hres]; clear Hre;
  [|inversion Hres as [Hwsfr | Hppo]; clear Hres;
  [inversion Hwsfr as [Hws | Hfr]; clear Hwsfr|]]].

    generalize (grfe_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy inter Hly Hrfe);
    intro; contradiction.  
    generalize (ab_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy Hab Hly);
    intro; contradiction.
    generalize (ws_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hws Hly);
    intro; contradiction.
    generalize (fr_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hfr Hlx Hly);
    intro; contradiction.
    generalize (ppo_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hpop Hdf Hmns Hdx Hdy Hly Hppo);
    intro; contradiction.

  inversion Hxy as [Hrfi | Hre]; clear Hxy;
  [| inversion Hre as [Hab | Hres]; clear Hre;
  [|inversion Hres as [Hwsfr | Hppo]; clear Hres;
  [inversion Hwsfr as [Hws | Hfr]; clear Hwsfr|]]].

    generalize (grfi_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy intra Hly Hrfi);
    intro; contradiction.   
    generalize (ab_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy Hab Hly);
    intro; contradiction.
    generalize (ws_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hws Hly);
    intro; contradiction.
    generalize (fr_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hfr Hlx Hly);
    intro; contradiction.
    generalize (ppo_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hpop Hdf Hmns Hdx Hdy Hly Hppo);
    intro; contradiction.

  inversion Hxy as [Hab | Hres]; clear Hxy;
  [|inversion Hres as [Hwsfr | Hppo]; clear Hres;
  [inversion Hwsfr as [Hws | Hfr]; clear Hwsfr|]].

    generalize (ab_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hdf Hmns Hdx Hdy Hab Hly);
    intro; contradiction.
    generalize (ws_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hws Hly);
    intro; contradiction.
    generalize (fr_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin  Hmns Hdx Hdy Hfr Hlx Hly);
    intro; contradiction.
    generalize (ppo_implies_p E p L dx dy x y Hwf Hwfl Hwfrfl Hlin Hpop Hdf Hmns Hdx Hdy Hly Hppo);
    intro; contradiction.
Qed.

Lemma wfl_e2 :
  forall E L e, 
  wf_labels E L -> wf_rf_labels E L ->
  events E e -> 
  (exists de, e = evt_of_devt de /\ L (f de)).
Proof.
intros E L e Hwfl Hwfrf He.

destruct Hwfl as [? [? [Hw ?]]].
destruct Hwfrf as [Hr ?].
assert (writes E e \/ reads E e) as Hor.
  case_eq (action e); intros d l v Hae; case_eq d; intro Hd.
    right; split; auto; exists l; exists v; subst; auto.
    left; split; auto; exists l; exists v; subst; auto.
destruct Hor as [Hwe | Hre].

  exists (DW e); split; auto; apply Hw; auto.
  generalize (Hr e Hre); intros [w [Hrfm Hl]];
    exists (DR (w,e)); split; auto; apply Hr; auto.
Qed.

Lemma tc_ghb_implies_tc_p :
  forall E p L dx dy x y,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L ->
  evt_of_devt dx = x ->
  evt_of_devt dy = y ->
  tc (AWmm.ghb E (p2X E p L)) x y ->
  L (f dx) -> L (f dy) ->
  tc p (f dx) (f dy).
Proof.
intros E p L dx dy x y Hwf Hwfl Hwfrfl Hlp Hpop Hdf Hmns Hdx Hdy Hxy Hlx Hly.
generalize dx dy Hdx Hdy Hlx Hly; clear Hdx Hdy Hlx Hly dx dy; induction Hxy. 

  intros dx dy Hdx Hdy Hlx Hly; apply trc_step; apply ghb_implies_p with E L x y; auto.

  intros dx dy Hdx Hdy Hlx Hly.
  assert (events E z) as Hez.
    apply dom_tc_ghb_in_evts with (p2X E p L) y; auto.
    apply path_ws_wf; auto. apply wfl_w; auto.
    apply path_rf_wf; auto.

  generalize (wfl_e2 E L z Hwfl Hwfrfl Hez);
  intros [dz [Hdz HLz]].
        
  apply trc_ind with (f dz);
    [apply IHHxy1 | apply IHHxy2]; auto.
Qed.

Lemma path_ac_ghb :
  forall E p L,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  linear_strict_order p L ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L ->
  acyclic (AWmm.ghb E (p2X E p L)).
Proof.
intros E p L Hwf Hwfl Hwfrfl Hlp Hpop Hdf Hmns x Hx.
  assert (events E x) as Hex.
    apply dom_tc_ghb_in_evts with (p2X E p L) x; auto.
      apply path_ws_wf; auto. apply wfl_w; auto.
      apply path_rf_wf; auto.    
  generalize (wfl_e2 E L x Hwfl Hwfrfl Hex); intros [dx [Hdx Hlx]].

  assert (evt_of_devt dx = x) as Hdx'.
    apply sym_eq; auto.
  generalize (tc_ghb_implies_tc_p E p L dx dx x x Hwf Hwfl Hwfrfl Hlp Hpop Hdf Hmns Hdx' Hdx' Hx Hlx Hlx).
  destruct_lin Hlp.
  rewrite trans_tc; auto.
    intro Hc; apply (Hac (f dx)); auto.
    intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
Qed.

Lemma b_p :
  forall E p  L s1 w w',
  well_formed_event_structure E ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  wf_labels E L ->
  wf_l_p E p L ->
  wf_buff E (B s1) ->
  B s1 w w' -> loc w = loc w' ->
  p (f (DW w)) (f (DW w')).
Proof.
intros E p L  s1 w w' Hwf Hlp Hmns Hwfl Hwflp Hwfb  Hww' Heql.

  assert (f (DW w) <> f (DW w')) as Hd.
    intro Heq; inversion Heq as [Heqe]; rewrite Heqe in Hww'.
    assert (tc (rrestrict (B s1) (fun e => loc e = loc w')) w' w') as Hc.
      apply trc_step; split; auto; split; unfold Ensembles.In; auto.
    destruct Hwfb as [? [Ht [Hac ?]]]; apply (Hac w' w' Hc).
assert (L (f (DW w))) as Hlfw.
  apply wfl_w with E; auto.
    destruct Hwfb as [Hw ?]; apply Hw; left; exists w'; auto.
assert (L (f (DW w'))) as Hlfw'.
  apply wfl_w with E; auto.
    destruct Hwfb as [Hw ?]; apply Hw; right; exists w; auto.    
destruct_lin Hlp; generalize (Htot (f (DW w)) (f (DW w')) Hd Hlfw Hlfw');
intro Hor; inversion Hor as [|Hp]; clear Hor; auto.

  assert (step E p L (Some s1) (f (DW w')) = None) as Hc.
     simpl; unfold flush_write_step.
    destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event =>
          rel_union (A.ppo E) (A.abc E (p2X E p L)) e w'))); auto.
    destruct (bemp (rrestrict (Q s1) (fun r' => loc r' = loc w' /\ po_iico E r' w'))); auto.
    destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w'))) w') as [Heq |]; auto.

   generalize (EPick.pick_in (fun w : Event =>
         ~
         (exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') dB w0) /\
         w = dB \/
         (exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') dB w0) /\
         rrestrict (B s1) (fun e : Event => loc e = loc w') dB w /\
         ~
         (exists w'0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') w'0 w)));

   intros Hinl; fold (pick_last (rrestrict (B s1) (fun e => (loc e) = (loc w')))) in Hinl;
    inversion Hinl as [[? Hd'] | [Hil' [? Hl']]]; clear Hinl. 
      rewrite Hd' in Heq.
  assert (events E w') as Hew'.
    assert (udr (B s1) w') as Hudrb.
      right; exists w; auto.
    destruct Hwfb as [Hwfb ?]; generalize (Hwfb w' Hudrb); intros [? ?]; auto.

        rewrite <- Heq in Hew'; generalize (dB_not_in_E E dB Hew'); intro Hc. 
        assert False as Ht.
          apply Hc; auto.
        inversion Ht.
      rewrite Heq in Hl'.
      assert (exists e' : Event, rrestrict (B s1) (fun e : Event => loc e = loc w') e' w') as Hc.
        exists w; split; auto. 
          unfold Ensembles.In; split; auto.
      contradiction.

  generalize (mns_implies_not_none E p L s1 (f (DW w')) Hlfw' Hwflp Hmns Hc); intro Ht; inversion Ht.
Qed.

Lemma ws_b :
  forall E p L s1 w w',
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  wf_l_p E p L ->
  wf_buff E (B s1) ->
  linear_strict_order p L ->
  machine_not_stuck E p L ->
  ran (B s1) w -> ran (B s1) w' ->
  ws (p2X E p L) w w' -> 
  B s1 w w'.
Proof.
intros E p L s1 w w' Hwf Hwfl Hwfrfl Hwflp Hwfb Hlp  Hmns Hbw Hbw' Hws;
  generalize Hwfb; intros [? [Ht Hacb]]. 
assert (forall w, writes E w -> L (f (DW w))) as Hlw.
  apply wfl_w; auto.
assert (loc w = loc w') as Hl.
  apply ABasic.ws_implies_same_loc with E (p2X E p L); auto.
    split; [apply path_ws_wf | apply path_rf_wf]; auto.
assert (w <> w') as Hd.
  generalize (path_ws_wf E p L Hlw Hlp); intros [Hlin ?].
  generalize (Hlin (loc w)); intros [? [? [Hac ?]]].
  generalize (path_ws_wf E p L Hlw Hlp); intro Hwswf.
  intro Heq; apply Hac with w; split; auto.
    rewrite <- Heq in Hws; auto.
      split; split. 
        apply ABasic.dom_ws_in_events with (p2X E p L) w'; auto.
        generalize (ABasic.dom_ws_is_write E (p2X E p L) w w' Hwswf Hws); intros [? [v Haw]];
          exists v; unfold loc; rewrite Haw; auto.
        apply ABasic.dom_ws_in_events with (p2X E p L) w'; auto.
        generalize (ABasic.dom_ws_is_write E (p2X E p L) w w' Hwswf Hws); intros [? [v Haw]];
          exists v; unfold loc; rewrite Haw; auto.

generalize (Ht w w' Hd Hbw Hbw' Hl); intro Hor; inversion Hor as [? | Hw'w]; auto.
assert (p (f (DW w)) (f (DW w'))) as Hp.
  destruct Hws as [Hw1 [Hw2 [? [l1 [l2 [H1 [H2 Hp]]]]]]]; subst; auto.

  assert (loc w' = loc w) as Hl'.
    apply sym_eq; auto.
  generalize (b_p E p L s1 w' w Hwf Hlp Hmns Hwfl Hwflp Hwfb Hw'w Hl'); intro Hc.
  assert (p (f (DW w)) (f (DW w')) /\ p (f (DW w')) (f (DW w))) as Hand.
    split; auto.
  destruct_lin Hlp; generalize (Hac (f (DW w)) (Htrans (f (DW w)) (f (DW w')) (f (DW w)) Hand));
  intro Hf; inversion Hf.
Qed.

Definition loc_of_label lb :=
  match lb with
  | d de | f de => loc (evt_of_devt de) end.

Lemma flush_read_same_mem :
  forall E p L s1 s2 w l wr r,
  flush_resolved_read_step E p L s1 wr r = Some s2 ->
  (M s1) l = w ->
  (M s2) l = w.
Proof.
intros E p L s1 s2 w l w' r' Hs Hms0.

  unfold flush_resolved_read_step in Hs;
  destruct (bin (ran (Q s1)) r'); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun e => (dp.dp E) e w'))); [|inversion Hs];
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r'))); [|inversion Hs];
  destruct (eqEv_dec w' (M s1 (loc r'))) as [Him | Hnim].
    destruct (bemp (rrestrict (B s1) (fun w => po_iico E w r' /\ loc r' = loc w))); [|inversion Hs].
      inversion Hs as [Heq]; auto. 
  destruct (bin (Intersection _ (ran (B s1)) (rrB r'))); [inversion Hs|].
  destruct (bin (fun w : Event => (B s1 w (waft E (B s1) r')) /\ loc r' = loc w) w'); [|inversion Hs].
    destruct (bin (fun w : Event => (w = wbef E (B s1) r' \/ B s1 (wbef E (B s1) r') w) /\ loc r' = loc w) w'); [|inversion Hs].
      inversion Hs as [Heq]; auto.
Qed.

Lemma not_flush_write_step_implies_same_mem :
  forall E p L  s1 s2 lb w l,
  step E p L (Some s1) lb = Some s2 ->
  ~(exists w', lb = f (DW w') /\ loc w' = l) -> 
  (M s1) l = w ->
  (M s2) l = w.
Proof.
intros E p L  s1 s2 lb w l Hs Hd Hms0.
case_eq lb; intros de Hlb; rewrite Hlb in Hs;
case_eq de; intros c Hde; rewrite Hde in Hs; simpl in Hs.



  
  destruct c as [w' r']; unfold buff_read_step in Hs;
  inversion Hs; auto.

  
  unfold buff_write_step in Hs;
  inversion Hs; auto.



  
  destruct c as [w' r'];
  apply flush_read_same_mem with E p L s1 w' r'; auto.

  
  unfold flush_write_step in Hs;
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e c))); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun r' : Event => loc r' = loc c /\ po_iico E r' c))); [|inversion Hs];
  destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc c))) c); [|inversion Hs];
  inversion Hs as [Heqs]; auto.
    simpl; unfold upd_mem.
      rewrite Hlb in Hd; rewrite Hde in Hd; simpl in Hd.
      destruct (eqLoc_dec l (loc c)) as [Heq | Hneq]; auto.
        rewrite Heq in Hd; assert False as Ht.
          apply Hd; auto.
            exists c; split; [|split]; auto.
        inversion Ht.
Qed.

Lemma wbef_after :
  forall E B w r,
  well_formed_event_structure E ->
  po_iico E w r -> loc w = loc r ->
  po_iico E w (wbef E B r).
Proof.
intros E B w r Hwf Hwr Heql.
generalize (EPick.pick_in (fun w : Event =>
   ran B w /\
   loc r = loc w /\
   po_iico E w r /\ ~ (exists w' : Event, loc r = loc w' /\ po_iico E w w' /\ po_iico E w' r)));
fold (wbef E B r); intros [Hr [Hl [Hpo Hn]]].

assert (events E w) as Hew.
  apply ABasic.po_iico_domain_in_events with r; auto.
assert (events E (wbef E B r)) as Hewbef.
  apply ABasic.po_iico_domain_in_events with r; auto.  
assert (proc_of w = proc_of (wbef E B r)) as Heqp.
  rewrite ABasic.po_implies_same_proc with E w r; auto.
  rewrite ABasic.po_implies_same_proc with E (wbef E B r) r; auto.  
  apply ABasic.po_iico_range_in_events with w; auto.
  apply ABasic.po_iico_range_in_events with w; auto.

generalize (ABasic.same_proc_implies_po w (wbef E B r) Hwf Heqp Hew Hewbef);
intro Hor; inversion Hor as [|Hwbefw]; auto; clear Hor.

assert (exists w' : Event, loc r = loc w' /\ po_iico E (wbef E B r) w' /\ po_iico E w' r) as Hc.
  exists w; split; [|split]; auto.
contradiction.
Qed.

Hypothesis wsi_ppo :
  forall E X x y,
  ws X x y -> po_iico E x y -> A.ppo E x y.

Lemma po_ws :
 forall (E : Event_struct) (p : Rln Label) (L : Ensemble Label)
   (w1 w2 : Event),
 well_formed_event_structure E ->
 Included Label (Union Label (dom p) (ran p)) L ->
 trans p ->
 acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
 machine_not_stuck E p L ->
 (forall e1 e2 : Event, po_iico E e1 e2 -> p (d (DW e1)) (d (DW e2))) ->
 (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  wf_labels E L ->
  writes E w1 -> writes E w2 ->
  pio_llh E w1 w2 -> 
  ws (p2X E p L) w1 w2.
Proof.
intros E p L w1 w2 Hwf Hudr Htp Hacp Htotp Hmns Hpop Hfifo Hwfl Hw1 Hw2 Hpio.

assert (write_serialization_well_formed (events E) (ws (p2X E p L))) as Hwswf.
  apply path_ws_wf; auto.
    destruct Hwfl as [? [? [? ?]]]; auto.
    split; [|split; [|split]]; auto.
      intros x1 x2 x3 [H12 H23]; apply Htp with x2; auto.
      intros x Hx; apply Hacp with x; apply trc_step; auto.

destruct (eqEv_dec w1 w2) as [Heq | Hneq].

  assert False as Ht.
    rewrite Heq in Hpio; destruct Hpio as [? [Hpo ?]].
    apply (ABasic.po_ac Hwf Hpo).
  inversion Ht.

assert (ws (p2X E p L) w1 w2 \/ ws (p2X E p L) w2 w1) as Hor.
  destruct Hwswf as [Hws_tot ?].
  generalize (Hws_tot (loc w1)); intro Hlin; destruct_lin Hlin.
  assert ((writes_to_same_loc_l (events E) (loc w1)) w1) as Hww1.
    destruct Hw1 as [? [l [v Haw]]]; 
    split; auto; exists v; unfold loc; rewrite Haw; auto.
  assert ((writes_to_same_loc_l (events E) (loc w1)) w2) as Hww2.
    destruct Hpio as [Hl ?]; rewrite Hl.
    destruct Hw2 as [? [l [v Haw]]]; 
    split; auto; exists v; unfold loc; rewrite Haw; auto.    
   
  generalize (Htot w1 w2 Hneq Hww1 Hww2); intros [[? ?]|[? ?]]; auto.

inversion Hor as [H12 | H21]; auto.

  generalize (path_uniws E p L w1 w2 Hwf Hudr Htp Hacp Hmns Hpop Hfifo Hpio);
  intro; contradiction.
Qed.

Lemma rb_implies_step_none :
  forall E p L s e1 e2,
 well_formed_event_structure E ->
 Included Label (Union Label (dom p) (ran p)) L ->
 trans p ->
 acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
 machine_not_stuck E p L ->
 (forall e1 e2 : Event, po_iico E e1 e2 -> p (d (DW e1)) (d (DW e2))) ->
 (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  wf_labels E L ->
  writes E e1 ->
  writes E e2 ->
  pio_llh E e1 e2 ->
  (ran (B s)) e1 ->
  machine_not_stuck E p L ->  
  step E p L (Some s) (f (DW e2)) = None.
Proof.
intros E p L s e1 e2 Hwf Hudr Htp Hacp Htotp Hmns Hpop Hfifo Hwfl Hw1 Hw2 [? [? ?]] Hb1.
generalize (excluded_middle (step E p L (Some s) (f (DW e2)) = None)); intro Hor;
inversion Hor as [|Hs]; auto.
simpl in Hs; unfold flush_write_step in Hs.
case_eq (bemp (rrestrict (rel_union (B s) (Q s)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2))); 
intro Hib; [|rewrite Hib in Hs]. 

generalize (bemp_axt (rel_union (B s) (Q s)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2) Hib); intro Hn.
assert (exists e, ran (rel_union (B s) (Q s)) e /\ rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2) as Hc.
  exists e1; split; auto.
    destruct Hb1 as [e Hd]; exists e; left; auto.
    left; apply wsi_ppo with (p2X E p L); auto. 
      apply po_ws; auto.
        split; [|split]; auto.
contradiction.

assert False as Ht.
  apply Hs; auto.
inversion Ht.
Qed.

Lemma rb_implies_contrad :
  forall E p L s e1 e2,
  well_formed_event_structure E ->
  wf_labels E L ->
  wf_rf_labels E L ->
  wf_l_p E p L ->
 Included Label (Union Label (dom p) (ran p)) L ->
 trans p ->
 acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
 (forall e1 e2 : Event, po_iico E e1 e2 -> p (d (DW e1)) (d (DW e2))) ->
 (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  writes E e1 ->
  writes E e2 ->
  pio_llh E e1 e2 ->
  (ran (B s)) e1 ->
  ~(machine_not_stuck E p L).
Proof.
intros E p L s e1 e2 Hwf Hwfl Hwfrf Hwflp Hudr Htp Hacp Htotp Hpop Hfifo Hw1 Hw2 H12 Hib Hmns.
generalize (rb_implies_step_none E p L s e1 e2 Hwf Hudr Htp Hacp Htotp Hmns Hpop Hfifo Hwfl Hw1 Hw2 H12 Hib Hmns); intro Hn.
assert (L (f (DW e2))) as Hle2.
  apply wfl_w with E; auto; simpl.
apply (mns_implies_not_none E p L s (f (DW e2)) Hle2 Hwflp Hmns Hn).
Qed.

Lemma pred_incl3 :
  forall (r:Rln Label) l1 l2,
  trans r ->
  r l1 l2 ->
  rel_incl (pred r l1) (pred r l2).
Proof.
intros r l1 l2 Htr H12 x y [? Hor]; split; auto.

   apply Htr with l1; auto.
  
Qed.

Lemma path_unifrrf : 
  forall E p L e1 e2,
  well_formed_event_structure E ->
  wf_l_p E p L ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p -> wf_labels E L -> wf_rf_labels E L ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  machine_not_stuck E p L ->  
  pio_llh E e1 e2 -> ~ (rel_seq (fr E (p2X E p L)) (rf (p2X E p L))) e2 e1.
Proof.
intros E p L e1 e2 Hwf Hwflp Hudr Htp Hacp Hwfl Hwfrfl Htotp Hmns [? [? Hnrr]] [w [Hfr Hrf]].
assert (rfmaps_well_formed E (events E) (rf (p2X E p L))) as Hrfwf.
    apply path_rf_wf; auto.
      split; [|split; [|split]]; auto.
        intros x1 x2 x3 [H12 H23]; apply Htp with x2; auto.
        intros x Hx; apply Hacp with x; apply trc_step; auto.

apply Hnrr; split; split.
  apply ABasic.ran_rf_in_events with (p2X E p L) w; auto.
  apply ABasic.ran_rf_is_read with E (p2X E p L) w; auto.
    destruct Hrfwf as [? [? ?]]; auto.

  apply ABasic.dom_fr_in_events with (p2X E p L) w; auto.
  apply ABasic.dom_fr_is_read with E (p2X E p L) w; auto.
Qed.

Lemma path_unifr : 
  forall E p L e1 e2,
  well_formed_event_structure E ->
  wf_l_p E p L ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p -> wf_labels E L -> wf_rf_labels E L ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall e1 e2, po_iico E e1 e2 -> p (d (DW e1)) (d (DW e2))) ->
  (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  (forall de : DEvent, L (f de) -> p (d de) (f de)) ->
  machine_not_stuck E p L ->  
  pio_llh E e1 e2 -> ~ fr E (p2X E p L) e2 e1.
Proof.
intros E p L e1 e2 Hwf Hwflp Hudrp Htp Hacp Hwflb Hwfrfl Htotp Hpop Hfifo Hdf Hmns H12 [He2 [He1 [e [Hrf Hws]]]].
generalize Hrf; intros Hrfee2.

assert (L (f (DR (e,e2)))) as Hlbe2.
  destruct Hrfee2 as [? ?]; auto.
generalize (Hmns (f (DR (e,e2))) Hlbe2);
intros [n [s0 [s1 [s2 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]]]; simpl in Hs; 
unfold flush_resolved_read_step in Hs.

destruct (bin (ran (Q s1)) e2); [|inversion Hs];
destruct (bemp (rrestrict (Q s1) (fun e0 => (dp.dp E) e0 e))); [|inversion Hs];
destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e e2))); [|inversion Hs].

destruct (eqEv_dec e (M s1 (loc e2))) as [Heq | Hneq].

case_eq (bemp (rrestrict (B s1) (fun w => po_iico E w e2 /\ loc e2 = loc w))); [intros Hemp | intros Hnemp; rewrite Hnemp in Hs; inversion Hs].
  rewrite Hemp in Hs; generalize (bemp_axt (B s1) (fun w => po_iico E w e2 /\ loc e2 = loc w) Hemp); intro Hn.
      assert (exists e, ran (B s1) e /\ po_iico E e e2 /\ loc e2 = loc e) as Hc.
        destruct H12 as [? [Hpo12 ?]]; exists e1; split; [|split]; auto.
          exists dB; apply in_order_implies_in_buff with E p L (f (DR (e,e2))) n s0 (f (DR (e,e2))); auto.
            apply wfl_e; auto.
            admit. admit. 
            assert (f (DR (e,e2)) <> f (DW e1)) as Hdiff.
              intros Heql; inversion Heql.
            assert (L (f (DR (e,e2)))) as Hlb2.
              auto.
            assert (L (f (DW e1))) as Hlb1.
              apply wfl_w with E; auto.
              split; auto.
                assert (linear_strict_order p L) as Hlp.
                  split; [|split; [|split]]; auto.
                    intros x1 x2 x3 [H12 H23]; apply Htp with x2; auto.
                    intros x Hx; apply Hacp with x; apply trc_step; auto.
                generalize (ABasic.ran_ws_is_write E (p2X E p L) e e1 (path_ws_wf E p L (wfl_w E L Hwflb) Hlp) Hws); auto;
                intros [v [l Hw]]; exists l; exists v; auto. 
            generalize (Htotp (f (DR (e,e2))) (f (DW e1)) Hdiff Hlb2 Hlb1); intros [? | Hc]; auto.
            generalize (in_mem_implies_no_interv_write E p L (f (DR (e,e2))) n s0 s1 (f (DR (e,e2))) e (loc e2) Hacp Htp Hm Hwfm Heq); intro Hniw.
            assert (exists w', loc w' = loc e2 /\ p (f (DW e)) (f (DW w')) /\ p (f (DW w')) (f (DR (e,e2)))) as Hiw. 
              exists e1; split; [|split]; auto.
                apply ws_implies_p' with E L e e1; auto. 
                  split; [|split; [|split]]; auto.
                    intros x1 x2 x3 [H12 H23]; apply Htp with x2; auto.
                    intros x Hx; apply (Hacp x); apply trc_step; auto.
                  
                apply wfl_w with E; auto.
                  generalize (Hwfm (loc e2)); intros [Hmw ?]; 
                  rewrite Heq; apply Hmw; auto.
            contradiction.

            exists (pred p (f (DR (e,e2)))); split; [|split; [|split]]; auto.
              apply td_tpred; auto.
              intros x y Hxy; auto.
              apply no_interv_last; auto; intros x y Hxy; auto.
      contradiction.

destruct (bin (Intersection _ (ran (B s1)) (rrB e2))); [inversion Hs|].
destruct (bin (fun w => (B s1 w (waft E  (B s1) e2)) /\ loc e2 = loc w) e); [|inversion Hs].
  case_eq (bin (fun w => (w = wbef E (B s1) e2 \/ B s1 (wbef E (B s1) e2) w) /\ loc e2 = loc w) e); intro Hbef; rewrite Hbef in Hs; [|inversion Hs].
  generalize (bin_In (fun w => (w = wbef E (B s1) e2 \/ B s1 (wbef E (B s1) e2) w) /\ loc e2 = loc w) e Hbef);
  intros [Hb Heql].

  assert (linear_strict_order p L) as Hlp.
    split; [|split; [|split]]; auto.
      intros x1 x2 x3 [Hp12 Hp23]; apply Htp with x2; auto.
      intros x Hx; apply Hacp with x; apply trc_step; auto.
  assert (ran (B s1) e) as Hre.
    inversion Hb as [Hwbef | Hrwbefe].

      generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc e2 = loc w /\ po_iico E w e2 /\
        ~ (exists w' : Event,  loc e2 = loc w' /\ po_iico E w w' /\ po_iico E w' e2)));
      fold (wbef E (B s1) e2); intros [Hr [Hl [Hpo Hn]]].
      rewrite Hwbef; auto.

    
    exists (wbef E (B s1) e2); auto.

  assert (ran (B s1) e1) as Hre1.
    exists dB; apply in_order_implies_in_buff with E p L (f (DR (e,e2))) n s0 (f (DW e)); auto.
      apply wfl_e; auto.
      admit.  admit. 
      destruct Hws as [? [? [? [? [? [? [? ?]]]]]]]; subst; auto.
      admit. 
      exists (pred p (f (DR (e,e2)))); split; [|split; [|split]]; auto.
        apply td_tpred; auto.
        apply pred_incl3; auto.
          rewrite <- trans_tc with Label p; auto. 
          apply in_buff_implies_f_after with E L (f (DR (e,e2))) n s0 s1; auto.
            apply wfl_w; auto. destruct Hwflb as [? [? [? ?]]]; auto.
            assert (udr (B s1) e) as Hudr.
              right; auto.
            destruct Hwfb as [Hwfb ?].
            generalize (Hwfb e Hudr); intros [? ?]; auto.
             intro Heq; inversion Heq.
            intro Heq; inversion Heq.
            exists (pred p (f (DR (e,e2)))); split; [|split]; auto. 
              intros x y Hxy; auto.
              apply no_interv_last; auto; intros x y Hxy; auto.
            right; auto.
            apply no_interv_last; auto; intros x y Hxy; auto.

  generalize (ws_b E p L s1 e e1 Hwf Hwflb Hwfrfl Hwflp Hwfb Hlp  Hmns Hre Hre1 Hws); intro Hbee1.
  
  destruct H12 as [Hl12 [Hpo12 ?]];
  generalize (wbef_after E (B s1) e1 e2 Hwf Hpo12 Hl12); intro Hpo1wb.

  assert (writes E e1) as Hw1.
    assert (udr (B s1) e1) as Hudr.
      right; auto.
    destruct Hwfb as [Hwfb ?]; apply (Hwfb e1 Hudr).
  assert (pio_llh E e1 (wbef E (B s1) e2)) as H1wb.
    split; [|split]; auto. 
      rewrite Hl12; 
      generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc e2 = loc w /\ po_iico E w e2 /\
        ~ (exists w' : Event, loc e2 = loc w' /\ po_iico E w w' /\ po_iico E w' e2)));
      fold (wbef E (B s1) e2); intros [Hr [Hl [Hpo Hn]]]; auto.

      intros [[? [? [? Har]]] ?].
      destruct Hw1 as [? [? [? Haw]]]; rewrite Haw in Har; inversion Har.
    assert (writes E (wbef E (B s1) e2)) as Hwb.
      apply Hwfb.
        inversion Hb as [Heq | Hbw].
          rewrite <- Heq; right; auto.
          left; exists e; auto.

  generalize (rb_implies_contrad E p L s1 e1 (wbef E (B s1) e2) Hwf Hwflb Hwfrfl Hwflp Hudrp Htp Hacp Htotp Hpop Hfifo Hw1 Hwb H1wb Hre1); intro; contradiction.
Qed.

Lemma path_uniproc :
  forall E p L,
  well_formed_event_structure E ->
  wf_l_p E p L ->
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L -> wf_labels E L -> wf_rf_labels E L ->
  (forall e1 e2, pio_llh E e1 e2 -> ~ ABasic.hb' E (p2X E p L) e2 e1).
Proof.
intros E p L Hwf Hwflp Hudrp Htp Hacp Htotp Hdf Hpod Hfifo Hmns Hwfl Hwfrfl e1 e2 H12 H21; 
  inversion H21 as [[[[Hrf | Hfr]|Hws]|Hwsrf] | Hfrrf]; clear H21.

  generalize Hrf; fold (~(rf (p2X E p L) e2 e1)); apply path_unirf; auto.

  generalize Hfr; fold (~(fr E (p2X E p L) e2 e1)); apply path_unifr; auto.

  generalize Hws; fold (~(ws (p2X E p L) e2 e1)); apply path_uniws; auto.

  generalize Hwsrf; fold (~(rel_seq (ws (p2X E p L)) (rf (p2X E p L)) e2 e1)); apply path_uniwsrf; auto.
    apply wfl_w; auto.

  generalize Hfrrf; fold (~(rel_seq (fr E (p2X E p L)) (rf (p2X E p L)) e2 e1)); apply path_unifrrf; auto.
Qed.

Theorem from_mach_to_ax :
  forall E p L,
  well_formed_event_structure E ->
  wf_labels E L -> wf_rf_labels E L ->
  wf_l_p E p L ->
  linear_strict_order p L ->
  (forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> p (d de1) (d de2)) ->
  
  (forall w1 w2, loc w1 = loc w2 -> p (d (DW w1)) (d (DW w2)) -> p (f (DW w1)) (f (DW w2))) ->
  (forall de, p (d de) (f de)) ->
  machine_not_stuck E p L ->
  AWmm.valid_execution E (p2X E p L).
Proof.
intros E p L Hwf Hwfl Hwfrfl Hwflp Hlp Hpod Hfifo Hdf Hmns.
assert (forall w, writes E w -> L (f (DW w))) as Hlw.
  apply wfl_w; auto.
generalize (path_ws_wf E p L Hlw Hlp); intro Hwswf.
generalize (path_rf_wf E p L Hlp  Hwfrfl); intro Hrfwf.
split; [|split; [|split; [|split]]]; auto.

    destruct_lin Hlp.
    assert (trans p) as Htp.
      intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.      
     apply path_uniproc; auto.
        apply irr_trans_ac; auto; intros [x Hx]; apply (Hac x Hx).  

  intros x Hx.
    assert (~(exists x, rf (p2X E p L) x x)) as Hrf_irr.
      intros [e He]; destruct Hrfwf as [? [Hrfm ?]];
      generalize (ABasic.ran_rf_is_read E (p2X E p L) e e Hrfm He); intros [? [? Hre]];
      generalize (ABasic.dom_rf_is_write E (p2X E p L) e e Hrfm He); intros [? [? Hwe]];
        rewrite Hre in Hwe; inversion Hwe; auto.   
    assert (~(exists x, dp.dp E x x)) as Hdp_irr.
      intros [e He]; generalize (dp.dp_valid E); intros [Hi ?]; 
        generalize (Hi e e He). apply ABasic.po_ac; auto.
    assert (~(exists x, rel_union (rf (p2X E p L)) (dp.dp E) x x)) as Hu_irr.
      intros [e He]; inversion He as [Hrf | Hdp].
        apply Hrf_irr; exists e; auto.
        apply Hdp_irr; exists e; auto.
    assert (trans (dp.dp E)) as Htdp.
      apply dp.dp_valid.
    assert (trans (rf (p2X E p L))) as Htrf.
      intros e1 e2 e3 H12 H23; destruct Hrfwf as [? [Hrfm ?]];
      generalize (ABasic.ran_rf_is_read E (p2X E p L) e1 e2 Hrfm H12); intros [? [? Hr2]];
      generalize (ABasic.dom_rf_is_write E (p2X E p L) e2 e3 Hrfm H23); intros [? [? Hw2]];
        rewrite Hr2 in Hw2; inversion Hw2; auto.      
    generalize (union_cycle_implies_seq_cycle2 Hrf_irr Hdp_irr Hu_irr Htdp Htrf Hx); 
    intros [y Hy]; clear Hrf_irr Hdp_irr Hu_irr Htdp Htrf x Hx; 
    generalize (tc_dec Hy); intros [z [Hyz [Hzy | Heqzy]]].
      destruct_lin Hlp.
      assert (trans p) as Htp.
        intros e1 e2 e3 H12 H23; apply Htrans with e2; auto.
      generalize Hzy; apply path_thin; auto.
        apply irr_trans_ac; auto; intros [x Hx]; apply (Hac x Hx).  
      rewrite Heqzy in Hyz.
        destruct_lin Hlp; generalize Hyz; apply rf_dp_irr; auto.
          intros x1 x2 x3 H12 H23; apply Htrans with x2; auto. 
          apply irr_trans_ac; auto. 
            intros x1 x2 x3 H12 H23; apply Htrans with x2; auto. 
            intros [x Hx]; apply (Hac x Hx).

  apply path_ac_ghb; auto.
Qed.

End Soundness.

Module Completeness (A:Archi) (dp:Dp). 

  Module S := Soundness A dp.
  Import S AMach AWmm ABasic.

Definition Delay := Event_struct -> Execution_witness -> Rln Event. 
Record wf_delay d := mkd {
ghb_ndelay : forall E X x y, AWmm.ghb E X x y -> ~d E X x y;
delay_antisym : forall E X x y, d E X x y -> ~d E X y x;
delay_tc : forall E X x y z, ~d E X x y -> ~d E X y z -> ~d E X x z;
delay_irr : forall E X x, ~(~d E X x x);
ndelay_dom_evt : forall E X x y, ~d E X x y -> events E x;
ndelay_ran_evt : forall E X x y, ~d E X x y -> events E y}.

Definition labels_of_es E X :=
  fun l => (exists w, writes E w /\ (l = d (DW w) \/ l = f (DW w))) \/
           (exists w, exists r, rf X w r /\ (l = d (DR (w,r)) \/ l = f (DR (w,r)))).

Inductive X2p (E:Event_struct) (X:Execution_witness) (delay:Delay) : Rln Label :=
  | X2p_df : forall de, events E (evt_of_devt de) -> 
               labels_of_es E X (f de) -> 
               X2p E X delay (d de) (f de)
  | X2p_ghb : forall de1 de2, ~delay E X (evt_of_devt de1) (evt_of_devt de2) ->
               labels_of_es E X (f de1) -> 
               labels_of_es E X (f de2) -> 
                X2p E X delay (f de1) (f de2)
  | X2p_po : forall de1 de2, po_iico E (evt_of_devt de1) (evt_of_devt de2) -> 
               labels_of_es E X (f de1) ->
               labels_of_es E X (f de2) ->  
               X2p E X delay (d de1) (d de2)
  | X2p_rf : forall de1 de2, rf X (evt_of_devt de1) (evt_of_devt de2) ->
               labels_of_es E X (f de1) -> 
               labels_of_es E X (f de2) -> 
               X2p E X delay (d de1) (f de2)
  | X2p_nghb : forall de1 de2, delay E X (evt_of_devt de1) (evt_of_devt de2) ->
               labels_of_es E X (f de1) -> 
               labels_of_es E X (f de2) -> 
                 X2p E X delay (f de2) (f de1)
  | X2p_ws : forall de1 de2, ws X (evt_of_devt de1) (evt_of_devt de2) ->
               labels_of_es E X (f de1) -> 
               labels_of_es E X (f de2) -> 
                X2p E X delay (f de1) (f de2)
  | X2p_fr : forall de1 de2, fr E X (evt_of_devt de1) (evt_of_devt de2) ->
               labels_of_es E X (f de1) -> 
               labels_of_es E X (f de2) -> 
                X2p E X delay (f de1) (f de2).

Lemma ran_q_contrad :
  forall Q e r,
  ~(ran Q e) ->
  ran (del_queue Q r) e ->
  False.
Proof.
intros Q e r Hnr [e' [Hq ?]]; apply Hnr; exists e'; auto.
Qed.

Lemma flush_read_q_step :
  forall E p (L:set Label) s0 s1 de w r,
  reads E (evt_of_devt de) ->
  ~ran (Q s0) (evt_of_devt de) -> 
  ran (Q s1) (evt_of_devt de) ->
  flush_resolved_read_step E p L s0 w r = Some s1 ->
  f (DR (w,r)) = d de.  
Proof.
intros E p L s0 s1 de w r  Hisr Hni0 Hi1 Hs.
    unfold flush_resolved_read_step in Hs.
      destruct (bin (ran (Q s0)) r); [|inversion Hs];
      destruct (bemp (rrestrict (Q s0) (fun e : Event => dp.dp E e w))); [|inversion Hs];
      destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].

      destruct (eqEv_dec w (M s0 (loc r))) as [Him | Hnim].
        destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r /\ loc r = loc w))); [|inversion Hs].
          inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1.
          generalize (ran_q_contrad (Q s0) (evt_of_devt de) r Hni0 Hi1); intro Ht; inversion Ht.

      destruct (bin (Intersection _ (ran (B s0)) (rrB r))); [inversion Hs|].
      destruct (bin (fun w : Event => (B s0 w (waft E(B s0) r)) /\ loc r = loc w) w); [|inversion Hs].
        destruct (bin (fun w : Event => (w = wbef E (B s0) r \/ B s0 (wbef E (B s0) r) w) /\ loc r = loc w) w); [|inversion Hs].
          inversion Hs as [Heq]; auto; rewrite <- Heq in Hi1; simpl in Hi1.
          generalize (ran_q_contrad (Q s0) (evt_of_devt de) r Hni0 Hi1); intro Ht; inversion Ht.
Qed. 

Lemma q_step :
  forall E p (L:set Label) s0 s1 lb de,
  (forall (w : Event), L (d (DW w)) -> writes E w) ->
  reads E (evt_of_devt de) ->
  wf_labels E L ->
  wf_rf_labels E L ->
  ~ran (Q s0) (evt_of_devt de) -> 
  ran (Q s1) (evt_of_devt de) ->
  L lb -> L (d de) ->
  step E p L (Some s0) lb = Some s1 ->
  lb = d de.
Proof.
intros E p L s0 s1 lb de Hwflw Hisr Hwfl Hwfr Hni0 Hi1 Hllb Hlde Hs.
case_eq lb; intros de' Hdlb.

  case_eq de'.
  intros [w r] Hde'. rewrite Hdlb in Hs; rewrite Hde' in Hs.
  
    unfold step in Hs; unfold buff_read_step in Hs.

    inversion Hs as [Heq]; rewrite <- Heq in Hi1; simpl in Hi1.
    destruct Hi1 as [e' Hi]; inversion Hi as [Hiq | Hdq].
      assert (ran (Q s0) (evt_of_devt de)) as Hc.
         exists e'; auto.
      contradiction. 

      destruct Hdq as [HdQ Hr]. 
        case_eq de.
          intros [w' r'] Hde.
            rewrite Hde in Hr; simpl in Hr; rewrite Hr; auto.
            assert (L (d (DR (w,r)))) as Hldw.
              rewrite Hde' in Hdlb; rewrite Hdlb in Hllb; auto.

            assert (L (d (DR (w',r')))) as Hldw'.
              rewrite Hde in Hlde; auto.

            rewrite Hr in Hldw'; destruct Hwfr as [? [Hwfr ?]];
            generalize (Hwfr r w w' Hldw Hldw'); intro Heqw;
            rewrite Heqw; auto.  

          intros w' Hde.
            rewrite Hde in Hlde;
            generalize (Hwflw w' Hlde); intros [? [? [? Haw]]];
            rewrite Hde in Hisr; simpl in Hisr; destruct Hisr as [? [? [? Har]]];
            rewrite Har in Haw; inversion Haw.

  intros e Hde'; rewrite Hde' in Hdlb; rewrite Hdlb in Hs;
  simpl in Hs; unfold buff_write_step in Hs.
  inversion Hs as [Heq].
    rewrite <- Heq in Hi1; simpl in Hi1; contradiction.

  case_eq de'.

  intros [w r] Hde'; rewrite Hdlb in Hs; rewrite Hde' in Hs; simpl in Hs.
  apply flush_read_q_step with E p L s0 s1; auto.
  
  intros w Hde'; rewrite Hdlb in Hs; rewrite Hde' in Hs; simpl in Hs.

    unfold flush_write_step in Hs.
      destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs];
      destruct (bemp (rrestrict (Q s0) (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs];
      destruct (eqEv_dec (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc w))) w); inversion Hs as [Heq].
      rewrite <- Heq in Hi1; simpl in Hi1; contradiction.
Qed.

Module OEL.
Parameter LE : Rln Label -> Rln Label.
Hypothesis OE : forall (s S:set Label) (r:Rln Label),
  Included _ s S ->
  partial_order r s -> 
  rel_incl r (LE r) /\
  linear_strict_order (LE r) S.
Hypothesis le_lso : forall (s:set Label) (r:Rln Label),
  linear_strict_order r s -> LE r = r.
End OEL.

Lemma oel_tot :
  forall p L,
  partial_order (tc p) L ->
  (forall e1 e2, e1 <> e2 -> 
   L e1 -> L e2 ->
   (OEL.LE (tc p)) e1 e2 \/ (OEL.LE (tc p)) e2 e1).
Proof.
intros p L Hpart.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; destruct_lin Hlin; auto.
Qed.

Lemma oel_trans :
  forall p L,
  partial_order (tc p) L ->
  trans (OEL.LE (tc p)).
Proof.
intros p L Hpart.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; destruct_lin Hlin; auto.
intros x y z Hxy Hyz; apply Htrans with y; auto.
Qed.

Lemma oel_part :
  forall p L,
  partial_order (tc p) L ->
  partial_order (tc (OEL.LE (tc p))) L.
Proof.
intros p L Hpart.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; rewrite trans_tc; [|apply oel_trans with L]; auto.
split; [|split]; destruct_lin Hlin; auto.
Qed.

Lemma oel_ac :
  forall p L,
  partial_order (tc p) L ->
  acyclic (OEL.LE (tc p)).
Proof.
intros p L Hpart.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; destruct_lin Hlin; auto.
intros x Hx; apply Hac with x.
  rewrite trans_tc in Hx; auto.
  apply oel_trans with L; auto.
Qed.

Lemma oel_incl :
  forall p L,
  partial_order (tc p) L ->
  rel_incl (tc p) (OEL.LE (tc p)).
Proof.
intros p L Hpart.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [Hi ?]; auto.
Qed.

Lemma oel_dom :
  forall p L x y,
  partial_order (tc p) L ->
  (OEL.LE (tc p)) x y ->
  L x.
Proof.
intros p L x y Hpart Hxy.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; destruct_lin Hlin; auto.
apply Hdr; left; exists y; auto.
Qed.

Lemma oel_ran :
  forall p L x y,
  partial_order (tc p) L ->
  (OEL.LE (tc p)) x y ->
  L y.
Proof.
intros p L x y Hpart Hxy.
assert (Included _ L L) as Htl.
  intros l Hl; auto.
generalize (OEL.OE L L (tc p) Htl Hpart);
intros [? Hlin]; destruct_lin Hlin; auto.
apply Hdr; right; exists x; auto.
Qed.

Lemma oel_udr :
  forall p L,
  partial_order (tc p) L ->
  Included _ (Union _ (dom (OEL.LE (tc p))) (ran (OEL.LE (tc p)))) L.
Proof.
intros p L Hpart ? [x [y Hd] | x [y Hr]].
  apply oel_dom with p y; auto.
  apply oel_ran with p y; auto.
Qed.

Lemma path_dec :
  forall E X delay l1 l2,
  X2p E X delay l1 l2 ->
  ( 
    (exists de, l1 = d de /\ l2 = f de) \/

    
    (exists de1, exists de2, 
     f de1 = l1 /\ f de2 = l2 /\
     ~delay E X (evt_of_devt de1) (evt_of_devt de2)) \/

    
    (exists de1, exists de2,
     d de1 = l1 /\ d de2 = l2 /\
     (po_iico E) (evt_of_devt de1) (evt_of_devt de2)) \/

    
    (exists de2, exists de1,
     f de2 = l2 /\ f de1 = l1 /\
     delay E X (evt_of_devt de2) (evt_of_devt de1)) \/
 
    (exists de1, exists de2,
     d de1 = l1 /\ f de2 = l2 /\
     rf X (evt_of_devt de1) (evt_of_devt de2)) \/

    
    (exists de1, exists de2, 
     f de1 = l1 /\ f de2 = l2 /\
     ws X (evt_of_devt de1) (evt_of_devt de2)) \/

    
    (exists de1, exists de2, 
     f de1 = l1 /\ f de2 = l2 /\
     fr E X (evt_of_devt de1) (evt_of_devt de2))
   ).
Proof. 
intros E X delay l1 l2 H12.

  inversion H12 as [e Hdf | e1 e2 Hnd | e1 e2 Hpo | e1 e2 Hrf | e1 e2 Hd | e1 e2 Hws | e1 e2 Hfr].
    left; exists e; split; auto.
    right; left; exists e1; exists e2; split; [|split]; auto.
    right; right; left; exists e1; exists e2; split; [|split]; auto.
    right; right; right; right; left; exists e1; exists e2; split; [|split]; auto.
    right; right; right; left; exists e1; exists e2; split; [|split]; auto.
    right; right; right; right; right; left; exists e1; exists e2; split; [|split]; auto.
    right; right; right; right; right; right; exists e1; exists e2; split; [|split]; auto.
Qed.

Lemma tc_path_dec :
  forall E X delay l1 l2,
  wf_delay delay ->
  tc (X2p E X delay) l1 l2 ->
  ((exists de, l1 = d de /\ l2 = f de) \/
     ~delay E X (evt_of_label l1) (evt_of_label l2) \/
     ~delay E X (evt_of_label l2) (evt_of_label l1)).
Proof.
intros E X delay l1 l2 Hwfd H12; induction H12 as [x y Hs|].

  generalize (path_dec E X delay x y Hs);  
  intros [? | [[dx [dy [Hx [Hy Hxy]]]] | [[dx [dy [Hx [Hy Hxy]]]] | [[dx [dy [Hx [Hy Hxy]]]] | [[dx [dy [Hx [Hy Hxy]]]] | [[dx [dy [Hx [Hy Hxy]]]] | [dx [dy [Hx [Hy Hxy]]]] ]]]]]]; auto.
    right; left; rewrite <- Hx; rewrite <- Hy; auto. 
    generalize (excluded_middle (delay E X (evt_of_devt dx) (evt_of_devt dy))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; rewrite <- Hx; rewrite <- Hy; simpl; auto.
        apply delay_antisym; auto.
       right; left; rewrite <- Hx; rewrite <- Hy; auto.
    rewrite <- Hx; rewrite <- Hy; simpl; right; left; apply delay_antisym; auto.
    generalize (excluded_middle (delay E X (evt_of_devt dx) (evt_of_devt dy))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; rewrite <- Hx; rewrite <- Hy; simpl; auto.
        apply delay_antisym; auto.
       right; left; rewrite <- Hx; rewrite <- Hy; auto.
    generalize (excluded_middle (delay E X (evt_of_devt dx) (evt_of_devt dy))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; rewrite <- Hx; rewrite <- Hy; simpl; auto.
        apply delay_antisym; auto.
       right; left; rewrite <- Hx; rewrite <- Hy; auto.
    generalize (excluded_middle (delay E X (evt_of_devt dx) (evt_of_devt dy))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; rewrite <- Hx; rewrite <- Hy; simpl; auto.
        apply delay_antisym; auto.
       right; left; rewrite <- Hx; rewrite <- Hy; auto.    

  inversion IHtc1 as [Hdf1 | [Hnd1 | Hd1]];
  inversion IHtc2 as [Hdf2 | [Hnd2 | Hd2]];
  clear IHtc1 IHtc2 H12_ H12_0. 

    destruct Hdf1 as [de1 [Hd1 Hf1]]; destruct Hdf2 as [de2 [Hd2 Hf2]]; rewrite Hd2 in Hf1; inversion Hf1.
    
    destruct Hdf1 as [de1 [Hd1 Hf1]]; rewrite Hf1 in Hnd2; rewrite Hd1; simpl;
    right; left; auto.

    destruct Hdf1 as [de1 [Hd1 Hf1]]; rewrite Hf1 in Hd2; rewrite Hd1; simpl;
    right; right; auto.

    destruct Hdf2 as [de2 [Hd2 Hf2]]; rewrite Hd2 in Hnd1; rewrite Hf2; simpl;
    right; left; auto. 

    right; left; apply delay_tc with (evt_of_label z); auto.

    generalize (excluded_middle (delay E X (evt_of_label x) (evt_of_label y))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; apply delay_antisym; auto.
        right; left; auto.

   destruct Hdf2 as [de2 [Hd2 Hf2]]; rewrite Hd2 in Hd1; rewrite Hf2; simpl;
    right; right; auto.

    generalize (excluded_middle (delay E X (evt_of_label x) (evt_of_label y))); intro Hor;
      inversion Hor as [Hd | Hnd]; clear Hor.
        right; right; apply delay_antisym; auto.
        right; left; auto.

    right; right; apply delay_tc with (evt_of_label z); auto.
Qed.

Lemma ac_ghb_implies_ac_X2p :
  forall E X delay,
  well_formed_event_structure E ->
  valid_execution E X ->
  wf_delay delay ->
  acyclic (X2p E X delay).
Proof.
intros E X delay Hwf Hv Hwfd x Hx.
generalize (tc_path_dec E X delay x x Hwfd Hx); intros [Hdf | [Hnd | Hnd]].

  destruct Hdf as [dx [Hd Hf]]; rewrite Hf in Hd; inversion Hd.
  apply (delay_irr delay Hwfd E X (evt_of_label x)  Hnd).
  apply (delay_irr delay Hwfd E X (evt_of_label x)  Hnd).
Qed.

Lemma X2p_d_po :
  forall E X delay l1 l2 de1 de2,
  well_formed_event_structure E ->
  l1 = d de1 -> l2 = d de2 ->
  X2p E X delay l1 l2 ->
  po_iico E (evt_of_devt de1) (evt_of_devt de2).
Proof.
intros E X delay l1 l2 de1 de2 Hwf H1 H2 H12.
rewrite H1 in H12; rewrite H2 in H12.
inversion H12; auto.
Qed.

Lemma label_implies_evt :
  forall E X l de,
  well_formed_event_structure E ->
  rfmaps_well_formed E (events E) (rf X) ->
  labels_of_es E X l ->
  l = d de \/ l = f de ->
  events E (evt_of_devt de).
Proof.
intros E X l de Hwf Hrfwf [Hw | Hr] Hor.

  destruct Hw as [w [Hw [Hdw | Hfw]]]; inversion Hor as [Hd | Hf]; clear Hor.
    rewrite Hdw in Hd; inversion Hd; simpl; destruct Hw; auto.
    rewrite Hdw in Hf; inversion Hf.
    rewrite Hfw in Hd; inversion Hd.
    rewrite Hfw in Hf; inversion Hf; simpl; destruct Hw; auto.

  destruct Hr as [w [r [Hrf [Hdr | Hfr]]]]; inversion Hor as [Hd | Hf]; clear Hor.
    rewrite Hdr in Hd; inversion Hd; simpl; apply ABasic.ran_rf_in_events with X w; auto.
    rewrite Hdr in Hf; inversion Hf.
    rewrite Hfr in Hd; inversion Hd.
    rewrite Hfr in Hf; inversion Hf; simpl; apply ABasic.ran_rf_in_events with X w; auto.
Qed.

Lemma dom_X2p_is_label :
  forall E X delay l l',
  well_formed_event_structure E ->
  write_serialization_well_formed (events E) (ws X) ->
  rfmaps_well_formed E (events E) (rf X) ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) ->
  wf_rf_labels E (labels_of_es E X) ->
  tc (X2p E X delay) l l' ->
  labels_of_es E X l.
Proof.
intros E X delay l l' Hwf Hwfws Hrfwf Hwfd Hwfl Hwfrfl Hll'.

induction Hll' as [x y Hxy|]; auto.

generalize Hwfl; intros Hwfl'; destruct Hwfl' as [? [? [? Hldf]]];
inversion Hxy as [? | ? | ? | ? | ? | ? | ?]; auto.
Qed.

Lemma ran_X2p_is_label :
  forall E X delay l l',
  well_formed_event_structure E ->
  write_serialization_well_formed (events E) (ws X) ->
  rfmaps_well_formed E (events E) (rf X) ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) ->
  wf_rf_labels E (labels_of_es E X) ->
  tc (X2p E X delay) l l' ->
  labels_of_es E X l'.
Proof.
intros E X delay l l' Hwf Hwfws Hrfws Hwfd Hwfl Hwfrfl Hll'.
induction Hll' as [x y Hxy|]; auto.

generalize Hwfl; intros Hwfl'; destruct Hwfl' as [? [? [? Hldf]]];
inversion Hxy as [? | ? | ? | ? | ? | ? | ?]; auto.
Qed.

Lemma wfl_es :
  forall E X,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_labels E (labels_of_es E X).
Proof.
intros E X Hwf Hv; split; [|split; [|split]].
  
  split; intros w [Hw|Hr].
    destruct Hw as [w' [Hw' [Heq | Heq]]]; inversion Heq; auto.
    destruct Hr as [w' [r' [? [Heq | Heq]]]]; inversion Heq.
    destruct Hw as [w' [Hw' [Heq | Heq]]]; inversion Heq; auto.
    destruct Hr as [w' [r' [? [Heq | Heq]]]]; inversion Heq.

  split; intros w r [Hw|Hr].
    destruct Hw as [w' [Hw' [Heq | Heq]]]; inversion Heq.
    destruct Hr as [w' [r' [? [Heq | Heq]]]]; inversion Heq; auto.
      destruct_valid Hv; apply Hrf_cands; auto.
    destruct Hw as [w' [Hw' [Heq | Heq]]]; inversion Heq.
    destruct Hr as [w' [r' [? [Heq | Heq]]]]; inversion Heq; auto.
      destruct_valid Hv; apply Hrf_cands; auto.

  intros w Hw; left; exists w; split; auto.

  intros de [Hw|Hr].
    destruct Hw as [w [Hw [Hl|Hl]]]; left; exists w; inversion Hl; auto.
    destruct Hr as [w [r [Hrf [Hl|Hl]]]]; right; exists w; exists r; inversion Hl; auto.
Qed.
    
Lemma wfl_rf_es :
  forall E X,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_rf_labels E (labels_of_es E X).
Proof.
intros E X Hwf Hv; split.

  intros r Hr; destruct_valid Hv; generalize (Hrf_init r Hr); intros [w [Hew Hrf]];
  exists w; split; auto.
    right; exists w; exists r; split; auto. 

  split; intros r w1 w2 [Hl1 | Hl1] [Hl2 | Hl2].
    destruct Hl1 as [w1' [Hw1' [Hl1 | Hl1]]]; inversion Hl1.
    destruct Hl1 as [w1' [Hw1' [Hl1 | Hl1]]]; inversion Hl1.
    destruct Hl2 as [w2' [Hw2' [Hl2 | Hl2]]]; inversion Hl2.

    destruct Hl1 as [w1' [r1' [Hrf1 [Hl1 | Hl1]]]]; destruct Hl2 as [w2' [r2' [Hrf2 [Hl2 | Hl2]]]];
    inversion Hl1; inversion Hl2.
      rewrite <- H0 in * |- *; rewrite H1 in * |- *; rewrite <- H2 in * |- *; rewrite <- H3 in * |- *; auto.
      destruct_valid Hv; apply Hrf_uni with r1'; auto.

    destruct Hl1 as [w1' [Hw1' [Hl1 | Hl1]]]; inversion Hl1.
    destruct Hl1 as [w1' [Hw1' [Hl1 | Hl1]]]; inversion Hl1.
    destruct Hl2 as [w2' [Hw2' [Hl2 | Hl2]]]; inversion Hl2.

    destruct Hl1 as [w1' [r1' [Hrf1 [Hl1 | Hl1]]]]; destruct Hl2 as [w2' [r2' [Hrf2 [Hl2 | Hl2]]]];
    inversion Hl1; inversion Hl2.
      rewrite <- H0 in * |- *; rewrite H1 in * |- *; rewrite <- H2 in * |- *; rewrite <- H3 in * |- *; auto.
      destruct_valid Hv; apply Hrf_uni with r1'; auto.
Qed.

Lemma tc_X2p_partial_order :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  partial_order (tc (X2p E X delay)) (labels_of_es E X).
Proof.
intros E X delay Hwf Hv Hwfd; generalize Hv; intro Hv';
  generalize (wfl_es E X Hwf Hv); intro Hwfl;
  generalize (wfl_rf_es E X Hwf Hv); intro Hwfrfl;
  destruct_valid Hv; split; [|split].
  intros ? [l [l' Hd] | l [l' Hr]].
    apply dom_X2p_is_label with delay l'; auto; split; auto.
    apply ran_X2p_is_label with delay l'; auto; split; auto.

  intros x1 x2 x3 [H12 H23]; apply trc_ind with x2; auto.

  intros x; apply ac_ghb_implies_ac_X2p; auto.
Qed.

Lemma in_queue_implies_in_order :
  forall E X p lst delay n s0 s1 lb de,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_labels E (labels_of_es E X) ->
  wf_delay delay ->
  rel_incl p (pred (OEL.LE (tc (X2p E X delay))) lb) ->
  machine E p (labels_of_es E X) lst n (Some s0) (Some s1) ->
  wf_rf_labels E (labels_of_es E X) ->
  wf_queue E (Q s1) ->
  ran (Q s1) (evt_of_devt de) ->
  labels_of_es E X (d de) ->
  (OEL.LE (tc (X2p E X delay))) (d de) lb .
Proof.
intros E X p lst delay n s0 s2 lb de Hwf Hv Hwfl Hwfd Hi Hm  Hwfr Hwfq Hq Hlde.
generalize E X p lst delay lb s2 Hwf Hv Hwfl Hwfd Hi Hm  Hwfr Hwfq Hq Hlde;
clear E X p lst delay lb s2 Hwf Hv Hwfl Hwfd Hi Hm  Hwfr Hwfq Hq Hlde;
induction n; intros E X p lst delay lb s2 Hwf Hv Hwfl Hwfd Hi Hm  Hwfr Hwfq Hq Hlde.

  inversion Hm as [Heq [? [? [Hiq ?]]]]; inversion Heq as [Heqs]; rewrite <- Heqs in Hq.
  assert (ran (Q s0) (evt_of_devt de)) as Hudr.
    auto.
  generalize (Hiq de); intros [dir bak]; generalize (dir Hudr); intros [x [y Hd] | y [x Hd]];
  generalize (Hi x y Hd); intros [? ?]; auto.
    generalize (oel_trans (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd));
    intros Htrans; apply Htrans with y; auto.

  destruct Hm as [lb' [s1 [Hd [Hm [Hs [? Hwfs]]]]]];
  fold (machine E (pred p lb') (labels_of_es E X) lb' n (Some s0) (Some s1)) in Hm.
  generalize (excluded_middle (ran (Q s1) (evt_of_devt de))); 
  intro Hor; inversion Hor as [Hqs1 | Hnqs1]; clear Hor.
    assert (rel_incl (pred p lb') (pred (OEL.LE (tc (X2p E X delay))) lb)) as Hip.
      intros e1 e2 [H12 Hor]; apply Hi; auto.
    destruct Hwfs as [? [? Hwfq1]];
    apply (IHn E X (pred p lb') lb' delay lb s1 Hwf Hv Hwfl Hwfd Hip Hm Hwfr Hwfq1 Hqs1 Hlde).

    assert (ran p lb') as Hran.
      apply last_in_ran; auto.
    assert (lb' = d de) as Hlb'.
      apply q_step with E p (labels_of_es E X)  s1 s2; auto.
        destruct Hwfl as [[Hw_d Hw_f] ?]; auto.
        generalize (oel_udr (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd));
        intro Hudr; apply Hudr; auto.

          destruct Hran as [x Hran]; generalize (Hi x lb' Hran); intros [? ?]; right; exists x; auto.
          

    destruct Hran as [lb'' Hdlb'']. 
    generalize (Hi lb'' lb' Hdlb''); intros [Hplb'lb'' Hor].
      rewrite Hlb' in Hor; auto. 
Qed.

Lemma dreads_dont_block :
  forall E X lst delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> wf_labels E (labels_of_es E X) ->
  wf_mem E (M s1) -> wf_queue E (Q s1) ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (d (DR (w1,r1)))) (labels_of_es E X) lst n (Some s0) (Some s1) ->
  ~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (d (DR (w1,r1))) = None).
Proof. 
intros E X lst delay n s0 s1 w1 r1 Hwf Hv Hwfl Hwfm Hwfq Hm Hs; simpl in Hs;
unfold buff_read_step in Hs.

  inversion Hs.
Qed.

Lemma dwrites_dont_block :
  forall E X lst delay n s0 s1 w1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> wf_labels E (labels_of_es E X) ->
  wf_mem E (M s1) -> wf_queue E (Q s1) ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (d (DW w1))) (labels_of_es E X) lst n (Some s0) (Some s1) ->
  ~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (d (DW w1)) = None).
Proof. 
intros E X lst delay n s0 s1 w1 Hwf Hv Hwfl Hwfm Hwfq Hm Hs; simpl in Hs;
unfold buff_write_step in Hs.

  inversion Hs.
Qed.

Lemma in_queue_bef_aft_implies_not_flush :
  forall E p L s1 s2 l dr r,
  wf_labels E L -> 
  reads E r ->
  evt_of_devt dr = r ->  L l ->
  step E p L (Some s1) l = Some s2 ->
  ran (Q s1) r -> ran (Q s2) r ->
  l <> f dr.
Proof.
intros E p L s1 s2 l dr r Hwfl Hrr Hdr Hl Hs H1 H2 Heq; rewrite Heq in Hs.
assert (exists w, dr = DR (w,r)) as Hisr.
  case_eq dr.
    intros [w r'] Heqdr;
    exists w; rewrite Heqdr in Hdr; simpl in Hdr; rewrite Hdr; auto.
    intros w Heqdr; rewrite Heqdr in Heq; rewrite Heq in Hl.
    destruct Hwfl as [[? Hfw] ?]; generalize (Hfw w Hl); intros [? [? [? Haw]]].
    rewrite Heqdr in Hdr; simpl in Hdr; rewrite <- Hdr in Hrr; destruct Hrr as [? [? [? Har]]].
    rewrite Haw in Har; inversion Har.
       
destruct Hisr as [w Heqdr]; rewrite Heqdr in Hs;
unfold step in Hs; unfold flush_resolved_read_step in Hs.

  destruct (bin (ran (Q s1)) r); [|inversion Hs].
  destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); [|inversion Hs].
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1))
                (fun e : Event =>
                 rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].
  destruct (eqEv_dec w (M s1 (loc r))).
    destruct (bemp (rrestrict (B s1)
              (fun w : Event => po_iico E w r /\ loc r = loc w))); [|inversion Hs].
    inversion Hs as [Heqs]; rewrite <- Heqs in H2; simpl in H2.
    destruct H2 as [x Hi]; unfold del_queue in Hi.
      destruct Hi as [? Hc]; apply Hc; auto.

  destruct (bin (Intersection Event (ran (B s1)) (rrB r)) w); [inversion Hs|].
  destruct (bin (fun w : Event => B s1 w (waft E (B s1) r) /\ loc r = loc w) w); [|inversion Hs].
  destruct (bin (fun w : Event =>
              (w = wbef E (B s1) r \/ B s1 (wbef E (B s1) r) w) /\
              loc r = loc w) w); [|inversion Hs].
    inversion Hs as [Heqs]; rewrite <- Heqs in H2; simpl in H2.
    destruct H2 as [x Hi]; unfold del_queue in Hi.
      destruct Hi as [? Hc]; apply Hc; auto.    
Qed.

Lemma in_queue_implies_f_after2 :
  forall E p (L:set Label) lst n s0 s1 lb dr r,
  wf_labels E L ->
  wf_rf_labels E L ->
  wf_queue E (Q s1) ->
  events E r ->
  evt_of_devt dr = r ->
  lb <> f dr ->
  L lb -> L (f dr) ->
  lst <> f dr ->
  L lst ->
  Included _ (Union _ (dom p) (ran p)) L ->
  acyclic p ->
  trans p ->
  (forall l1 l2, l1 <> l2 -> (L l1) -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall de, L (f de) -> tc p (d de) (f de)) ->
  (~(exists l', tc p lst l' /\ tc p l' lb) \/ lst = lb) ->
  ran (Q s1) r ->
  (exists r, rel_incl r (pred p lb) /\ 
     (~(exists l', tc p (last r) l' /\ tc p l' lst) \/ last r = lst) /\
     machine E r L lst n (Some s0) (Some s1)) ->
  tc p lb (f dr).
Proof.
intros E p L lst n s0 s2 lb dr r  Hwfl Hwfr Hwfq Her Hdr Hdiff Hlb Hldr Hnr Hllst Hudrp Hacp Htp Htotp Hdf Hlst Hw Hm; 
generalize p lst lb s2 Hdiff Hlb Hnr Hllst Hacp Htp Htotp  Hdf Hwfr Hwfq Hm Hw Hudrp Hlst; 
clear p lst lb s2 Hdiff Hlb Hnr Hllst Hacp Htp Htotp  Hdf Hwfr Hwfq Hm Hw Hudrp Hlst; induction n;
intros p lst lb s2 Hdiff Hlb Hnr Hllst Hacp Htp Htotp  Hdf Hwfr Hwfq Hm Hw Hudrp Hlst.
  destruct Hm as [dd [Hid [Hldd Hm]]].

  assert (L (f dr)) as Hlf.
    apply Hudrp; rewrite dom_ran_tc; auto; right; exists (d dr); apply Hdf; auto.
  destruct Hm as [Heq [? [? [Hiqd [Hiqf ?]]]]].
  generalize (Htotp lb (f dr) Hdiff Hlb Hlf); intros [|Hrlb]; [apply trc_step|]; auto.
  generalize (Htotp lst (f dr) Hnr Hllst Hlf); intros [Hlstr | Hrlst].
    inversion Hlst as [Hni | Heqlstlb].
      assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
        exists (f dr); split; apply trc_step; auto.
      contradiction.
      rewrite <- Heqlstlb in Hrlb. 
        assert (tc p (f dr) (f dr)) as Hc.
          apply trc_ind with lst; apply trc_step; auto.
        generalize (Hacp (f dr) Hc); intro Ht; inversion Ht.
  destruct (eqLabel_dec (last dd) (f dr)) as [Heql | Hneql].
  
  assert (udr dd (f dr)) as Hifr.
    right; apply last_in_ran; auto.
  generalize (Hiqf dr); intros [? bakf]; generalize (bakf Hifr); 
  rewrite Hdr; inversion Heq as [Heqs]; intro; contradiction.

  assert (L (last dd)) as HLdd.
    apply Hudrp.
      assert (last dd = last dd) as Ht.
        auto.
      generalize (last_in_ran dd (last dd) Ht); intros [x Hx];
      generalize (Hid x (last dd) Hx); intros [? ?]; left; exists lb; auto.
  generalize (Htotp (last dd) (f dr) Hneql HLdd Hlf); intro Hor.
  inversion Hor as [Hlr | Hrl]; clear Hor.
    inversion Hldd as [Hni | Heqllst].
      assert (exists l' : Label, tc p (last dd) l' /\ tc p l' lst) as Hc.
        exists (f dr); split; apply trc_step; auto.
      contradiction.
      rewrite Heqllst in Hlr.
        assert (tc p (f dr) (f dr)) as Hc.
          apply trc_ind with lst; apply trc_step; auto.
        generalize (Hacp (f dr) Hc); intro Ht; inversion Ht.
    
  assert (udr dd (f dr)) as Hifr.
    apply before_last_implies_in with p; auto.
    intros x y Hxy; generalize (Hid x y Hxy); intros [? ?]; auto.
  generalize (Hiqf dr); intros [? bakf]; generalize (bakf Hifr); 
  rewrite Hdr; inversion Heq as [Heqs]; intro; contradiction.

  destruct Hm as [dd [Hid [Hldd Hm]]].
  destruct Hm as [lb' [s1 [Hd [Hm [Hs12 [? [? [? Hwfq1]]]]]]]]; fold (machine E (pred dd lb') L lb' n (Some s0) (Some s1)) in Hm. 
  generalize (excluded_middle (ran (Q s1) r)); intro Hor; inversion Hor as [Hi | Hni]; clear Hor; auto.

  assert (L lb') as Hlb'.
    apply Hudrp.
      generalize (last_in_ran dd lb' Hd); intros [x Hr];
      generalize (Hid x lb' Hr); intros [? ?]; left; exists lb; auto.

  assert (lb' <> f dr) as Hdiff'.
    apply in_queue_bef_aft_implies_not_flush with E dd L s1 s2 r; auto.
  assert (~ (exists l' : Label, tc p lb' l' /\ tc p l' lb') \/ lb' = lb') as Hllb'.
    auto.
  assert (exists r0 : Rln Label,
    rel_incl r0 (pred p lb') /\
     (~(exists l', tc p (last r0) l' /\ tc p l' lb') \/ last r0 = lb') /\
    machine E r0 L lb' n (Some s0) (Some s1)) as Hed.
    exists (pred dd lb'); split; [|split]; auto.
       intros x y [Hxy Hylb'];  
       generalize (Hid x y Hxy); intros [? ?]; 
       generalize (Hid y lb' Hylb'); intros [? ?]; split; auto.
       left; apply no_interv_last; auto. 
         intros e1 e2 H12; generalize (Hid e1 e2 H12); intros [? ?]; auto.     

  generalize (IHn p lb' lb' s1 Hdiff' Hlb' Hdiff' Hlb' Hacp Htp Htotp Hdf Hwfr Hwfq1 Hed Hi Hudrp Hllb'); auto.
  intro Hlb'r.

  assert (L (f dr)) as Hlf.
    apply Hudrp; auto.
      rewrite dom_ran_tc; right; exists lb'; auto.
  generalize (Htotp lb (f dr) Hdiff Hlb Hlf); intros [|Hrlb]; [apply trc_step; auto|].
  generalize (Htotp lst (f dr) Hnr Hllst Hlf); intros [Hlstr | Hrlst].
  inversion Hlst as [Hn | Heq].
    assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
      exists (f dr); split; apply trc_step; auto.
    contradiction.
    assert (tc p lb lb) as Hc.
      apply trc_ind with (f dr); apply trc_step; auto; rewrite <- Heq; auto.
    generalize (Hacp lb Hc); intro Ht; inversion Ht.
  
    inversion Hldd as [Hnldd | Heqldd].
    assert (exists l' : Label, tc p (last dd) l' /\ tc p l' lst) as Hco.
      exists (f dr); split; auto.
        rewrite Hd; auto.
        apply trc_step; auto.
    contradiction.
    rewrite Hd in Heqldd; rewrite Heqldd in Hlb'r.
    assert (tc p lst lst) as Hc.
      apply trc_ind with (f dr); auto; apply trc_step; auto.
    generalize (Hacp lst Hc); intro Ht; inversion Ht.

    assert (lb' = d dr) as Hlb'.
      apply q_step with E dd L s1 s2; auto.
        destruct Hwfl as [[? ?] ?]; auto.
        apply Hwfq; rewrite Hdr; auto.
        rewrite Hdr; auto. rewrite Hdr; auto.
        apply Hudrp.
          generalize (last_in_ran dd lb' Hd); intros [x Hxlb'];
          generalize (Hid x lb' Hxlb'); intros [? ?]; left; exists lb; auto.
        destruct Hwfl as [? [? [? Hdfl]]]; apply Hdfl; auto.
    assert (tc p lb' (f dr)) as Hpdf.
      rewrite Hlb'. apply Hdf; auto.

  assert (L (f dr)) as Hlf.
    apply Hudrp; rewrite dom_ran_tc; right; exists lb'; auto.

    generalize (Htotp lb (f dr) Hdiff Hlb Hlf); intro Hor; 
      inversion Hor as [? | Hflb]; clear Hor; apply trc_step; auto.

      rewrite Hd in Hldd.
      generalize (Htotp lst (f dr) Hnr Hllst Hlf); intros [Hlstr | Hrlst].
        inversion Hlst as [Hnir | Heqlstlb].
          assert (exists l' : Label, tc p lst l' /\ tc p l' lb) as Hc.
            exists (f dr); split; apply trc_step; auto.
          contradiction.
          assert (tc p lb lb) as Hc.
            apply trc_ind with (f dr); rewrite Heqlstlb in Hlstr; apply trc_step; auto.
          generalize (Hacp lb Hc); intro Ht; inversion Ht.

        inversion Hldd as [Hnldd | Heqldd].
        assert (exists l' : Label, tc p lb' l' /\ tc p l' lst) as Hc.
          exists (f dr); split; auto; apply trc_step; auto.
        contradiction.

      rewrite Heqldd in Hpdf.
      assert (tc p lst lst) as Hc.
        apply trc_ind with (f dr); auto; apply trc_step; auto.
      generalize (Hacp lst Hc); intro Ht; inversion Ht.        
Qed.

Lemma not_flush_write_step_implies_same_mem :
  forall E p L lst n s0 s1 s2 lb w l,
  wf_mem E (M s1) ->
  machine E (pred p lb) L lst n (Some s0) (Some s1) ->
  step E p L (Some s1) lb = Some s2 ->
  
  trans p ->
  ~(exists w', loc w' = loc w /\ (p (f (DW w)) (f (DW w'))) /\ (p (f (DW w')) lb \/ (f (DW w')) = lb)) -> 
  (M s1) l = w ->
  (M s2) l = w.
Proof.
intros E p L lst n s0 s1 s2 lb w l Hwfm Hm Hs Htp Hd Hms0.
case_eq lb; intros de Hlb; rewrite Hlb in Hs;
case_eq de; intros c Hde; rewrite Hde in Hs; simpl in Hs.



  
  destruct c as [w' r']; unfold buff_read_step in Hs;
  inversion Hs; auto.

  
  unfold buff_write_step in Hs;
  inversion Hs; auto.


  destruct c as [w' r'];
  apply flush_read_same_mem with E p L s1 w' r'; auto.

  
  unfold flush_write_step in Hs;
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e c))); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun r' : Event => loc r' = loc c /\ po_iico E r' c))); [|inversion Hs];
  destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc c))) c); [|inversion Hs];
  inversion Hs as [Heqs]; auto.
    simpl; unfold upd_mem.
      rewrite Hlb in Hd; rewrite Hde in Hd; simpl in Hd.
      destruct (eqLoc_dec l (loc c)) as [Heq | Hneq]; auto.
       assert False as Ht.
          apply Hd; auto.
            exists c; split; [|split]; auto.
            rewrite <- Heq; rewrite <- Hms0; apply sym_eq; apply Hwfm.

            assert ((exists d : Rln Label,
                     trans d /\
                     rel_incl d (pred p (f (DW c))) /\ machine E d L lst n (Some s0) (Some s1))) as Hex.
              rewrite Hde in Hlb; rewrite Hlb in Hm; exists (pred p (f (DW c))); 
                split; [|split]; auto.
                  apply td_tpred; auto. intros x y Hxy; auto.      
            assert (w = M s1 l) as Heqw.
              apply sym_eq; auto.
            apply (in_mem_implies_in_order E p L lst n s0 s1 (f (DW c)) w l Htp Hex Heqw).
       
        inversion Ht.
Qed.

Lemma in_order_implies_in_mem : 
  forall E p L lst n s0 s1 w lb,
  wf_mem E (M s1) ->
  trans p -> 
  ~(exists w', loc w' = loc w /\ (p (f (DW w)) (f (DW w'))) /\ (p (f (DW w')) lb \/ (f (DW w')) = lb)) -> 
  (exists d, trans d /\ rel_incl d (pred p lb) /\ machine E d L lst n (Some s0) (Some s1)) ->
  M s1 (loc w) = w.
Proof. 
intros E p L lst n s0 s2 w lb'' Hwfm Htp  Hniw Hm; 
generalize s2 lst Hwfm Hm; clear s2 lst Hwfm Hm; induction n; 
intros s2 lst Hwfm [pd [Htpd [Hi Hm]]].

  inversion Hm.
    admit. 

  destruct Hm as [lb [s1 [Hdlb [Hm [Hs [? [Hwfm1 ?]]]]]]];
  fold (machine E (pred pd lb) L lb n (Some s0) (Some s1)) in Hm.
  assert (exists d, trans d /\ rel_incl d (pred p lb'') /\
         machine E d L lb n (Some s0) (Some s1)) as Hex.
    exists (pred pd lb); split; [|split]; auto. 
      apply td_tpred; auto.
      intros x y [Hxy ?]; apply Hi; auto.
  generalize (IHn s1 lb Hwfm1 Hex); intro His1.

  apply not_flush_write_step_implies_same_mem with E pd L lb n s0 s1 lb; auto.
  intros [w' [Heql [Hww' Hor]]].
  assert (exists w', loc w' = loc w /\ (p (f (DW w)) (f (DW w'))) /\ (p (f (DW w')) lb'' \/ (f (DW w') = lb''))) as Hc.
    exists w'; split; [|split]; auto.
       
      generalize (Hi (f (DW w)) (f (DW w')) Hww'); intros [? ?]; auto.
      inversion Hor as [Hw'lb | Heq]; clear Hor.
        generalize (Hi (f (DW w')) lb Hw'lb); intros [Hpw'lb Hplblb'']; auto.
          
            left; apply Htp with lb; auto.
            
      generalize (Hi (f (DW w)) (f (DW w')) Hww'); intros [? ?]; auto.

  contradiction.        
Qed.

Lemma in_order_implies_ws :
  forall E X delay w w',
  well_formed_event_structure E ->
  valid_execution E X ->
  wf_delay delay ->
  loc w = loc w' ->
  wf_labels E (labels_of_es E X) ->
  OEL.LE (tc (X2p E X delay)) (f (DW w)) (f (DW w')) ->
  ws X w w'.
Proof.
intros E X delay w w' Hwf Hv Hwfd Heql Hwfl Hpff.

generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.

destruct (eqEv_dec w w') as [Heq | Hneq].
  rewrite Heq in Hpff. 
  generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpart);
  intros Hac; rewrite <- trans_tc with Label (OEL.LE (tc (X2p E X delay))) in Hpff.
  generalize (Hac (f (DW w')) Hpff); intro Hf; inversion Hf.
    apply oel_trans with (labels_of_es E X); auto.

assert (Included Label
    (Union Label (dom (OEL.LE (tc (X2p E X delay))))
       (ran (OEL.LE (tc (X2p E X delay))))) (labels_of_es E X)) as Hudr.
  apply oel_udr; auto.

assert (labels_of_es E X (f (DW w))) as Hlw.
  apply Hudr; left; exists (f (DW w')); auto.

assert (labels_of_es E X (f (DW w'))) as Hlw'.
  apply Hudr; right; exists (f (DW w)); auto.

assert (ws X w w' \/ ws X w' w) as Hor.
  destruct_valid Hv.
  generalize (Hws_tot (loc w)); intro Hlin; destruct_lin Hlin.
  assert ((writes_to_same_loc_l (events E) (loc w)) w) as Hww.
    destruct Hwfl as [[? Hfw] ?]; generalize (Hfw w Hlw); intros [? [l [v Haw]]]; 
    split; auto; exists v; unfold loc; rewrite Haw; auto.
  assert ((writes_to_same_loc_l (events E) (loc w)) w') as Hww'.
    rewrite Heql; destruct Hwfl as [[? Hfw] ?]; generalize (Hfw w' Hlw'); intros [? [l [v Haw]]]; 
    split; auto; exists v; unfold loc; rewrite Haw; auto.
  generalize (Htot w w' Hneq Hww Hww'); intros [[? ?]|[? ?]]; auto.

inversion Hor as [|Hw'w]; auto.

assert ((tc (X2p E X delay)) (f (DW w')) (f (DW w))) as Hpw'w.
  apply trc_step; apply X2p_ws; auto.
     

generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd));
intro Hi; generalize (Hi (f (DW w')) (f (DW w)) Hpw'w); intro Hc.
generalize (oel_trans (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd) (f (DW w)) (f (DW w')) (f (DW w)) Hpff Hc);
intro Hww.
assert (tc (OEL.LE (tc (X2p E X delay))) (f (DW w)) (f (DW w))) as Htc.
  apply trc_step; auto.
generalize (oel_ac (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd) (f (DW w)) Htc);
intro Ht; inversion Ht.
Qed.

Lemma wfl_le2 :
  forall E X,
  well_formed_event_structure E ->
  valid_execution E X ->
  ((forall l : Label, labels_of_es E X l -> events E (evt_of_label l))).
Proof.
intros E X Hwf Hv l [Hw | Hr]; destruct_valid Hv.
  destruct Hw as [w [[Hew ?] [Hl | Hl]]]; rewrite Hl; auto.
  destruct Hr as [w [r [Hrf [Hl | Hl]]]]; rewrite Hl; simpl;
  apply ABasic.ran_rf_in_events with X w; auto; split; auto.
Qed.

Lemma in_order_implies_mem_or_buff_state :
  forall E X  delay n s0 s1 w r,
  well_formed_event_structure E ->
  valid_execution E X ->
  wf_delay delay ->
  wf_mem E (M s1) -> 
  wf_labels E (labels_of_es E X) -> 
  labels_of_es E X (f (DW w)) -> labels_of_es E X (f (DR (w,r))) ->
  rf X w r ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w,r)))) (labels_of_es E X) (f (DR (w,r))) n (Some s0) (Some s1) ->
  (M s1 (loc r) = w \/ ran (B s1) w).
Proof. 
intros E X  delay n s0 s1 w r Hwf Hv Hwfd Hwfm Hwfl Hlfw Hlfwr Hrf Hm.
generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
generalize (oel_tot (X2p E X delay) (labels_of_es E X) Hpart); intro Htotp.
assert (f (DW w) <> f (DR (w,r))) as Hdiff.
  intro Hfeq; inversion Hfeq as [Heq].
generalize (Htotp (f (DW w)) (f (DR (w,r))) Hdiff Hlfw Hlfwr); intro Hor;
inversion Hor as [Hpwr | Hprw]; clear Hor.

  assert (loc w = loc r) as Heql.
    assert (rfmaps_well_formed E (events E) (rf X)) as Hwfrf.
      destruct_valid Hv; split; auto.
      apply rf_implies_same_loc2 with E X; auto; split; auto.
        destruct_valid Hv; split; auto.
  rewrite <- Heql; left; apply in_order_implies_in_mem with E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X)  (f (DR (w,r))) n s0 (f (DR (w,r))); auto.
    apply oel_trans with (labels_of_es E X) ; auto.
    intros [w' [Heqlw' [Hww' Hw'r]]].
      assert (loc w = loc w') as Heql'.
        apply sym_eq; auto.
      
      generalize (in_order_implies_ws E X delay w w' Hwf Hv Hwfd Heql' Hwfl Hww'); intro Hws.
      assert (fr E X r w') as Hfr.
          split; [|split]; auto.
          apply ABasic.ran_rf_in_events with X w; auto.
            destruct_valid Hv; split; auto.
          apply ABasic.ran_ws_in_events with X w; auto.
            destruct_valid Hv; split; auto.
          exists w; split; auto.
      
         
      assert (labels_of_es E X (f (DW w'))) as Hflw'.
        destruct Hwfl as [? [? [Hw ?]]]; apply Hw. 
        split. 
          apply ABasic.ran_ws_in_events with X w; auto.
            destruct_valid Hv; split; auto.
          assert (write_serialization_well_formed (events E) (ws X)) as Hwswf.
            destruct_valid Hv; split; auto.
          generalize (ABasic.ran_ws_is_write E X w w' Hwswf Hws); intros [v [l Haw]]; exists l; exists v; auto.
      
      generalize (X2p_fr E X delay (DR (w,r)) (DW w')  Hfr Hlfwr Hflw'); intro Hprw'.
      apply (oel_ac (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd) (f (DR (w,r)))).
      generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi;
      inversion Hw'r as [Htcw'r | Heqw'r]; clear Hw'r.
        apply trc_ind with (f (DW w')); apply trc_step; auto.
           apply Hi; apply trc_step; auto.
        rewrite Heqw'r in Hprw'; apply trc_step; apply Hi; apply trc_step; auto.

    


    exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w,r)))); split; [|split]; auto.
      apply td_tpred; auto. apply oel_trans with (labels_of_es E X) ; auto.
      intros x y Hxy; auto. 

  assert (Included _ (labels_of_es E X) (labels_of_es E X)) as Htl.
    intros e He; auto.
  generalize (OEL.OE (labels_of_es E X) (labels_of_es E X) (tc (X2p E X delay)) Htl Hpart);
    clear Htl; intros [Hil Hl].
  destruct_lin Hl.
  assert (trans (OEL.LE (tc (X2p E X delay)))) as Htr.
    intros x1 x2 x3 H12 H23; apply Htrans with x2; auto.
  assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hacy.
    intros e He. 
      rewrite trans_tc in He; auto; apply (Hac e He).
    assert (exists d : Rln Label, trans d /\ rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w, r)))) /\
      ~(exists l' : Label,
         tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
         tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w, r)))) /\
    machine E d (labels_of_es E X) (f (DR (w, r))) n (Some s0) (Some s1)) as Hed.
      exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w,r)))); split; [|split; [|split]]; auto.
        apply td_tpred; auto.
        intros e1 e2 H12; auto.
        apply no_interv_last; auto; intros e1 e2 H12; auto.
     
  assert (forall l : Label, labels_of_es E X l -> events E (evt_of_label l)) as Hle.
    apply wfl_le2; auto.
  generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hi.
  assert (OEL.LE (tc (X2p E X delay)) (d (DW w)) (f (DR (w, r)))) as Hwr.
    apply Hi; apply trc_step; apply X2p_rf; auto.
  assert (forall de : DEvent, labels_of_es E X (f de) -> OEL.LE (tc (X2p E X delay)) (d de) (f de)) as Hdf.
    intros de Hde; apply Hi; apply trc_step; apply X2p_df; auto.
      apply label_implies_evt with X (f de); auto.
        destruct_valid Hv; split; auto.
  assert (~ (exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f (DR (w, r))) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w, r)))) \/
  f (DR (w, r)) = f (DR (w, r))) as Hni.
    auto.


  right; generalize (in_order_implies_in_buff E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DR (w,r))) n s0 s1 w (f (DR (w,r))) Hdr Htr Hacy Htotp Hle Hwr Hwr Hprw Hdf Hni Hed); 
  intro; exists dB; auto.
Qed. 

Lemma rf_p2X_X2p :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  rf (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X)) = rf X.
Proof.
intros E X delay Hwf Hv Hwfd Hwfl Hwfrfl; apply ext_rel; intros x y; split; intro Hxy.

  simpl in Hxy; unfold p2rf in Hxy.
  
  assert (reads E y) as Hry.    
    destruct Hxy as [[? [? [l [? [[v Hry] ?]]]]] ?].
    split; auto; exists l; exists v; auto.
  destruct_valid Hv; generalize (Hrf_init y Hry); intros [w [Hew Hrf]].
  
  destruct (eqEv_dec x w) as [Heq | Hneq]; subst; auto.
  destruct Hwfrfl as [? [? Huni]].
  assert (labels_of_es E X (f (DR (w,y)))) as Hlfw.
    right; exists w; exists y; split; auto.
  assert (labels_of_es E X (f (DR (x,y)))) as Hlfx.
    destruct Hxy; auto.
  generalize (Huni y w x Hlfw Hlfx); auto;
  intro; assert False as Ht.
    apply Hneq; auto.
  inversion Ht.

  simpl; unfold p2rf; split. 
    destruct_valid Hv; generalize (Hrf_cands x y Hxy); intro Hrfm; auto.
    right; exists x; exists y; split; auto.
Qed.

Lemma ws_p2X_X2p :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_labels E (labels_of_es E X) -> wf_delay delay ->
  ws (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X)) = ws X.
Proof.
intros E X delay Hwf Hv Hwfl Hwfd; apply ext_rel; intros x y; split; intro Hxy.

  simpl in Hxy; unfold p2ws in Hxy.
  destruct Hxy as [Hwx [Hwy [Hl [l1 [l2 [Hl1 [Hl2 H12]]]]]]].
  generalize (oel_ac (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hacp. 
  
  assert (x <> y) as Hdiff.
    intro Heq; rewrite Heq in Hl1; rewrite Hl1 in H12; rewrite Hl2 in H12.
    apply Hacp with (f (DW y)); apply trc_step; auto.

  assert ((writes_to_same_loc_l (events E) (loc x)) x) as Hwwx.
    destruct Hwx as [? [lx [vx Hax]]]; split; auto.
    exists vx; unfold loc; rewrite Hax; auto.

  assert ((writes_to_same_loc_l (events E) (loc x)) y) as Hwwy.
    destruct Hwy as [? [ly [vy Hay]]]; split; auto.
    rewrite Hl; exists vy; unfold loc; rewrite Hay; auto.

  generalize Hv; intro Hv'; destruct_valid Hv'; 
    generalize (Hws_tot (loc x)); intro Hlin; destruct_lin Hlin.
  generalize (Htot x y Hdiff Hwwx Hwwy); intros [[Hxy?] | [Hyx ?]]; auto.
  assert (OEL.LE (tc (X2p E X delay)) l2 l1) as H21.
    rewrite Hl1; rewrite Hl2.
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi;
    apply Hi; apply trc_step.
      destruct Hwfl as [? [? [Hw ?]]];   
      apply X2p_ws; auto. 
    assert (tc (OEL.LE (tc (X2p E X delay))) l1 l1) as Hc.
      apply trc_ind with l2; auto; apply trc_step; auto.
    generalize (Hacp l1 Hc); intro Ht; inversion Ht.

  simpl; unfold p2ws.
  generalize Hv; intro Hv'; destruct_valid Hv'.
  generalize (Hws_cands x y Hxy); intros [l [[? [vx Hwx]] [? [vy Hwy]]]].
    split; split; auto.
      exists l; exists vx; auto.
      split; auto; exists l; exists vy; auto.
        split. unfold loc; rewrite Hwx; rewrite Hwy; auto.
          exists (f (DW x)); exists (f (DW y)); split; [|split]; auto.
              destruct Hwfl as [? [? [Hw ?]]]; 
              generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi;
              apply Hi; apply trc_step; apply X2p_ws; auto.
                 
                  apply Hw; auto. split; auto; exists l; exists vx; auto.
                  apply Hw; auto. split; auto; exists l; exists vy; auto.
Qed.

Lemma p2X_X2p :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) -> 
  wf_delay delay ->
  p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) = X.
Proof.
intros E X delay Hwf Hv Hwfl Hwfrfl Hwfd.
assert (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) = 
  mkew (ws (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) )) 
    (rf (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X)) )) as Heq.
  simpl; auto.
rewrite Heq; rewrite ws_p2X_X2p; auto; rewrite rf_p2X_X2p; simpl; auto.
destruct X; simpl; auto.
Qed.

Lemma contrad_ppo :
  forall E X  delay n s0 s1 de,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) -> 
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  events E (evt_of_devt de) ->
  labels_of_es E X (f de) ->
  (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f de)) /\
    (~(exists l' : Label,
       tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
       tc (OEL.LE (tc (X2p E X delay))) l' (f de)) \/ last d = f de) /\
     machine E d (labels_of_es E X)  (f de) n (Some s0) (Some s1)) ->
  bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X) )) e (evt_of_devt de))) = false ->
  False.
Proof.
intros E X  delay n s0 s1 de' Hwf Hv Hwfl Hwfrfl Hwfd Hwfm Hwfb Hwfq Hede' Hlde' Hem Hppo.
  generalize Hwfl; intros Hwfl'.
  generalize (bemp_axf (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X) )) e (evt_of_devt de')) Hppo);
  clear Hppo; intros [e [Hi Hppo]].
    assert (rel_incl (pred (OEL.LE (tc (X2p E X delay))) (f de')) (pred (OEL.LE (tc (X2p E X delay))) (f de'))) as Ht.
      intros x y Hxy; auto.
    
      destruct Hi as [e' Hd].
        assert (events E e) as Hee.
          inversion Hd as [Hb | Hq].
            assert (udr (B s1) e) as Hudr.
              right; exists e'; auto.
            destruct Hwfb as [Hwfb ?]; generalize (Hwfb e Hudr); intros [? ?]; auto.
            assert (ran (Q s1) e) as Hran.
              exists e'; auto.
            generalize (Hwfq e Hran); intros [? ?]; auto.

       generalize (wfl_e2 E (labels_of_es E X) e Hwfl Hwfrfl Hee);
       intros [de [Hde Hlde]].

        assert (f de' <> f de) as Hdiff.
          intros Heqf; inversion Heqf as [Heq]; rewrite <- Heq in Hde; simpl in Hde.
          inversion Hppo as [Hpo | Hab].
            generalize (A.ppo_valid Hpo); auto. subst; apply ABasic.po_ac; auto. 
            subst; generalize Hab; apply A.ab_irr.

        assert (Included _ (labels_of_es E X) (labels_of_es E X)) as Htl.
          intros l Hl; auto.
        assert (partial_order (tc (X2p E X delay)) (labels_of_es E X)) as Hpl.
          apply tc_X2p_partial_order; auto.
        generalize (OEL.OE (labels_of_es E X) (labels_of_es E X) (tc (X2p E X delay)) Htl Hpl);
          clear Htl; intros [Hil Hl].
        assert (trans (OEL.LE (tc (X2p E X delay)))) as Httc.
          destruct_lin Hl; intros x y z Hxy Hyz; apply Htrans with y; auto.

  assert (tc (OEL.LE (tc (X2p E X delay))) (f de') (f de)) as Hfw1fe.
      case_eq de.

      intros [w r] Hldr.
      inversion Hd as [Hb | Hq].
        destruct Hwfl as [? [[? Hr] ?]].
        rewrite Hldr in Hlde; generalize (Hr w r Hlde); intros [? [? [l [? [[v Har] ?]]]]].
        assert (udr (B s1) e) as Hudr.
          right; exists e'; auto.
        destruct Hwfb as [Hw ?]; generalize (Hw e Hudr); intros [? [? [? Haw]]].
        rewrite Hldr in Hde; simpl in Hde; rewrite <- Hde in Har; rewrite Har in Haw; inversion Haw.
    
        assert (ran (Q s1) e) as Hiq.
          exists e'; auto.
        destruct Hwfl as [Hwflw ?].
        generalize (Hwfq e Hiq); intro Hre.

        assert (forall de, labels_of_es E X (f de) -> tc (OEL.LE (tc (X2p E X delay))) (d de) (f de)) as Hdf.
          intros de'' Hde''; generalize (X2p_df E X delay de''); intro Hip; apply trc_step; apply Hil; apply trc_step; auto.
          apply Hip; auto. 
            apply label_implies_evt with X (f de''); auto.
              destruct_valid Hv; split; auto.

        assert (evt_of_devt de = e) as Hde2.
          apply sym_eq; auto.
        assert (~(exists l' : Label,
          tc (OEL.LE (tc (X2p E X delay))) (f de') l' /\
          tc (OEL.LE (tc (X2p E X delay))) l' (f de')) \/ f de' = f de') as Hlst.
          auto.
        assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hacp.
          apply oel_ac with (labels_of_es E X); auto.
        rewrite <- Hldr; destruct_lin Hl; 
        apply (in_queue_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f de') n s0 s1 (f de') de e Hwfl' Hwfrfl Hwfq Hee Hde2 Hdiff Hlde' Hlde Hdiff Hlde' Hdr Hacp Httc Htot Hdf Hlst Hiq Hem).

        intros ? Hdw.
        inversion Hd as [Hb | Hq].

        rewrite Hdw in Hde; simpl in Hde.
        assert (udr (B s1) e) as Hran.
          right; exists e'; auto.
        assert (de = DW e) as Hdwe.
          rewrite Hde; auto.

        assert (forall w, writes E w -> tc (OEL.LE (tc (X2p E X delay))) (d (DW w)) (f (DW w))) as Hdwf.
          intros w Hew; generalize (X2p_df E X delay (DW w)); intro Hip; apply trc_step; apply Hil; apply trc_step; auto.
          apply Hip; auto. simpl; destruct Hew; auto.
          destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto.

        assert (writes E e) as Hwe. 
          apply Hwfb; right; exists e'; auto.
    assert (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f de') l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f de')) \/ f de' = f de') as Hfde'.
      auto.

    assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hac'.   
      unfold acyclic; auto.
      intros x Hx. 
        rewrite trans_tc in Hx.
        destruct_lin Hl; apply (Hac x Hx); auto.
          apply oel_trans with (labels_of_es E X).
            apply tc_X2p_partial_order; auto.

        rewrite <- Hdw; rewrite Hdwe in *|-*; destruct_lin Hl;    
        apply (in_buff_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f de') n s0 s1 (f de') e Hdr Hwe Hdiff Hlde' Hdiff Hlde' Httc Hac' Htot Hdwf Hfde' Hem Hran).

    rewrite Hdw in Hlde; destruct Hwfl as [[? Hfw] ?].
    rewrite Hdw in Hde; simpl in Hde; rewrite <- Hde in Hlde.
    generalize (Hfw e Hlde); intros [? [? [? Haw]]].
    assert (ran (Q s1) e) as Hudr.
      exists e'; auto.
    generalize (Hwfq e Hudr); intros [? [? [? Har]]]; rewrite Har in Haw; inversion Haw.

    assert (~ (delay E X (evt_of_devt de) (evt_of_devt de'))) as Hnd.
      apply ghb_ndelay; auto; inversion Hppo as [Hpo | Hab].
        apply ppo_in_ghb; subst; auto.
        apply ab_in_ghb; subst; auto. 
        rewrite (p2X_X2p E X delay Hwf Hv Hwfl) in Hab; auto.
    generalize (X2p_ghb E X delay de de' Hnd Hlde Hlde'); intro Hew1.
    assert (tc (OEL.LE (tc (X2p E X delay))) (f de) (f de)) as Hc.
    apply trc_ind with (f de'); auto.
      apply trc_step; apply Hil; apply trc_step; auto.
    rewrite trans_tc in Hc; auto.
    destruct_lin Hl; apply (Hac (f de)); auto.
Qed.  

Lemma in_order_implies_queue_state :
  forall E X  delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  valid_execution E X ->
  wf_delay delay ->
  wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X)  (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  bin (ran (Q s1)) r1 = true.
Proof.
intros E X  delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfrfl Hlwr1 Hm.
generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
generalize (oel_trans (X2p E X delay) (labels_of_es E X) Hpart); intro Htrans.
generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpart); intro Hac.
generalize (oel_tot (X2p E X delay) (labels_of_es E X) Hpart); intro Htot.
assert (OEL.LE (tc (X2p E X delay)) (f (DR (w1,r1))) (f (DR (w1,r1))) \/ f (DR (w1,r1)) = f (DR (w1,r1))) as Hor.
  auto.
assert (exists d : Rln Label, trans d /\ rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ 
        ~(exists l' : Label,
          tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
          tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) /\
        machine E d (labels_of_es E X)  (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed.
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split; [|split]]; auto.
    apply td_tpred; auto; intros e1 e2 H12; auto.
    intros e1 e2 H12; auto.
    apply no_interv_last; auto; intros e1 e2 H12; auto.
assert (Included Label
    (Union Label (dom (OEL.LE (tc (X2p E X delay))))
       (ran (OEL.LE (tc (X2p E X delay))))) (labels_of_es E X)) as Hudr.
  apply oel_udr; auto.
assert (~ (exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f (DR (w1, r1))) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/
  f (DR (w1, r1)) = f (DR (w1, r1))) as Hni.
  auto.
assert (forall de, labels_of_es E X (f de) -> OEL.LE (tc (X2p E X delay)) (d de) (f de)) as Hdf.
  intros de Hde; generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hi;
  apply Hi; apply trc_step; apply X2p_df; auto.
    apply label_implies_evt with X (f de); auto.
    destruct_valid Hv; split; auto.
assert (OEL.LE (tc (X2p E X delay)) (d (DR (w1,r1))) (f (DR (w1,r1)))) as Hpdf.
  apply Hdf; auto.
generalize (in_order_implies_in_queue E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DR (w1,r1))) n s0 s1 (f (DR (w1,r1))) w1 r1  Hudr Htrans Hac Htot Hwfrfl Hlwr1 Hdf Hpdf Hpdf Hor Hni Hed);
  intro Hran.
apply In_bin; auto.
Qed.

Lemma pred_pred :
  forall (p:Rln Label) l l',
  trans p ->
  p l' l ->
  pred (pred p l) l' = pred p l'.
Proof.
intros p l l' Htp Hl'l; apply ext_rel; intros x y; split; 
  intros [H1 H2]; split; auto.
    destruct H1; auto.
    destruct H2; auto.

    split; auto.
        apply Htp with l'; auto.
    split; auto.
Qed.

Lemma same_last_eq :
  forall L d p ld lp,
  L lp -> L ld ->
  rel_incl d p ->
  trans p ->
  (forall l1 l2, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  last (pred p lp) = last (pred d ld) ->
  ld = lp.
Proof.
intros L d p ld lp Hlp Hld Hi Htp Htotp Heql.
destruct (eqLabel_dec ld lp) as [|Hdiff]; auto.
assert (p ld lp \/ p lp ld) as Hor.
  apply Htotp; auto.

inversion Hor as [Hdp | Hpd]; clear Hor.
  assert (~(exists l', p (last (pred p lp)) l' /\ p l' lp)) as Hn.
    rewrite <- trans_tc with Label p; auto.
    apply no_interv_last; auto; intros x y Hxy; auto.
      rewrite <- trans_tc with Label p; auto.
  assert (exists l', p (last (pred p lp)) l' /\ p l' lp) as Hy.
    rewrite Heql; exists ld; split; auto.
      apply Hi. 
      assert (last (pred d ld) = last (pred d ld)) as Heq.
        auto.
      generalize (last_in_ran (pred d ld) (last (pred d ld)) Heq); intros [x [? ?]]; auto.    
  contradiction.

  assert (~(exists l', p (last (pred d ld)) l' /\ p l' ld)) as Hn.
    rewrite <- trans_tc with Label p; auto.
    apply no_interv_last; auto; intros x y Hxy; auto.
  assert (exists l', p (last (pred d ld)) l' /\ p l' ld) as Hy.
    rewrite <- Heql; exists lp; split; auto.
      assert (last (pred p lp) = last (pred p lp)) as Heq.
        auto.
      generalize (last_in_ran (pred p lp) (last (pred p lp)) Heq); intros [x [? ?]]; auto.
  contradiction.  
Qed.

Lemma ws_pp_p :
  forall E p pp L l e el,
  rel_incl pp p ->
  (pp l (last pp) \/ last pp = l) ->
  (el = evt_of_label l) ->
  ws (p2X E pp L) e el ->
  ws (p2X E p L) e el.
Proof.
intros E p pp L l e el Hip Hlp Hel [? [? [? [l1 [l2 [? [? H12]]]]]]];
split; [|split; [|split]]; auto.
exists l1; exists l2; split; [|split]; auto; apply Hip; auto.
Qed.

Lemma step_monot :
  forall E p pp L s1 s2 l,
  rel_incl pp p ->
  
  
  (pp l (last pp) \/ last pp = l) ->
  step E pp L (Some s1) l = s2 ->
  step E p L (Some s1) l = s2.  
Proof.
intros E p pp L s1 s2 l  Hi Hlast Hs.
case_eq l; intros de Hl; 
rewrite Hl in Hs; case_eq de.

  intros [w r] Hde; rewrite Hde in Hs; simpl in * |- *;
  unfold buff_read_step in * |- *; auto.

  intros w Hde; rewrite Hde in Hs; simpl in * |- *;
  unfold buff_write_step in * |- *; auto.

  intros [w r] Hde; rewrite Hde in Hs; simpl in * |- *;
  unfold flush_resolved_read_step in * |- *.
    destruct (bin (ran (Q s1)) r); auto.
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); auto.

    case_eq (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E pp L)) e r))); intros Hbqp; auto;
    rewrite Hbqp in Hs. Focus 2.
      assert (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E p L)) e r)) = false) as Hbq.
        case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) 
         (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); intro Hb; auto.
        generalize (bemp_axf (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E pp L)) e r) Hbqp);
          intros [e [Hbq Hrr]].
        generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r) Hb); intro Hn.
        assert (exists e : Event, ran (rel_union (B s1) (Q s1)) e /\
          rel_union (A.ppo E) (A.abc E (p2X E p L)) e r) as Hc.
          exists e; split; auto.
            inversion Hrr; [left | right]; auto.
            apply A.ab_wit_incl with (p2X E pp L); auto.
            
        contradiction.
      rewrite Hbq; auto.
      assert (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E p L)) e r)) = true) as Hbq.
        case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) 
         (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); intro Hb; auto.
        generalize (bemp_axf (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r) Hb);
          intros [e [Hbq Hrr]].
        generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E pp L)) e r) Hbqp); intro Hn.
        assert (exists e : Event, ran (rel_union (B s1) (Q s1)) e /\
          rel_union (A.ppo E) (A.abc E (p2X E pp L)) e r) as Hc.
          exists e; split; auto.
            inversion Hrr; [left | right]; auto.
            apply A.ab_wit_incl with (p2X E p L); auto.
            
             
        contradiction.
      rewrite Hbq; auto.

  intros w Hde; rewrite Hde in Hs; simpl in * |- *;
  unfold flush_write_step in * |- *.

    case_eq (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E pp L)) e w))); intros Hbqp; auto;
    rewrite Hbqp in Hs. Focus 2.
      assert (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E p L)) e w)) = false) as Hbq.
        case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) 
         (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); intro Hb; auto.
        generalize (bemp_axf (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E pp L)) e w) Hbqp);
          intros [e [Hbq Hrr]].
        generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w) Hb); intro Hn.
        assert (exists e : Event, ran (rel_union (B s1) (Q s1)) e /\
          rel_union (A.ppo E) (A.abc E (p2X E p L)) e w) as Hc.
          exists e; split; auto.
            inversion Hrr; [left | right]; auto.
            apply A.ab_wit_incl with (p2X E pp L); auto.
            
        contradiction.
      rewrite Hbq; auto.
      assert (bemp
         (rrestrict (rel_union (B s1) (Q s1))
            (fun e : Event =>
             rel_union (A.ppo E)
               (A.abc E (p2X E p L)) e w)) = true) as Hbq.
        case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) 
         (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); intro Hb; auto.
        generalize (bemp_axf (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w) Hb);
          intros [e [Hbq Hrr]].
        generalize (bemp_axt (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E pp L)) e w) Hbqp); intro Hn.
        assert (exists e : Event, ran (rel_union (B s1) (Q s1)) e /\
          rel_union (A.ppo E) (A.abc E (p2X E pp L)) e w) as Hc.
          exists e; split; auto.
            inversion Hrr; [left | right]; auto.
            apply A.ab_wit_incl with (p2X E p L); auto.
        contradiction.
      rewrite Hbq; auto.
Qed.

Lemma mach_det :
  forall E (L:set Label) n s0 s1d s1p d p ld lp,
  L ld -> L lp ->
  Included _ (Union _ (dom p) (ran p)) L ->
  rel_incl d p ->
  trans d -> trans p ->
  (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  machine E (pred d ld) L ld n (Some s0) (Some s1d) ->
  machine E (pred p lp) L lp n (Some s0) (Some s1p) ->
  ld = lp /\ s1d = s1p.
Proof.
intros E L n s0 s2d s2p d p ld lp Hld Hlp Hudrp Hi Htd Htp Htotp Hmd Hmp.
generalize s2d s2p ld lp Hld Hlp Hudrp Hmd Hmp; clear s2d s2p ld lp Hld Hlp Hudrp Hmd Hmp;
induction n; intros s2d s2p ld lp Hld Hlp Hudrp Hmd Hmp.
      inversion Hmd as [Heqd ?]; inversion Hmp as [Heqp ?]; 
      inversion Heqd; inversion Heqp; split; subst; auto.
      admit. 

    destruct Hmd as [ld' [s1d [Hld' [Hmd [Hsd ?]]]]];
    destruct Hmp as [lp' [s1p [Hlp' [Hmp [Hsp ?]]]]].
    fold (machine E (pred (pred d ld) ld') L ld' n (Some s0) (Some s1d)) in Hmd.
    fold (machine E (pred (pred p lp) lp') L lp' n (Some s0) (Some s1p)) in Hmp.
    assert (machine E (pred d ld') L ld' n (Some s0) (Some s1d)) as Hmd'.
      
      rewrite <- pred_pred with d ld ld'; auto.
      generalize (last_in_ran (pred d ld) ld' Hld'); intros [? [? ?]]; auto.
    assert (machine E (pred p lp') L lp' n (Some s0) (Some s1p)) as Hmp'.
      
      rewrite <- pred_pred with p lp lp'; auto.
      generalize (last_in_ran (pred p lp) lp' Hlp'); intros [? [? ?]]; auto.
    assert (L ld') as Hisld'.
      apply Hudrp.
        generalize (last_in_ran (pred d ld) ld' Hld'); intros [x [Hr ?]].
        generalize (Hi x ld' Hr); intros; right; exists x; auto.
    assert (L lp') as Hislp'.
      apply Hudrp.
        generalize (last_in_ran (pred p lp) lp' Hlp'); intros [x [Hr ?]].
        right; exists x; auto.
    generalize (IHn s1d s1p ld' lp' Hisld' Hislp' Hudrp Hmd' Hmp'); intros [Heql' Heqs'].

    assert (ld = lp) as Heql.
      rewrite Heql' in Hld'; rewrite <- Hld' in Hlp'.
      apply same_last_eq with L d p; auto.

    split; auto.

    rewrite Heqs' in Hsd; rewrite Heql' in Hsd.
    assert (rel_incl (pred d ld) (pred p lp)) as Hipp.
      intros x y [Hxy ?]; split.
        apply Hi; auto.
        rewrite <- Heql; apply Hi; auto.

    assert ((pred d ld) lp' (last (pred d ld)) \/ last (pred d ld) = lp') as Hllp'.
      right; subst; auto.

    generalize (step_monot E (pred p lp) (pred d ld) L s1p (Some s2d) lp' Hipp Hllp' Hsd);
    intro Hsp'; rewrite Hsp' in Hsp; inversion Hsp; auto.
Qed.

Lemma incl_and_last_is_pred :
  forall d p L l,
  Included Label (Union Label (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall d l, rel_incl d (pred p l) ->
     (forall x1 x2, x1 <> x2 -> ( p x1 l) -> ( p x2 l) ->
      d x1 x2 \/ d x2 x1)) -> 
  rel_incl d (pred p l) ->
  ~(exists l', tc p (last d) l' /\ tc p l' l) \/ last d = l ->
  d = pred p l.
Proof.
intros d p L l Hudrp Htp Hacp Htotp Htot Hi Hor; apply ext_rel; intros x y; split; intro Hxy.
  apply Hi; auto.
  destruct Hxy as [Hxy Hyl].
  inversion Hor as [Hn | Hl]; clear Hor.
    destruct (eqLabel_dec (last d) y) as [Heq | Hneq].
      assert (d x y \/ d y x) as Hord.
        apply Htot with l; auto.
          intros Heqxy; rewrite Heqxy in Hxy.
            rewrite <- trans_tc with Label p in Hxy; auto; apply (Hacp y Hxy); auto.
            apply Htp with y; auto.
      inversion Hord as [|Hdyx]; clear Hord; auto.
        generalize (Hi y x Hdyx); intros [Hyx ?].
        assert (tc p x x) as Hc.
          apply trc_ind with y; apply trc_step; auto.
        generalize (Hacp x Hc); intro Ht; inversion Ht; auto.
      assert (p (last d) y \/ p y (last d)) as Hor.
        apply Htotp; auto.
          assert (last d = last d) as Ht.
            auto.
          apply Hudrp; generalize (last_in_ran d (last d) Ht); intros [e He];
          generalize (Hi e (last d) He); intros [? ?]; right; exists e; auto.
          apply Hudrp; right; exists x; auto.
      inversion Hor as [Hldy | Hyld]; clear Hor.
        assert (exists l' : Label, tc p (last d) l' /\ tc p l' l) as Hc.
          exists y; split; apply trc_step; auto.
        contradiction.
        assert (d x y \/ d y x) as Hord.
          apply Htot with l; auto.
            intros Heqxy; rewrite Heqxy in Hxy.
              rewrite <- trans_tc with Label p in Hxy; auto; apply (Hacp y Hxy); auto.
              apply Htp with y; auto.
        inversion Hord as [|Hdyx]; clear Hord; auto.
          generalize (Hi y x Hdyx); intros [Hyx ?].
          assert (tc p x x) as Hc.
            apply trc_ind with y; apply trc_step; auto.
          generalize (Hacp x Hc); intro Ht; inversion Ht; auto.
      assert (d x y \/ d y x) as Hord.
        apply Htot with l; auto.
          intros Heqxy; rewrite Heqxy in Hxy.
            rewrite <- trans_tc with Label p in Hxy; auto; apply (Hacp y Hxy); auto.
            apply Htp with y; auto.
      inversion Hord as [|Hdyx]; clear Hord; auto.
        generalize (Hi y x Hdyx); intros [Hyx ?].
          assert (tc p x x) as Hc.
            apply trc_ind with y; apply trc_step; auto.
          generalize (Hacp x Hc); intro Ht; inversion Ht; auto.
Qed.

Lemma mns_fwd :
  forall E (p:Rln Label) L n s0 s1 l1 l2,
  Included Label (Union Label (dom p) (ran p)) L ->
  trans p -> acyclic p ->
  (forall l1 l2 : Label, l1 <> l2 -> L l1 -> L l2 -> p l1 l2 \/ p l2 l1) ->
  (forall d l, rel_incl d (pred p l) ->
     (forall x1 x2, x1 <> x2 -> ( p x1 l) -> ( p x2 l) ->
      d x1 x2 \/ d x2 x1)) -> 
  p l1 l2 -> 
  (exists d, trans d /\ rel_incl d (pred p l1) /\ 
      (~(exists l' : Label, tc p (last d) l' /\ tc p l' l1) \/ last d = l1) /\
     machine E d L l1 n (Some s0) (Some s1)) ->
  (exists d, rel_incl d (pred p l2) /\ 
      (~(exists l' : Label, tc p (last d) l' /\ tc p l' l2) \/ last d = l2) /\
     machine E d L l2 n (Some s0) (Some s1)).
Proof.
intros E p L n s0 s1 l1 l2 Hudrp Htp Hacp Htotp Htpred H12 [d [Htd [Hi [Hlast Hm]]]].

  generalize l1 l2 d s0 s1 H12 Htd Hi Hlast Hm; clear l1 l2 d s0 s1 H12 Htd Hi Hlast Hm;
  induction n; intros l1 l2 d s0 s1 H12 Htd Hi Hlast Hm.

    inversion Hm.
      admit. 
      
    destruct Hm as [l [s [Hl [Hm [Hs Hwf]]]]];
    fold (machine E (pred d l) L l n (Some s0) (Some s)) in Hm.
    assert (p l l2) as Hl2.
      generalize (last_in_ran d l Hl); intros [x Hdxl];
      generalize (Hi x l Hdxl); intros [? Hpll1];
      apply Htp with l1; auto.
    assert (rel_incl (pred d l) (pred p l)) as Hip.
      intros x y [Hxy Hyl]; 
      generalize (Hi x y Hxy); intros [? ?]; 
      generalize (Hi y l Hyl); intros [? ?]; split; auto.
    assert ((~ (exists l' : Label, tc p (last (pred d l)) l' /\ tc p l' l) \/ last (pred d l) = l)) as Hlp.
      left; apply no_interv_last; auto.
        intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto.
    assert (trans (pred d l)) as Htpd.
      apply td_tpred; auto.
    generalize (IHn l l2 (pred d l) s0 s Hl2 Htpd Hip Hlp Hm); intros [d' [Hid' [Hld' Hm']]].

      exists d; split; [|split]; auto.
        intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; split; auto.
          apply Htp with l1; auto.
        assert (d' = pred p l2) as Hd'.
          apply incl_and_last_is_pred with L; auto.
        rewrite Hd' in Hm'.
        assert (rel_incl d p) as Hidp.
          intros x y Hxy; generalize (Hi x y Hxy); intros [? ?]; auto.
        assert (L l) as Hll.
          apply Hudrp.
            generalize (last_in_ran d l Hl); intros [x Hr]; right; exists x; apply Hidp; auto.
        assert (L l2) as Hll2.
          apply Hudrp; right; exists l; auto.
      
        generalize (mach_det E L n s0 s s d p l l2 Hll Hll2 Hudrp Hidp Htd Htp Htotp Hm Hm'); intros [Heqll2 ?].
        rewrite Hl; rewrite Heqll2; auto.
          exists l; exists s; split; [|split; [|split]]; auto.
Qed.

Lemma oel_tpred :
  forall E X delay,
  (forall d l, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) l) ->
     (forall x1 x2, x1 <> x2 -> 
      ((OEL.LE (tc (X2p E X delay))) x1 l) -> ((OEL.LE (tc (X2p E X delay))) x2 l) ->
      d x1 x2 \/ d x2 x1)).
Proof.
Admitted.  

Lemma contrad_dp :
  forall (E : Event_struct) (X : Execution_witness) delay (n : nat) (s0 s1 : State) (w1 r1 : Event),
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_mem E (M s1) ->
  wf_queue E (Q s1) ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  rf X w1 r1 ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  bemp (rrestrict (Q s1) (fun e => dp.dp E e w1)) = false -> False.
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfq Hwfl Hwfrfl Hlbr Hrf Hm Hdp.
  generalize Hwfl; intros Hwfl'.
  generalize (bemp_axf (Q s1) (fun e => dp.dp E e w1) Hdp);
  clear Hdp; intros [e [Hi Hdp]].
    assert (rel_incl (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1))))) as Ht.
      intros x y Hxy; auto.
    
      destruct Hi as [e' Hd].
        assert (events E e) as Hee.
          apply ABasic.po_iico_domain_in_events with w1; auto.
            apply dp.dp_valid; auto.

        generalize (wfl_e2 E (labels_of_es E X) e Hwfl Hwfrfl Hee); intros [de [Hde Hlde]].
        assert (f (DR (w1,r1)) <> f de) as Hdiff.
          intros Heqf; inversion Heqf as [Heq]; rewrite <- Heq in Hde; simpl in Hde.
            destruct_valid Hv; apply (Hth r1); apply trc_ind with w1; apply trc_step; [right; rewrite Hde in Hdp | left]; auto.
        assert (labels_of_es E X (f (DR (w1,r1)))) as Hlw1.
          auto.
        assert (Included _ (labels_of_es E X) (labels_of_es E X)) as Htl.
          intros l Hl; auto.
        assert (partial_order (tc (X2p E X delay)) (labels_of_es E X)) as Hpl.
          apply tc_X2p_partial_order; auto.
        generalize (OEL.OE (labels_of_es E X) (labels_of_es E X) (tc (X2p E X delay)) Htl Hpl);
          clear Htl; intros [Hil Hl].
        assert (trans (OEL.LE (tc (X2p E X delay)))) as Httc.
          destruct_lin Hl; intros x y z Hxy Hyz; apply Htrans with y; auto.
        assert (forall de, labels_of_es E X (f de) -> tc (OEL.LE (tc (X2p E X delay))) (d de) (f de)) as Hdf.
          intros de'' Hde''; generalize (X2p_df E X delay de''); intro Hip; apply trc_step; apply Hil; apply trc_step; auto.
          apply Hip; auto. apply label_implies_evt with X (f de''); auto.
            destruct_valid Hv; split; auto.
        assert (exists r, rel_incl r (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\
                  (~(exists l' : Label,
                  tc (OEL.LE (tc (X2p E X delay))) (last r) l' /\
                  tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/ last r = f (DR (w1,r1))) /\
                  machine E r (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hem.
          exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split]; auto.
            left; apply no_interv_last; auto; intros x y Hxy; auto.

  assert (tc (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1))) (f de)) as Hfw1fe.

        assert (ran (Q s1) e) as Hiq.
          exists e'; auto.
        destruct Hwfl as [Hwflw ?].
        generalize (Hwfq e Hiq); intro Hre.
        assert (evt_of_devt de = e) as Hdesym.
          apply sym_eq; auto.
        assert (~(exists l' : Label, 
          tc (OEL.LE (tc (X2p E X delay))) (f (DR (w1, r1))) l' /\
          tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/
          f (DR (w1, r1)) = f (DR (w1, r1))) as Hlst.
          auto.

        assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hacp.
          apply oel_ac with (labels_of_es E X); auto.

        destruct_lin Hl; 
        apply (in_queue_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DR (w1,r1))) n s0 s1 (f (DR (w1,r1))) de e Hwfl' Hwfrfl Hwfq Hee Hdesym Hdiff Hlw1 Hlde Hdiff Hlw1 Hdr Hacp Httc Htot Hdf Hlst Hiq Hem).

  assert (tc (OEL.LE (tc (X2p E X delay))) (f de) (f (DR (w1,r1)))) as Hfefw1.
  
    assert (ran (Q s1) r1) as Hiq.
      apply bin_In; apply in_order_implies_queue_state with E X  delay n s0 w1; auto.

        assert (rfmaps_well_formed E (events E) (rf X)) as Hrfwf.
          destruct_valid Hv; split; auto.

    assert (exists d, trans d /\ rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ 
                      (~(exists l', tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
                         tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/ last d = f (DR (w1,r1)))/\
                      machine E d (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hedr1.
      exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split; [|split]]; auto.
        apply td_tpred; auto.
        left; apply no_interv_last; auto; intros x y Hxy; auto.
    assert (trans (OEL.LE (tc (X2p E X delay)))) as Htr.
      apply oel_trans with (labels_of_es E X) ; auto.  
    rewrite trans_tc in Hfw1fe; auto.
    generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpl); intro Hacp.
    generalize (oel_udr (X2p E X delay) (labels_of_es E X) Hpl); intro Hudrp.
    generalize (oel_tot (X2p E X delay) (labels_of_es E X) Hpl); intro Htotp.
    generalize (oel_tpred E X delay); intro Htpred.
    generalize (mns_fwd E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) n s0 s1 (f (DR (w1,r1))) (f de) Hudrp Htr Hacp Htotp Htpred Hfw1fe Hedr1); intro Hedde.
    destruct Hwfl as [Hwflw ?].

    assert (events E r1) as Her1.
      generalize (Hwfq r1 Hiq); intros [? ?]; auto.
    assert (evt_of_devt (DR (w1,r1)) = r1) as Hdr1.
      simpl; auto.
    assert (f de <> f (DR (w1,r1))) as Hdiff'.
      intro Heq; apply Hdiff; auto.
    assert (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f de) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f de)) \/ f de = f de) as Hlst.
      auto.

    destruct_lin Hl; 
    apply (in_queue_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f de) n s0 s1 (f de) (DR (w1,r1)) r1 Hwfl' Hwfrfl Hwfq Her1 Hdr1 Hdiff' Hlde Hlbr Hdiff' Hlde Hdr Hacp Httc Htot Hdf Hlst Hiq Hedde); auto.

    assert (tc (OEL.LE (tc (X2p E X delay))) (f de) (f de)) as Hc.
    apply trc_ind with (f (DR (w1,r1))); auto.
    rewrite trans_tc in Hc; auto.
    destruct_lin Hl; apply (Hac (f de)); auto.
Qed.

Definition loc_of_label lb :=
  match lb with
  | d de | f de => loc (evt_of_devt de) end.

Lemma flush_read_same_mem :
  forall E p L s1 s2 w l w' r',
  flush_resolved_read_step E p L s1 w' r' = Some s2 ->
  (M s1) l = w ->
  (M s2) l = w.
Proof.
intros E p L s1 s2 w l w' r' Hs Hms0.

  unfold flush_resolved_read_step in Hs;
  destruct (bin (ran (Q s1)) r'); [|inversion Hs];
  destruct (bemp (rrestrict (Q s1) (fun e => (dp.dp E) e w'))); [|inversion Hs];
  destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r'))); [|inversion Hs].

  destruct (eqEv_dec w' (M s1 (loc r'))) as [Him | Hnim].
    destruct (bemp (rrestrict (B s1) (fun w => po_iico E w r' /\ loc r' = loc w))); [|inversion Hs].
      inversion Hs as [Heq]; auto.
  destruct (bin (Intersection _ (ran (B s1)) (rrB r'))); [inversion Hs|]. 
  destruct (bin (fun w : Event => (B s1 w (waft E (B s1) r')) /\ loc r' = loc w) w'); [|inversion Hs].
    destruct (bin (fun w : Event => (w = wbef E (B s1) r' \/ B s1 (wbef E (B s1) r') w) /\ loc r' = loc w) w'); [|inversion Hs].
      inversion Hs as [Heq]; auto.
Qed.

Lemma pick_last_loc :
  forall B w,
  (pick_last (rrestrict B (fun e => loc e = loc w))) <> dB ->
  loc (pick_last (rrestrict B (fun e => loc e = loc w))) = loc w.
Proof.
intros B w HndB.

  generalize (EPick.pick_in (fun w0 : Event =>
          ~
          (exists w1 : Event,
             rrestrict B (fun e : Event => loc e = loc w) dB w1) /\ w0 = dB \/
          (exists w1 : Event,
             rrestrict B (fun e : Event => loc e = loc w) dB w1) /\
          rrestrict B (fun e : Event => loc e = loc w) dB w0 /\
          ~
          (exists w' : Event,
             rrestrict B (fun e : Event => loc e = loc w) w' w0)));

  intros Hinl; fold (pick_last (rrestrict B (fun e => loc e = loc w))) in Hinl.
  inversion Hinl as [[? ?] |[? [[? [? ?]] ?]]]; [contradiction|]; auto.
Qed.

Lemma upd_buff_id :
  forall (B:Buffer) e,
  B dB e ->
  upd_buff B e = B.
Proof.
intros B e Hi; apply ext_rel; intros x y; split; intro Hxy.
  unfold upd_buff in Hxy; auto.
  case_eq (bin (udr B) e); intro Hib.
    rewrite Hib in Hxy; auto.
    assert (bin (udr B) e = true) as Hib'.
      apply In_bin; right; exists dB; auto.
    rewrite Hib in Hib'; inversion Hib'.

  unfold upd_buff; 
  destruct (bin (udr B) e); auto.
  left; auto.
Qed.

Lemma incl_lasts_implies_eq_pick :
  forall B (B':Buffer) a,
  (forall w1 w2, 
   w1 <> w2 ->
    ran B' w1 ->
    ran B' w2 ->
   loc w1 = loc w2 -> B' w1 w2 \/ B' w2 w1) ->
   rel_incl B B' -> 
  ~ (exists e', rrestrict B' (fun e => loc e = a) e' (pick_last (rrestrict B (fun e => loc e = a)))) -> 
  ~ (pick_last (rrestrict B (fun e => loc e = a)) = dB) -> 
  ~ (pick_last (rrestrict B' (fun e => loc e = a)) = dB) ->
  pick_last (rrestrict B (fun e => loc e = a)) = 
  pick_last (rrestrict B' (fun e => loc e = a)).
Proof.
intros B'' B' a Hwfb Hincl Hfwd Hnemp Hnemp'.
destruct (eqEv_dec (pick_last (rrestrict B'' (fun e => loc e = a))) 
                   (pick_last (rrestrict B' (fun e => loc e = a)))) as [|Hneq]; auto.

  generalize (EPick.pick_in (fun e => (~(exists w, rrestrict B'' (fun e => loc e = a) dB w) /\ e = dB) \/ 
  ((exists w, rrestrict B'' (fun e => loc e = a) dB w) /\ rrestrict B'' (fun e => loc e = a) dB e /\
    ~ (exists e', rrestrict B'' (fun e => loc e = a) e' e)))); 
  intros Hinl; fold (pick_last (rrestrict B'' (fun e => loc e = a))) in Hinl;
  inversion Hinl as [[? Hd] | [? [Hil Hl]]]; clear Hinl; 
  [contradiction | auto].

  assert (rrestrict B' (fun e => loc e = a) dB (pick_last (rrestrict B'' (fun e => loc e = a)))) as Hi.
    destruct Hil as [Hil [? ?]]; split; auto.

  generalize (EPick.pick_in (fun e => (~(exists w, rrestrict B' (fun e => loc e = a) dB w) /\ e = dB) \/ 
  ((exists w, rrestrict B' (fun e => loc e = a) dB w) /\ rrestrict B' (fun e => loc e = a) dB e /\
    ~ (exists e', rrestrict B' (fun e => loc e = a) e' e)))); 
  intros Hinl; fold (pick_last B') in Hinl; 
  inversion Hinl as [[? Hd] | [? [Hil' Hl']]]; clear Hinl; [contradiction|].

  assert (loc (pick_last (rrestrict B'' (fun e => loc e = a))) = 
    loc (pick_last (rrestrict B' (fun e => loc e = a)))) as Heql.
    destruct Hi as [? [? Hl'']]; auto; destruct Hil' as [? [? Hl2']]. 
      rewrite Hl''; fold (pick_last (rrestrict B' (fun e : Event => loc e = a))) in Hl2'; rewrite Hl2'; auto.
  destruct Hi as [Hi [? Hlb]]; destruct Hil' as [Hil' [? Hlb']].
  assert (ran B' (pick_last (rrestrict B'' (fun e : Event => loc e = a)))) as Hi'.
    exists dB; auto.
  assert (ran B' (pick_last (rrestrict B' (fun e : Event => loc e = a)))) as Hil''.
    exists dB; auto.
  generalize (Hwfb (pick_last (rrestrict B'' (fun e => loc e = a)))
    (pick_last (rrestrict B' (fun e => loc e = a))) Hneq Hi' Hil'' Heql); intro Hor; inversion Hor as [Hbb' | Hb'b].
  
  assert (exists e', rrestrict B' (fun e => loc e = a) e' (pick_last (rrestrict B' (fun e =>  loc e = a)))) as Hcl'.
    exists (pick_last (rrestrict B'' (fun e => loc e = a))); split; auto.
  contradiction.

  assert (exists e', rrestrict B' (fun e => loc e = a) e' (pick_last (rrestrict B'' (fun e =>  loc e = a)))) as Hcfwd.
    exists (pick_last (rrestrict B' (fun e => loc e = a) )); split; auto.
  contradiction.   
Qed.

Lemma upd_wfb :
  forall (B:Buffer) w, 
  (forall w1 w2, 
   w1 <> w2 ->
   ran B w1 ->
   ran B w2 ->
   loc w1 = loc w2 -> B w1 w2 \/ B w2 w1) ->
  (forall w1 w2, 
   w1 <> w2 ->
    ran (upd_buff B w) w1 ->
    ran (upd_buff B w) w2 ->
   loc w1 = loc w2 -> upd_buff B w w1 w2 \/ upd_buff B w w2 w1).
Proof.
intros B' w Hwfb w1 w2 Hd H1 H2 Hl.

  unfold upd_buff in * |- *; case_eq (bin (udr B') w); intro Hib; 
  rewrite Hib in * |-*; auto.
    destruct H1 as [x1 H1]; destruct H2 as [x2 H2].
    inversion H1 as [Hb1 | Heq1]; clear H1; 
    inversion H2 as [Hb2 | Heq2]; clear H2.
      assert (ran B' w1) as Hr1.
        exists x1; auto.
      assert (ran B' w2) as Hr2.
        exists x2; auto.
      generalize (Hwfb w1 w2 Hd Hr1 Hr2 Hl); intro Hor; inversion Hor as [H12 | H21]; clear Hor; [left | right]; left; auto.
      destruct Heq2 as [? Heq2]; left; right; split; auto;  subst; auto; split; auto; right; exists x1; auto.
      destruct Heq1 as [? Heq1]; right; right; split; auto;  subst; auto; split; auto; right; exists x2; auto.
      destruct Heq1 as [? Heq1]; destruct Heq2 as [? Heq2]; rewrite <- Heq2 in Heq1; contradiction.   
Qed.

Lemma last_not_dummy :
  forall (B:Buffer),
  pick_last B <> dB ->
  exists w, pick_last B = w.
Proof.
intros B HndB.
generalize (EPick.pick_in (fun e => (~(exists w, B dB w) /\ e = dB) \/ 
  ((exists w, B dB w) /\ B dB e /\ ~ (exists e', B e' e)))); 
  intros Hinl; fold (pick_last B) in Hinl.
inversion Hinl as [[? ?] |]; [contradiction|].
exists (pick_last B); auto.
Qed.
 
Lemma dummy_not_last :
  forall (B:Buffer) w,
  acyclic B ->
  B dB w ->
  pick_last B <> dB.
Proof.
intros B w Hac Hbw Heq.
generalize (EPick.pick_in (fun e => (~(exists w, B dB w) /\ e = dB) \/ 
  ((exists w, B dB w) /\ B dB e /\ ~ (exists e', B e' e)))); 
  intros Hinl; fold (pick_last B) in Hinl.
inversion Hinl as [[? ?] | [? [Hib ?]]].

  assert (exists w, B dB w) as Hc.
    exists w; auto.
  contradiction.

  rewrite Heq in Hib; apply Hac with dB; apply trc_step; auto.
Qed.
  
Lemma tc_updb :
  forall B w a x y,
  w <> dB ->
  trans B ->
  tc (rrestrict (upd_buff B w) (fun e => loc e = a)) x y ->
  tc (rrestrict B (fun e => loc e = a)) x y \/ ((x = dB \/ udr B x) /\ (y <> dB /\ ~(udr B y))).
Proof.
intros B w a x y HndB Htr Hxy.
induction Hxy as [x y [Hs ?] |]. 
  
  unfold upd_buff in Hs; case_eq (bin (udr B) w); 
    intro Hi; rewrite Hi in Hs.
    left; apply trc_step; split; auto.
    
      inversion Hs as [Hb | [[Hib ?] Heq]]; auto.
      left; apply trc_step; split; auto.
      right; split; auto. 
        subst; split; auto. 
          intro Hibw; generalize (In_bin (udr B) w Hibw); intro Ht;
          rewrite Ht in Hi; inversion Hi.

  inversion IHHxy1 as [Hxz | [Hx [? Hnz]]];
  inversion IHHxy2 as [Hzy | [Hz Hny]].

    left; apply trc_ind with z; auto.
    right; split; auto.
      rewrite trans_tc in Hxz; auto. 
        destruct Hxz as [? ?]; auto; right; left; exists z; auto.
        intros e1 e2 e3 [H12 [? ?]] [H23 [? ?]]; split; [|split]; auto.
          apply Htr with e2; auto.

    rewrite trans_tc in Hzy; auto.
    
      assert (udr B z) as Hc.
        destruct Hzy as [? ?]; left; exists y; auto.
      contradiction.
        intros e1 e2 e3 [H12 [? ?]] [H23 [? ?]]; split; [|split]; auto.
          apply Htr with e2; auto.      

    right; split; auto.
Qed.

Lemma acb_persistent :
  forall B w,
  w <> dB ->
  trans B ->
  acyclic (rrestrict B (fun e => loc e = loc w)) ->
  acyclic (rrestrict (upd_buff B w) (fun e => loc e = loc w)).
Proof.
intros B w HndB Htr Hac x Htc.
generalize (tc_updb B w (loc w) x x HndB Htr Htc); intro Hor.
inversion Hor as [Hc | [Hx [HndBx Hnx]]].

  apply Hac with x; auto.

  inversion Hx; contradiction.
Qed.

Lemma nemp_upd_buff :
  forall E B w',
  w' <> dB ->
  wf_buff E B ->
  pick_last (rrestrict B (fun e => (loc e) = loc w')) <> dB ->
  pick_last (rrestrict (upd_buff B w') (fun e => (loc e) = loc w')) <> dB.
Proof.
intros E B' w' HndB Hwfb Hnemp.
generalize (last_not_dummy (rrestrict B' (fun e => (loc e) = loc w')) Hnemp);
  intros [e He].
generalize (EPick.pick_in (fun e => (~(exists w, rrestrict B' (fun e => loc e = loc w') dB w) /\ e = dB) 
  \/ ((exists w, rrestrict B' (fun e => loc e = loc w') dB w) /\ rrestrict B' (fun e => (loc e) = loc w') dB e /\ 
    ~ (exists e', rrestrict B' (fun e => (loc e) = loc w') e' e)))); 
  intros Hinl; fold (pick_last (rrestrict B' (fun e => (loc e) = loc w'))) in Hinl;
  inversion Hinl as [[? Hd] | [? [Hil' Hl']]]; clear Hinl; [contradiction |].

rewrite He in Hil'; destruct Hil' as [? [? ?]];
apply dummy_not_last with e; auto.
  apply acb_persistent; auto; destruct Hwfb as [? [? [? [? ?]]]]; auto.

  split; auto. 
  unfold upd_buff; destruct (bin (udr B') w'); auto.
  left; auto.
Qed.

Lemma upd_buff_last :
  forall E (B:Buffer) w ,
   w <> dB ->
  wf_buff E B -> 
  pick_last (rrestrict B (fun e => loc e = loc w)) <> dB ->
  ~ B dB  w ->
  pick_last (rrestrict B (fun e => loc e = loc w)) = 
  pick_last (rrestrict (upd_buff B w) (fun e => loc e = loc w)).
Proof.
intros E B' w  HndB Hwfb Hnemp Hnin.
apply incl_lasts_implies_eq_pick; auto.
  apply upd_wfb; destruct Hwfb as [? [? ?]]; auto. 

  intros e1 e2; unfold upd_buff; destruct (bin (udr B') w); auto; left; auto.
  intros [e He].
  generalize (EPick.pick_in (fun w0 : Event =>
           ~
           (exists w1 : Event,
              rrestrict B' (fun e : Event => loc e = loc w) dB w1) /\ 
           w0 = dB \/
           (exists w1 : Event,
              rrestrict B' (fun e : Event => loc e = loc w) dB w1) /\
           rrestrict B' (fun e : Event => loc e = loc w) dB w0 /\
           ~
           (exists w' : Event,
              rrestrict B' (fun e : Event => loc e = loc w) w' w0)));
  intros Hinl; fold (pick_last (rrestrict B' (fun e => loc e = loc w))) in Hinl.
    inversion Hinl as [[? Hd] | [? [Hil' Hl']]]; clear Hinl; [contradiction |].

    destruct He as [He ?]; unfold upd_buff in He.
        assert (exists e', (rrestrict B' (fun e => loc e = loc w)) e' (pick_last (rrestrict B' (fun e =>  loc e = (loc w))))) as Hc.
          exists e; split; auto.
          destruct (bin (udr B') w); auto.
          inversion He as [Hb | [? Heq]]; auto. 
            rewrite <- Heq in Hnin; destruct Hil'; contradiction.
        contradiction.

  apply nemp_upd_buff with E; auto.
Qed.

Lemma eq_del_lasts :
  forall E B a w',
  events E w' ->
  a <> loc w' ->
  (fun e => (~(exists w, rrestrict B (fun e => loc e = a) dB w) /\ e = dB) \/ 
      ((exists w, rrestrict B (fun e => loc e = a) dB w) /\ rrestrict B (fun e => loc e = a) dB e /\ 
     ~ (exists e', rrestrict B (fun e => loc e = a) e' e))) =
  (fun e => (~(exists w, rrestrict (del_buff B w') (fun e => loc e = a) dB w) /\ e = dB) \/ 
      ((exists w, rrestrict (del_buff B w') (fun e => loc e = a) dB w) /\ rrestrict (del_buff B w') (fun e => loc e = a) dB e /\ 
     ~ (exists e', rrestrict (del_buff B w') (fun e => loc e = a) e' e))).
Proof.
intros E B' a w' Hew' Hor; apply Extensionality_Ensembles; unfold Same_set; split; intros e Hi.

  inversion Hi as [[Hemp Heq] | [Hnemp [[Hin [HdB He]] Hn]]]; clear Hi. 
  left; split; auto.
    intros [w [[Hib ?] ?]].
    assert (exists w : Event, rrestrict B' (fun e : Event => loc e = a) dB w) as Hc.
      exists w; split; auto.
    contradiction.
    
  right; split; auto.
    destruct Hnemp as [w [Hib [? Hl]]]; exists w; split; auto.
      split; auto. 
        split. intro Heq. 
        assert (w' = dB) as Heq'.
          auto.
        generalize Heq'; 
        apply (dB_not_in_E E w' Hew'); auto.
        unfold Ensembles.In in Hl; rewrite <- Hl in Hor;
        intro Heq; rewrite Heq in Hor; apply Hor; auto.

    split.

    split; subst; auto.
      split; auto; destruct He as [Hp Hl]; subst; auto.
        split. 
          generalize (dB_not_in_E E w' Hew'); auto.
          intro Heq; rewrite Heq in Hor; apply Hor; auto.
    intros [e' He']; assert (exists e', rrestrict B' (fun e => loc e = a) e' e) as Hex.
      exists e'; destruct He' as [Hirr ?]; split; auto.
      inversion Hirr as [?]; auto.
    contradiction.

  inversion Hi as [[Hemp Heq] | [Hnemp [[Hin [HdB He]] Hn]]]; clear Hi. 
  left; split; auto.
    intros [w [Hib [? Hl]]].
    apply Hemp; exists w; split; auto.
      split; auto. 
        split. intro Heqw'. 
        assert (w' = dB) as Heqw''.
          auto.
        generalize Heqw''; 
        apply (dB_not_in_E E w' Hew'); auto.
        
        unfold Ensembles.In in Hl; intro Heqw; rewrite <- Hl in Hor; 
        rewrite Heqw in Hor; apply Hor; auto.

  right; split; auto.
    destruct Hnemp as [w [[Hib ?] ?]]; exists w; split; auto.
    split.

    split; subst; auto.
      inversion Hin as [?]; auto.
    intros [e' He']; assert (exists e', rrestrict (del_buff B' w') (fun e =>  loc e = a) e' e) as Hex.
      exists e'; destruct He' as [Hirr [Hle' Hle]]; split; auto. 
      split; auto; subst; auto.
        split. intro Heq; rewrite Heq in Hle'; auto.
        unfold Ensembles.In in Hle; rewrite <- Hle in Hor; intro Heq; 
        rewrite Heq in Hor; apply Hor; auto.
          
    contradiction.
Qed.

Lemma eq_lasts_implies_eq_pick :
  forall B B',
  (fun e => (~(exists w, B dB w) /\ e = dB) \/ ((exists w, B dB w) /\ B dB e /\ ~ (exists e', B e' e))) =
  (fun e => (~(exists w, B' dB w) /\ e = dB) \/ ((exists w, B' dB w) /\ B' dB e /\ ~ (exists e', B' e' e))) ->
  pick_last B = pick_last B'.
Proof.
intros B'' B' Heq; unfold pick_last. 
rewrite <- Heq; auto.
Qed. 

Lemma not_same_loc_implies_same_last :
  forall E B w w',
  events E w' ->
  loc w <> loc w' ->
  pick_last (rrestrict B (fun e => loc e = loc w)) = 
  pick_last ((rrestrict (del_buff B w')) (fun e => loc e = loc w)).
Proof.
intros E B' w w' Hew' Hnl; apply eq_lasts_implies_eq_pick.
apply eq_del_lasts with E; auto.
Qed.

Lemma eq_upd_lasts :
  forall E B' c we,
  events E c ->
  loc c <> loc we ->
(fun e : Event =>
 ~ (exists w : Event, rrestrict B' (fun e0 : Event => loc e0 = loc we) dB w) /\
 e = dB \/
 (exists w : Event, rrestrict B' (fun e0 : Event => loc e0 = loc we) dB w) /\
 rrestrict B' (fun e0 : Event => loc e0 = loc we) dB e /\
 ~ (exists e' : Event, rrestrict B' (fun e0 : Event => loc e0 = loc we) e' e)) =
(fun e : Event =>
 ~
 (exists w : Event,
    rrestrict (upd_buff B' c) (fun e0 : Event => loc e0 = loc we) dB w) /\
 e = dB \/
 (exists w : Event,
    rrestrict (upd_buff B' c) (fun e0 : Event => loc e0 = loc we) dB w) /\
 rrestrict (upd_buff B' c) (fun e0 : Event => loc e0 = loc we) dB e /\
 ~
 (exists e' : Event,
    rrestrict (upd_buff B' c) (fun e0 : Event => loc e0 = loc we) e' e)).
Proof.
intros E B' c we Hec Hor; apply Extensionality_Ensembles; unfold Same_set; split; intros e Hi.

  inversion Hi as [[Hemp Heq] | [Hnemp [[Hin [HdB He]] Hn]]]; clear Hi. 
  left; split; auto.
    intros [w [Hib [? Heql]]].
    assert (exists w : Event, rrestrict B' (fun e : Event => loc e = loc we) dB w) as Hc.
      exists w; split; auto.
      unfold upd_buff in Hib.
        case_eq (bin (udr B') c); intro Hi; rewrite Hi in Hib; auto.
        inversion Hib as [? | [? Heqw]]; auto.
          unfold Ensembles.In in Heql; rewrite Heqw in Heql; contradiction.
          
    contradiction.
    
  right; split; auto.
    destruct Hnemp as [w [Hib [? Hl]]]; exists w; split; auto.
      unfold upd_buff; destruct (bin (udr B') c); auto; left; auto.
      split; auto. 
        split; auto. 
          unfold upd_buff; destruct (bin (udr B') c); auto; left; auto.


    intros [e' He']; assert (exists e', rrestrict B' (fun e => loc e = loc we) e' e) as Hex.
      exists e'; destruct He' as [Hirr [? Heql]]; split; auto.
      unfold upd_buff in Hirr.
        case_eq (bin (udr B') c); intro Hi; rewrite Hi in Hirr; auto.
        inversion Hirr as [? | [? Heqw]]; auto.
          unfold Ensembles.In in Heql; rewrite Heqw in Heql; contradiction.

    contradiction.

  inversion Hi as [[Hemp Heq] | [Hnemp [[Hin [HdB He]] Hn]]]; clear Hi. 
  left; split; auto.
    intros [w [Hib [? Hl]]].
    apply Hemp; exists w; split; auto.
      unfold upd_buff in Hib.
        unfold upd_buff; case_eq (bin (udr B') c); intro Hi; auto; left; auto.

  right; split; auto.
    destruct Hnemp as [w [Hib [? Hl]]]; exists w; split; auto.
      unfold upd_buff in Hib; destruct (bin (udr B') c); auto.
        inversion Hib as [? | [? Heqw]]; auto.
          unfold Ensembles.In in Hl; rewrite Heqw in Hl; contradiction.

      split; auto. 
        split; auto.
      unfold upd_buff in Hin; destruct (bin (udr B') c); auto.
        inversion Hin as [? | [? Heqw]]; auto.
          unfold Ensembles.In in He; rewrite Heqw in He; contradiction.

    intros [e' He']; assert (exists e', rrestrict (upd_buff B' c) (fun e =>  loc e = loc we) e' e) as Hex.
      exists e'; destruct He' as [Hirr [Hle' Hle]]; split; auto. 
      unfold upd_buff; destruct (bin (udr B') c); auto; left; auto.
    contradiction.
Qed.

Lemma not_same_loc_implies_same_last' :
  forall E B c we,
  events E c ->
  loc c <> loc we ->
  pick_last (rrestrict B (fun e => loc e = loc we)) = 
  pick_last ((rrestrict (upd_buff B c)) (fun e => loc e = loc we)).
Proof.
intros E B' c we Hec Hnl; apply eq_lasts_implies_eq_pick.
apply eq_upd_lasts with E; auto.
Qed.

Lemma flush_read_same_last :
  forall E p L s0 s1 we w r,
  pick_last (rrestrict (B s0) (fun e => loc e = loc we)) = we ->
  flush_resolved_read_step E p L s0 w r = Some s1 ->
  pick_last (rrestrict (B s1) (fun e => loc e = loc we)) = we.    
Proof.
intros E p L s0 s1 we w r Hs0 Hs.
  unfold flush_resolved_read_step in Hs;
  destruct (bin (ran (Q s0)) r); [|inversion Hs];
  destruct (bemp (rrestrict (Q s0) (fun e : Event => dp.dp E e w))); [|inversion Hs];
  destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs].
  destruct (eqEv_dec w (M s0 (loc r))) as [Him | Hnim].
    destruct (bemp (rrestrict (B s0) (fun w => po_iico E w r /\ loc r = loc w))); [|inversion Hs].
      inversion Hs as [Heq]; auto; rewrite <- Heq in Hs0; simpl in Hs0; contradiction.
  destruct (bin (Intersection _ (ran (B s0)) (rrB r))); [inversion Hs|].
  destruct (bin (fun w : Event => (B s0 w (waft E  (B s0) r)) /\ loc r = loc w) w); [|inversion Hs].
    destruct (bin (fun w : Event => (w = wbef E (B s0) r \/ B s0 (wbef E (B s0) r) w) /\ loc r = loc w) w); [|inversion Hs].
      inversion Hs as [Heq]; auto; rewrite <- Heq in Hs0; simpl in Hs0; contradiction.          
Qed.

Lemma not_flush_implies_same_last :
  forall E p X s0 s1 l we,
  well_formed_event_structure E ->
  valid_execution E X ->
  labels_of_es E X l ->
  wf_buff E  (B s0) ->
  pick_last (rrestrict (B s0) (fun e => loc e = loc we)) = we ->
  we <> dB ->
  l <> f (DW we) ->
  step E p (labels_of_es E X) (Some s0) l = Some s1 ->
  pick_last (rrestrict (B s1) (fun e => loc e = loc we)) = we.
Proof.
intros E p X s0 s1 l we Hwf Hv Hlb Hwfb Hs0 Hnemp Hl Hs.
case_eq l; intros de Hde; rewrite Hde in Hs;
case_eq de; intros c Hc; rewrite Hc in Hs; simpl in Hs.


  
  
  destruct c as [w r]; unfold buff_read_step in Hs;
  inversion Hs; auto.

  
  unfold buff_write_step in Hs; inversion Hs; simpl.
    rewrite <- Hs0; apply sym_eq. 
    rewrite pick_last_loc; auto.
    generalize (excluded_middle ((B s0) dB c)); intro Hor; inversion Hor as [Hi | Hni].
      rewrite upd_buff_id; auto.
      destruct (eqLoc_dec (loc c) (loc we)) as [Heql | Hneql]; [rewrite <- Heql|].
      apply upd_buff_last with E; auto.
        rewrite Hc in Hde; assert (l = d (DW c) \/ l = f (DW c)) as Horl; auto.
        assert (rfmaps_well_formed E (events E) (rf X)) as Hrfwf.
          destruct_valid Hv; split; auto.
        generalize (label_implies_evt E X l (DW c) Hwf Hrfwf Hlb Horl); intros He Heq;
        simpl in He; apply (dB_not_in_E E c He); auto.
        rewrite Heql; auto; rewrite <- Hs0 in Hnemp; auto.

        apply not_same_loc_implies_same_last' with E; auto.
        rewrite Hc in Hde; assert (l = d (DW c) \/ l = f (DW c)) as Horl; auto.
        assert (rfmaps_well_formed E (events E) (rf X)) as Hrfwf.
          destruct_valid Hv; split; auto.
        generalize (label_implies_evt E X l (DW c) Hwf Hrfwf Hlb Horl); intros He; simpl in He; auto. 
        intro Heq; rewrite Heq in Hs0; apply Hnemp; auto.


  
  destruct c as [w r];
  apply flush_read_same_last with E p (labels_of_es E X) s0 w r; auto.

  
  unfold flush_write_step in Hs;
  destruct (bemp (rrestrict (rel_union (B s0) (Q s0)) (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p (labels_of_es E X) )) e c))); [|inversion Hs];
  destruct (bemp (rrestrict (Q s0) (fun r' => loc r' = loc c /\ po_iico E r' c))); [|inversion Hs].
  destruct (eqEv_dec (pick_last (rrestrict (B s0) (fun e : Event => loc e = loc c))) c) as [HeqB|]; [|inversion Hs].
  destruct (eqLoc_dec (loc c) (loc we)) as [Heql | Hneql].
    rewrite Heql in HeqB; rewrite HeqB in Hs0;
    rewrite Hs0 in Hc; rewrite Hc in Hde; contradiction.

  inversion Hs as [Heq]; simpl; rewrite <- not_same_loc_implies_same_last with E (B s0) we c; auto.
    assert (evt_of_label l = c) as Heqe.
      rewrite Hde; rewrite Hc; simpl; auto.
    rewrite <- Heqe; apply wfl_le2 with X; auto.
Qed.

Lemma flushed_is_last :
  forall E X lst delay n s0 s1 w,
  well_formed_event_structure E ->
  valid_execution E X ->
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) ->
  w <> dB ->
  (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w))) /\ machine E d (labels_of_es E X) lst n (Some s0) (Some s1)) ->
  pick_last (rrestrict (B s1) (fun e => loc e = loc w)) = w. 
Proof.
intros E X lst delay n s0 s2 w Hwf Hv Hwfd Hwfm Hwfb  Hnemp Hmd;
generalize s2 lst Hwfm Hwfb Hmd; clear s2 lst Hwfm Hwfb Hmd; 
induction n.

intros s2 lst Hwfm Hwfb [d [Hi Hm]]; inversion Hm.
    admit. 

intros s2 lst Hwfm Hwfb Hmd.
destruct Hmd as [d [Hid [l [s1 [Hudr [Hm [Hs [? [Hwfm1 [Hwfb1 Hwfq1]]]]]]]]]]; 
fold (machine E (pred d l) (labels_of_es E X) l n (Some s0) (Some s1)) in Hm;
fold (step E d (labels_of_es E X) (Some s1) l) in Hs.

assert (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w))) /\ machine E  d (labels_of_es E X) l n (Some s0) (Some s1)) as Hed.
  exists (pred d l); split; auto.
  intros x y Hxy; apply Hid. 
    generalize Hxy; apply pred_incl; auto.
generalize (IHn s1 l Hwfm1 Hwfb1 Hed); intro Hneq.

destruct (eqLabel_dec l (f (DW w))) as [Heqf | Hneqf].


    rewrite Heqf in Hudr.

        assert (Included _ (labels_of_es E X) (labels_of_es E X)) as Htl.
          intros e He; auto.
        assert (partial_order (tc (X2p E X delay)) (labels_of_es E X)) as Hpl.
          apply tc_X2p_partial_order; auto.
        generalize (OEL.OE (labels_of_es E X) (labels_of_es E X) (tc (X2p E X delay)) Htl Hpl);
          clear Htl; intros [Hil Hl].

  assert (OEL.LE (tc (X2p E X delay)) (f (DW w)) (f (DW w))) as Hc.
      assert (ran d (f (DW w))) as Hran.  
        destruct Hudr; apply last_in_ran; auto.
      destruct Hran as [e Hran]; generalize (Hid e (f (DW w)) Hran);
      intros [Hde Hore]; auto. 
        
        

       

   destruct_lin Hl; generalize (Hac (f (DW w)) Hc); intro Ht; inversion Ht.


  apply not_flush_implies_same_last with E d X s1 l; auto.
  assert (Included _ (labels_of_es E X) (labels_of_es E X)) as Htl.
    intros e He; auto.
  generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
  generalize (OEL.OE (labels_of_es E X) (labels_of_es E X) (tc (X2p E X delay)) Htl Hpart);
  clear Htl; intros [Hil Hl].
  destruct_lin Hl; apply Hdr.
    assert (ran d l) as Hran.
      destruct Hudr; apply last_in_ran; auto.
    destruct Hran as [l' Hdll'];
    generalize (Hid l' l Hdll'); intros [Hll' ?]; right; exists l'; auto.
Qed.

Lemma po_fr :
  forall E X r w,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  reads E r -> writes E w ->
  po_iico E r w -> loc r = loc w ->
  fr E X r w.
Proof.
intros E X r w Hwf Hv Hr Hw Hpo Hl.
destruct_valid Hv.
generalize (Hrf_init r Hr); intros [wr [Hewr Hrf]].

assert (pio_llh E r w) as Hpio.
  split; [|split]; auto.
    intros [? [? [l [v Haw]]]];
    destruct Hw as [? [? [? Har]]].
    rewrite Haw in Har; inversion Har.

destruct (eqEv_dec w wr) as [Heq | Hneq].

assert (hb' E X w r) as Hhb.
  rewrite Heq; left; left; left; left; auto.
generalize (Hsp r w Hpio); intro; contradiction.

assert (ws X w wr \/ ws X wr w) as Hor.
  generalize (Hws_tot (loc w)); intro Hlin; destruct_lin Hlin.
  assert ((writes_to_same_loc_l (events E) (loc w)) w) as Hww.
    destruct Hw as [? [l [v Haw]]]; 
    split; auto; exists v; unfold loc; rewrite Haw; auto.
  assert ((writes_to_same_loc_l (events E) (loc w)) wr) as Hwwr.
    rewrite <- Hl; rewrite <- ABasic.rf_implies_same_loc2 with E X wr r; auto.
    split; auto.
    generalize (ABasic.dom_rf_is_write E X wr r Hrf_cands Hrf);
    intros [l [v Haw]]; exists v; unfold loc; rewrite Haw; auto.
      split; split; auto.
     
  generalize (Htot w wr Hneq Hww Hwwr); intros [[? ?]|[? ?]]; auto.

inversion Hor as [Hwwr | Hwrw].

  assert (hb' E X w r) as Hhb.
    left; right; exists wr; split; auto.
  generalize (Hsp r w Hpio); intro; contradiction.

  split; [|split]; auto. 
    destruct Hr; auto.
    destruct Hw; auto.
    exists wr; split; auto.
Qed.

Lemma unifr_contrad' :
  forall E X delay n s0 s1 w1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) ->
  wf_rf_labels E (labels_of_es E X) ->
  writes E w1 ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w1))) /\ 
    (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DW w1))) \/ last d = f (DW w1)) /\
     machine E d (labels_of_es E X) (f (DW w1)) n (Some s0) (Some s1)) ->
  bemp (rrestrict (Q s1) (fun r' => loc r' = loc w1 /\ po_iico E r' w1)) = true.
Proof.
intros E X delay n s0 s1 w1 Hwf Hv Hwfd Hwfl Hwfrfl Hw1 Hwfm Hwfb Hwfq Hed.
case_eq (bemp (rrestrict (Q s1) (fun r' => loc r' = loc w1 /\ po_iico E r' w1))); auto.
intro Hnemp; generalize (bemp_axf (Q s1) (fun r' => loc r' = loc w1 /\ po_iico E r' w1) Hnemp);
intros [r [Hq [Heql Hporw]]].
assert (events E r) as Her.
  apply ABasic.po_iico_domain_in_events with w1; auto.
assert (reads E r) as Hr.
  apply Hwfq; auto.

generalize Hv; intro Hv'; destruct_valid Hv'; 
generalize (Hrf_init r Hr); intros [wr [Hlfwr Hrf]].
assert (labels_of_es E X (f (DW w1))) as Hlb.
  destruct Hwfl as [? [? [Hle ?]]]; apply Hle; auto.
generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
assert (trans (OEL.LE (tc (X2p E X delay)))) as Htp.
  apply oel_trans with (labels_of_es E X); auto.
assert (forall l1 l2 : Label,
  l1 <> l2 ->
  labels_of_es E X l1 ->
  labels_of_es E X l2 ->
  OEL.LE (tc (X2p E X delay)) l1 l2 \/ OEL.LE (tc (X2p E X delay)) l2 l1) as Htotp.
  apply oel_tot; auto.
assert (forall de : DEvent, labels_of_es E X (f de) -> tc (OEL.LE (tc (X2p E X delay))) (d de) (f de)) as Hdf.
  intros de Hede; apply trc_step. 
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intros Hi; apply Hi; apply trc_step; apply X2p_df; auto.
    apply label_implies_evt with X (f de); auto.
      split; auto.
assert (f (DW w1) <> f (DR (wr,r))) as Hdiff.
  intro Heq; inversion Heq.
assert (evt_of_devt (DR (wr,r)) = r) as Hedr.
  auto. 
assert (Included Label
    (Union Label (dom (OEL.LE (tc (X2p E X delay))))
       (ran (OEL.LE (tc (X2p E X delay))))) (labels_of_es E X)) as Hudr. 
  apply oel_udr; auto.
assert (labels_of_es E X (f (DR (wr,r)))) as Hldr.
  right; exists wr; exists r; split; auto.
assert (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f (DW w1)) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DW w1))) \/
     f (DW w1) = f (DW w1)) as Hlst.
  auto.
assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hacp.
  apply oel_ac with (labels_of_es E X); auto.


generalize Hwfl; intros [Hwflw ?];
generalize (in_queue_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DW w1)) n s0 s1 (f (DW w1)) (DR (wr,r)) r Hwfl Hwfrfl Hwfq Her Hedr Hdiff Hlb Hldr Hdiff Hlb Hudr Hacp Htp Htotp Hdf Hlst Hq Hed);
intro Hw1r.



generalize (po_fr E X r w1 Hwf Hv Hr Hw1 Hporw Heql); intro Hfr.
generalize (X2p_fr E X delay (DR (wr,r)) (DW w1) Hfr); intro Hprw1.

generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpart); intro Hac.
assert False as Ht.
  apply (Hac (f (DR (wr,r)))); apply trc_ind with (f (DW w1)); auto; apply trc_step; auto.
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hi; apply Hi; apply trc_step; auto.
inversion Ht.
Qed.

Lemma fwrites_dont_block_ex :
  forall E X delay n s0 s1 w1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  writes E w1 ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w1))) /\ 
    (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DW w1))) \/ last d = f (DW w1)) /\
     machine E d (labels_of_es E X) (f (DW w1)) n (Some s0) (Some s1)) ->
  ~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DW w1)) = None).
Proof.
intros E X delay n s0 s1 w1 Hwf Hv Hwfd Hwfl Hwfrfl Hew1 Hwfm Hwfb Hwfq Hed' Hs; simpl in Hs;
unfold flush_write_step in Hs.
assert (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w1))) /\
     machine E d (labels_of_es E X) (f (DW w1)) n (Some s0) (Some s1)) as Hed.
  destruct Hed' as [d [? [? ?]]]; exists d; split; auto.

case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X) )) e w1))); 
  intro Hppo; rewrite Hppo in Hs.

  clear Hppo; case_eq (bemp (rrestrict (Q s1) (fun r' => loc r' = loc w1 /\ po_iico E r' w1)));
  intro Huniwr; rewrite Huniwr in Hs.

    clear Huniwr; destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e => loc e = loc w1))) w1).

      inversion Hs.
      
       assert (w1 <> dB) as HndB.
         intros Heq; rewrite Heq in Hew1; destruct Hew1 as [Hew1 ?]; simpl in Hew1. 
           apply (dB_not_in_E E dB); auto.
       generalize (flushed_is_last E X (f (DW w1)) delay n s0 s1 w1 Hwf Hv Hwfd Hwfm Hwfb HndB Hed); intro; contradiction.

    rewrite unifr_contrad' with E X delay n s0 s1 w1 in Huniwr; auto; inversion Huniwr.

  
  apply contrad_ppo with E X delay n s0 s1 (DW w1); auto.
    destruct Hew1 as [? ?]; auto. 
    destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto.
Qed.

Lemma fwrites_dont_block :
  forall E X delay n s0 s1 w1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) ->
  wf_rf_labels E (labels_of_es E X) ->
  writes E w1 ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DW w1))) (labels_of_es E X) (f (DW w1)) n (Some s0) (Some s1) ->
  ~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DW w1)) = None).
Proof.
intros E X delay n s0 s1 w1 Hwf Hv Hwfd Hwfl Hwfrfl Hew1 Hwfm Hwfb Hwfq Hm.
apply fwrites_dont_block_ex with n s0; auto.
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DW w1))); split; [|split]; auto. 
    intros x y Hxy; auto.
    left; apply no_interv_last; auto; intros x y Hxy; auto.
Qed.

Lemma b_ws :
  forall E X delay n s0 s1 w w',
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  (exists d, rel_incl d ((pred (OEL.LE (tc (X2p E X delay))) (f (DW w')))) /\ 
    (~(exists l' : Label,
       tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
       tc (OEL.LE (tc (X2p E X delay))) l' (f (DW w'))) \/ last d = f (DW w')) /\
     machine E d (labels_of_es E X) (f (DW w')) n (Some s0) (Some s1)) ->
  B s1 w w' -> loc w = loc w' ->
  ws X w w'.
Proof. 
intros E X delay n s0 s1 w w' Hwf Hv Hwfd Hwfl Hwfrfl Hwfm Hwfb Hwfq Hed' Hww' Heql.
assert (exists d, rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w'))) /\
     machine E d (labels_of_es E X) (f (DW w')) n (Some s0) (Some s1)) as Hed.
  destruct Hed' as [d [? [? ?]]]; exists d; split; auto.

generalize Hv; intro Hv'; destruct_valid Hv'.
generalize (Hws_tot (loc w)); intro Hl; destruct_lin Hl.
 clear Hws_tot Hws_cands Hrf_init Hrf_cands Hrf_uni Hdr Hac Htrans.
  assert (w <> w') as Hd.
    intro Heq; rewrite Heq in Hww'.
    assert (tc (rrestrict (B s1) (fun e => loc e = loc w')) w' w') as Hc.
      apply trc_step; split; auto.
        split; unfold Ensembles.In; auto.
    destruct Hwfb as [? [Ht [Hac ?]]]; apply (Hac w' w' Hc). 

  assert ((writes_to_same_loc_l (events E) (loc w)) w) as Hw.
    assert (udr (B s1) w) as Hbw.
      left; exists w'; auto.
    destruct Hwfb as [Hudrb [? ?]]; generalize (Hudrb w Hbw); intros [? [? [v Haw]]];
    split; auto; exists v; auto; unfold loc; rewrite Haw; auto.
  assert ((writes_to_same_loc_l (events E) (loc w)) w') as Hw'.
    rewrite Heql.
    assert (udr (B s1) w') as Hbw'.
      right; exists w; auto.
    destruct Hwfb as [Hudrb [? ?]]; generalize (Hudrb w' Hbw'); intros [? [? [v Haw']]];
    split; auto; exists v; auto; unfold loc; rewrite Haw'; auto.
  generalize (Htot w w' Hd Hw Hw'); intro Hor; 
    inversion Hor as [Hws|Hws]; clear Hor; destruct Hws as [Hws ?]; auto.
  assert (X2p E X delay (f (DW w')) (f (DW w))) as Hp.
    destruct Hwfl as [? [? [Hwl ?]]].
    apply X2p_ws; auto. 
       
    apply Hwl; auto. 
      destruct Hw' as [? [v Haw]]; auto;
      split; auto; exists (loc w); exists v; auto. 
    apply Hwl; auto.
      destruct Hw as [? [v Haw]]; auto;
      split; auto; exists (loc w); exists v; auto. 
  clear Htot H.

  assert (step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DW w')) = None) as Hc.
     simpl; unfold flush_write_step.
    destruct (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e : Event =>
          rel_union (A.ppo E) (A.abc E (p2X E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) )) e w'))); auto.
    destruct (bemp (rrestrict (Q s1) (fun r' => loc r' = loc w' /\ po_iico E r' w'))); auto.
    destruct (eqEv_dec (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w'))) w') as [Heq |]; auto.
    generalize (EPick.pick_in (fun w : Event =>
         ~
         (exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') dB w0) /\
         w = dB \/
         (exists w0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') dB w0) /\
         rrestrict (B s1) (fun e : Event => loc e = loc w') dB w /\
         ~
         (exists w'0 : Event,
            rrestrict (B s1) (fun e : Event => loc e = loc w') w'0 w)));
    intros Hinl; fold (pick_last (rrestrict (B s1) (fun e => (loc e) = (loc w')))) in Hinl;
    inversion Hinl as [[? Hd'] | [? [Hil' Hl']]]; clear Hinl. 
      rewrite Hd' in Heq.
  assert (events E w') as Hew'.
    apply ABasic.dom_ws_in_events with X w; auto.
      clear Hsp Hth Hvalid; destruct_valid Hv; split; auto.
        rewrite <- Heq in Hew'; generalize (dB_not_in_E E dB Hew'); intro Hc. 
        assert False as Ht.
          apply Hc; auto.
        inversion Ht.
      rewrite Heq in Hl'.
      assert (exists e' : Event, rrestrict (B s1) (fun e : Event => loc e = loc w') e' w') as Hc.
        exists w; split; auto. 
          unfold Ensembles.In; split; auto.
      contradiction.  

  assert (writes E w') as Hew'.
    split. apply ABasic.dom_ws_in_events with X w; auto.
      clear Hsp Hth Hvalid; destruct_valid Hv; split; auto.
      assert (write_serialization_well_formed (events E) (ws X)) as Hwswf.
        clear Hsp Hth Hvalid; destruct_valid Hv; split; auto.
      generalize (ABasic.dom_ws_is_write E X w' w Hwswf Hws); auto.
  generalize (fwrites_dont_block_ex E X delay n s0 s1 w' Hwf Hv Hwfd Hwfl Hwfrfl Hew' Hwfm Hwfb Hwfq Hed'); intro. contradiction.
Qed.

Lemma uniwsrf_contrad :
  forall E X delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> 
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  w1 <> M s1 (loc r1) ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DW w1)) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  rf X w1 r1 ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  bin (fun w => (B s1 w (waft E (B s1) r1)) /\ loc r1 = loc w) w1 = true.
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfb Hwfq Hnm Hwfl Hwfrfl Hlfw Hlfwr Hrf Hm; 
case_eq (bin (fun w => (B s1 w (waft E (B s1) r1)) /\ loc r1 = loc w) w1); auto.

assert (loc r1 = loc w1) as Heql.
  assert (rfmaps_well_formed E (events E) (rf X)) as Hwfrf.
    destruct_valid Hv; split; auto.
    apply sym_eq; apply ABasic.rf_implies_same_loc2 with E X; auto; split; auto.
      destruct_valid Hv; split; auto.

intro Hf; generalize (nbin_nIn (fun w => (B s1 w (waft E (B s1) r1)) /\ loc r1 = loc w) w1 Hf); clear Hf; intro Hf.
generalize (in_order_implies_mem_or_buff_state E X  delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfl Hlfw Hlfwr Hrf Hm); intros [Him | Hib].

rewrite Him in Hnm; assert False as Ht.
  apply Hnm; auto.
inversion Ht.

generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc r1 = loc w /\
                                    po_iico E r1 w /\
                                    ~ (exists w', loc r1 = loc w' /\ po_iico E r1 w' /\ po_iico E w' w)));
fold (waft E (B s1) r1); intros [Hr [Hl [Hpo Hn]]].

destruct (eqEv_dec w1 (waft E (B s1) r1)) as [Heq | Hneq].

  rewrite <- Heq in Hpo.
  assert (pio_llh E r1 w1) as Hpio.
    split; [|split]; auto.
      intros [? [? [? [? Har]]]].
      destruct_valid Hv; generalize (ABasic.dom_rf_is_write E X w1 r1 Hrf_cands Hrf);
      intros [? [? Haw]]; rewrite Haw in Har; inversion Har.
  assert (hb' E X w1 r1) as Hhb.
    left; left; left; left; auto.
  destruct_valid Hv; generalize (Hsp r1 w1 Hpio); intro Ht; contradiction.

  rewrite Heql in Hl; generalize Hwfb; intros [? [Htb Hacb]];
  generalize (Htb w1 (waft E (B s1) r1) Hneq Hib Hr Hl); intro Hor; inversion Hor as [H1aft | Haft1]; clear Hor.
    assert ((B s1 w1 (waft E (B s1) r1)) /\ loc r1 = loc w1) as Hc.
      split; auto.  
    contradiction.

  assert (pio_llh E r1 (waft E (B s1) r1)) as Hpio.
    split; [|split]; auto.
      rewrite <- Hl; apply sym_eq; apply ABasic.rf_implies_same_loc2 with E X; auto.
        destruct_valid Hv; split; split; auto.
        assert (udr (B s1) (waft E (B s1) r1)) as Hudr.
          left; exists w1; auto.
        destruct Hwfb as [Hwfb ?]; generalize (Hwfb (waft E (B s1) r1) Hudr);
        intros [? [? [? Haw]]]; intros [? [? [? [? Har]]]].
        rewrite Har in Haw; inversion Haw.

generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
generalize (oel_trans (X2p E X delay) (labels_of_es E X) Hpart); intro Htrans.
generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpart); intro Hac.

assert (exists r, trans r /\ rel_incl r (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ 
           (~(exists l' : Label,
            tc (OEL.LE (tc (X2p E X delay))) (last r) l' /\
            tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/ last r = f (DR (w1,r1))) /\
          machine E r (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed.
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split; [|split]]; auto.
    apply td_tpred; auto. 
    intros x y Hxy; auto. 
    left; apply no_interv_last; auto; intros x y Hxy; auto. 

  assert (hb' E X (waft E (B s1) r1) r1) as Hhb.
    left; right; exists w1; split; auto.
        apply b_ws with E delay n s0 s1; auto.

        apply mns_fwd with (f (DR (w1,r1))); auto.
        apply oel_udr; auto.
        apply oel_tot; auto.
        apply oel_tpred; auto.

rewrite <- trans_tc with Label (OEL.LE (tc (X2p E X delay))); auto.
apply in_buff_implies_f_after2 with E (labels_of_es E X) (f (DR (w1,r1))) n s0 s1; auto.
  apply oel_udr; auto. 
  split. apply ABasic.dom_rf_in_events with X r1; auto.
    destruct_valid Hv; split; auto.
    apply ABasic.dom_rf_is_write with E X r1; auto.
    destruct_valid Hv; auto.
  intro Heq; inversion Heq.
  intro Heq; inversion Heq.
  apply oel_tot; auto.
  generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hi.
  intros e He; apply trc_step; apply Hi; apply trc_step; apply X2p_df.
    simpl; auto. destruct He; auto. 
      destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto. 

  destruct Hed as [r [? [? [? ?]]]]; exists r; split; [|split]; auto.

  right; exists (waft E (B s1) r1); auto. 

  destruct_valid Hv; generalize (Hsp r1 (waft E (B s1) r1) Hpio); intro Ht; contradiction.
Qed. 

Lemma unifr_contrad :
  forall E X delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> 
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DW w1)) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  rf X w1 r1 ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  w1 <> M s1 (loc r1) ->
  bin (fun w => (w = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w) /\ loc r1 = loc w) w1 = true.
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfb Hwfq Hwfl Hwfrfl Hlfw Hlfwr Hrf Hm Hneqm;
case_eq (bin (fun w => (w = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w) /\ loc r1 = loc w) w1); auto.
assert (loc r1 = loc w1) as Heql.
  assert (rfmaps_well_formed E (events E) (rf X)) as Hwfrf.
    destruct_valid Hv; split; auto.
    apply sym_eq; apply ABasic.rf_implies_same_loc2 with E X; auto; split; auto.
      destruct_valid Hv; split; auto.

intro Hf; generalize (nbin_nIn (fun w => (w = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w) /\ loc r1 = loc w) w1 Hf); clear Hf; intro Hf.
generalize (in_order_implies_mem_or_buff_state E X  delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfl Hlfw Hlfwr Hrf Hm); intros [Him | Hib].

rewrite Him in Hneqm; assert False as Ht.
  apply Hneqm; auto.
inversion Ht.

generalize (EPick.pick_in (fun w => ran (B s1) w /\ loc r1 = loc w /\ po_iico E w r1 /\
   ~ (exists w' : Event, loc r1 = loc w' /\ po_iico E w w' /\ po_iico E w' r1)));
fold (wbef E (B s1) r1); intros [Hr [Hl [Hpo Hn]]].

destruct (eqEv_dec w1 (wbef E (B s1) r1)) as [Heq | Hneq].

  assert ((w1 = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w1) /\ loc r1 = loc w1) as Hc.
    split; auto. 
  contradiction.

  rewrite Heql in Hl; generalize Hwfb; intros [? [Htb Hacb]];
  generalize (Htb w1 (wbef E (B s1) r1) Hneq Hib Hr Hl); intro Hor; inversion Hor as [H1aft | Haft1]; clear Hor.

    generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
    generalize (oel_trans (X2p E X delay) (labels_of_es E X) Hpart); intro Htrans.
    generalize (oel_ac (X2p E X delay) (labels_of_es E X) Hpart); intro Hac.

  assert (exists d, rel_incl d ((pred (OEL.LE (tc (X2p E X delay))) (f (DW (wbef E (B s1) r1))))) /\ 
          (~(exists l' : Label,
            tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
            tc (OEL.LE (tc (X2p E X delay))) l' (f (DW (wbef E (B s1) r1)))) \/ last d = f (DW (wbef E (B s1) r1))) /\
            machine E d (labels_of_es E X) (f (DW (wbef E (B s1) r1))) n (Some s0) (Some s1)) as Hed.
    assert (exists r, trans r /\ rel_incl r (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ 
             (~(exists l' : Label,
               tc (OEL.LE (tc (X2p E X delay))) (last r) l' /\
               tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1,r1)))) \/ last r = f (DR (w1,r1))) /\
              machine E r (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed.
      exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split; [|split]]; auto. 
        apply td_tpred; auto.
        intros x y Hxy; auto. 
        left; apply no_interv_last; auto; intros x y Hxy; auto. 

    apply mns_fwd with (f (DR (w1,r1))); auto.
      apply oel_udr; auto. apply oel_tot; auto. apply oel_tpred; auto.

    rewrite <- trans_tc with Label (OEL.LE (tc (X2p E X delay))); auto.
    apply in_buff_implies_f_after2 with E (labels_of_es E X) (f (DR (w1,r1))) n s0 s1; auto.
      apply oel_udr; auto.
    assert (udr (B s1) (wbef E (B s1) r1)) as Hudr.
      right; exists w1; auto.
    destruct Hwfb as [Hwfb ?]; generalize (Hwfb (wbef E (B s1) r1) Hudr); intros [? ?]; auto.
  
    intro Heq; inversion Heq.
    intro Heq; inversion Heq.
    apply oel_tot; auto.
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hi.
      intros e He; apply trc_step; apply Hi; apply trc_step; apply X2p_df.
      simpl; auto. 
        destruct He; auto.
      destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto.
  
    destruct Hed as [r [? [? [? ?]]]]; exists r; split; [|split]; auto.
   
    right; exists w1; auto.

 generalize (b_ws E X delay n s0 s1 w1 (wbef E (B s1) r1) Hwf Hv Hwfd Hwfl Hwfrfl Hwfm Hwfb Hwfq Hed H1aft Hl); intro Hws.
  assert (fr E X r1 (wbef E (B s1) r1)) as Hfr.
    split; [|split]; auto.
      assert (rfmaps_well_formed E (events E) (rf X)) as Hwfrf.
        destruct_valid Hv; split; auto.
      apply ABasic.ran_rf_in_events with X w1; auto.
      apply ABasic.ran_ws_in_events with X w1; auto.
        destruct_valid Hv; split; auto.
      exists w1; split; auto.
  destruct_valid Hv.

  assert (pio_llh E (wbef E (B s1) r1) r1) as Hpio.
    split; [|split]; auto.
      apply sym_eq; apply ABasic.fr_implies_same_loc with E X; auto.
        split; split; auto.
      intros [[? [? [? Hrw]] ?]].
      assert (write_serialization_well_formed (events E) (ws X)) as Hwswf.
        split; auto.
      generalize (ABasic.ran_ws_is_write E X w1 (wbef E (B s1) r1) Hwswf Hws); intros [? [? Hww]].
      rewrite Hrw in Hww; inversion Hww.

  assert ((hb' E X) r1 (wbef E (B s1) r1)) as Hhb.
    left; left; left; right; auto.

  generalize (Hsp (wbef E (B s1) r1) r1 Hpio); intro Ht; contradiction.

 assert ((w1 = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w1) /\ loc r1 = loc w1) as Hc.
    split; auto. 
  contradiction.
Qed.

Lemma mb_ws :
  forall E X lst delay n s0 s1 w w',
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay -> wf_labels E (labels_of_es E X) ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  (exists d, trans d /\ rel_incl d ((pred (OEL.LE (tc (X2p E X delay))) (f (DW w')))) /\ machine E d (labels_of_es E X) lst n (Some s0) (Some s1)) ->
  w = M s1 (loc w) -> ran (B s1) w' -> loc w = loc w' ->
  ws X w w'.
Proof. 
intros E X lst delay n s0 s1 w w' Hwf Hv Hwfd Hwfl Hwfm Hwfb Hwfq [d [Htd [Hid Hm]]] Hmw Hbw' Heql.
generalize Hv; intro Hv'; destruct_valid Hv'.
generalize (Hws_tot (loc w)); intro Hl; destruct_lin Hl.
 clear Hws_tot Hws_cands Hrf_init Hrf_cands Hrf_uni Hdr Hac Htrans.
  generalize (oel_trans (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Htr.

  assert (exists d : Rln Label, trans d /\
          rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DW w'))) /\
          machine E d (labels_of_es E X) lst n (Some s0) (Some s1)) as Hed.
    exists d; split; [|split]; auto.

  assert (w <> w') as Hd.
    generalize (in_mem_implies_in_order E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) lst n s0 s1 (f (DW w')) w (loc w) Htr Hed Hmw); intro Hww'.
    intro Heq; rewrite Heq in Hww'. 
    apply (oel_ac (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd) (f (DW w'))); apply trc_step; auto.

  assert ((writes_to_same_loc_l (events E) (loc w)) w) as Hw.
      generalize (Hwfm (loc w)); intros [[Hew [l [v Haw]]] Hloc]. 
      split; auto.
      rewrite Hmw; auto.
      unfold loc in Hloc; fold (loc w) in Hloc; rewrite Haw in Hloc.
      rewrite <- Hloc; rewrite Hmw.
      exists v; rewrite Haw; auto.
  assert ((writes_to_same_loc_l (events E) (loc w)) w') as Hw'.
    rewrite Heql.
    assert (udr (B s1) w') as Hudrbw'.
      right; auto.
    destruct Hwfb as [Hudrb [? ?]]; generalize (Hudrb w' Hudrbw'); intros [? [? [v Haw']]];
    split; auto; exists v; auto; unfold loc; rewrite Haw'; auto.
  generalize (Htot w w' Hd Hw Hw'); intro Hor; 
    inversion Hor as [Hws|Hws]; clear Hor; destruct Hws as [Hws ?]; auto.
  assert (X2p E X delay (f (DW w')) (f (DW w))) as Hp.
    destruct Hwfl as [? [? [Hwl ?]]].
    apply X2p_ws; auto. 
       
    apply Hwl; auto. 
      destruct Hw' as [? [v Haw]]; auto;
      split; auto; exists (loc w); exists v; auto. 
    apply Hwl; auto.
      destruct Hw as [? [v Haw]]; auto;
      split; auto; exists (loc w); exists v; auto. 
  clear Htot H.

  generalize (in_mem_implies_in_order E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) lst n s0 s1 (f (DW w')) w (loc w) Htr Hed Hmw);
  intros Hww'.
  assert (OEL.LE (tc (X2p E X delay)) (f (DW w')) (f (DW w))) as Hw'w.
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intros Hi.
    apply Hi; apply trc_step; auto.
  generalize (Htr (f (DW w)) (f (DW w')) (f (DW w)) Hww' Hw'w); intro Hc.
  generalize (oel_ac (X2p E X delay)(labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intros Hac.
  rewrite <- trans_tc with Label (OEL.LE (tc (X2p E X delay))) in Hc; auto.
  generalize (Hac (f (DW w)) Hc); intro Ht; inversion Ht.
Qed.

Lemma mns_fwd' :
  forall E (p:Rln Label) (L:set Label) lst n s0 s1 l1 l2,
  trans p ->
  p l1 l2 -> 
  machine E (pred p l1) L lst n (Some s0) (Some s1) ->
  (exists d, trans d /\ rel_incl d (pred p l2) /\ machine E d L lst n (Some s0) (Some s1)).
Proof.
intros E p L lst n s0 s1 l1 l2 Htp H12 Hm.
exists (pred p l1); split; [|split]; auto.
  apply td_tpred; auto.
  intros x y [? Hor]; split; auto.
  
  apply Htp with l1; auto.
Qed.

Lemma unifr_contrad'' :
  forall E X delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> 
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  wf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DW w1)) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  rf X w1 r1 ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  w1 = M s1 (loc r1) ->
  bemp (rrestrict (B s1) (fun w : Event => po_iico E w r1 /\ loc r1 = loc w)) = true.
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfb Hwfq Hwfl Hlw1 Hlr1 Hrf Hm Him.
case_eq (bemp (rrestrict (B s1) (fun w : Event => po_iico E w r1 /\ loc r1 = loc w))); intro Hib; auto.
generalize (bemp_axf (B s1) (fun w : Event => po_iico E w r1 /\ loc r1 = loc w) Hib); clear Hib.
intros [w' [Hib [Hpo Hloc]]].

generalize (oel_trans (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Htr.
assert (exists d,  rel_incl d (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\
          (~(exists l' : Label,
            tc (OEL.LE (tc (X2p E X delay))) (last d) l' /\
            tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/ last d = f (DR (w1,r1))) /\
        machine E d (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed.
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split]; auto.
     intros e1 e2 H12; auto. 
    left; apply no_interv_last; auto; intros x y Hxy; auto. 
assert (tc (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1))) (f (DW w'))) as Hp.
  apply in_buff_implies_f_after2 with E (labels_of_es E X) (f (DR (w1,r1))) n s0 s1; auto.
    apply oel_udr; auto.
      apply tc_X2p_partial_order; auto.
    apply Hwfb; right; auto.
    intro Heq; inversion Heq. 
    intro Heq; inversion Heq.
    apply oel_ac with (labels_of_es E X); auto. apply (tc_X2p_partial_order E X delay Hwf Hv Hwfd). 
    apply oel_tot; auto. apply (tc_X2p_partial_order E X delay Hwf Hv Hwfd). 
    intros e He; apply trc_step. 
      generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi; 
      apply Hi; apply trc_step; apply X2p_df; auto. 
      simpl; destruct He; auto. 
      destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto.


    right; auto.

rewrite trans_tc in Hp; auto.
generalize (mns_fwd' E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DR (w1,r1))) n s0 s1 (f (DR (w1,r1))) (f (DW w')) Htr Hp Hm);
intro Hmnsw'.

assert (loc r1 = loc w1) as Heqlrw.
  apply sym_eq. apply ABasic.rf_implies_same_loc2 with E X; auto.
    destruct_valid Hv; split; split; auto.

rewrite Heqlrw in Him.
assert (loc w1 = loc w') as Heqlww'.
  rewrite <- Heqlrw; auto.
generalize (mb_ws E X (f (DR (w1,r1))) delay n s0 s1 w1 w' Hwf Hv Hwfd Hwfl Hwfm Hwfb Hwfq Hmnsw' Him Hib Heqlww');
intro Hws.

assert (fr E X r1 w') as Hfr.
  split; [|split]; auto.
    apply ABasic.po_iico_range_in_events with w'; auto.
    apply ABasic.po_iico_domain_in_events with r1; auto.
    exists w1; split; auto.

assert (pio_llh E w' r1) as Hpio.
    split; [|split]; auto.
      intros [[? [? [? Hw']]] ?].
      assert (write_serialization_well_formed (events E) (ws X)) as Hwswf.
        destruct_valid Hv; split; auto.
      generalize (ABasic.ran_ws_is_write E X w1 w' Hwswf Hws); intros [? [? Haw]];
      rewrite Haw in Hw'; inversion Hw'.  

assert ((hb' E X) r1 w') as Hhb.
  left; left; left; right; auto.

destruct_valid Hv; generalize (Hsp w' r1 Hpio); intro Ht; contradiction.
Qed.

Lemma grf_contrad :
  forall E X delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X -> 
  wf_delay delay ->
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  wf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DW w1)) ->
  labels_of_es E X (f (DR (w1,r1))) ->
  rf X w1 r1 ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  bin (Intersection _ (ran (B s1)) (rrB r1)) w1 = false.
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfm Hwfb Hwfq Hwfl Hlw1 Hlr1 Hrf Hm.
case_eq (bin (Intersection _ (ran (B s1)) (rrB r1)) w1); intro Hib; auto.
generalize (bin_In (Intersection _ (ran (B s1)) (rrB r1)) w1 Hib); clear Hib; intro Hinter.
destruct Hinter as [w1 Hbw Hrrw].

assert (events E w1) as Hew1.
  assert (udr (B s1) w1) as Hudr.
    right; auto.
  destruct Hwfb as [Hwfb ?]; generalize (Hwfb w1 Hudr); intros [? ?]; auto.
assert (f (DR (w1,r1)) <> f (DW w1)) as Hdiff.
  intro Heq; inversion Heq.
assert (acyclic (OEL.LE (tc (X2p E X delay)))) as Hacp.
  apply oel_ac with (labels_of_es E X) ; auto.
    apply tc_X2p_partial_order; auto.
assert (trans (OEL.LE (tc (X2p E X delay)))) as Htp.
  apply oel_trans with (labels_of_es E X) ; auto.
    apply tc_X2p_partial_order; auto.
assert (forall l1 l2, l1 <> l2 -> labels_of_es E X l1 -> labels_of_es E X l2 -> (OEL.LE (tc (X2p E X delay))) l1 l2 \/ (OEL.LE (tc (X2p E X delay))) l2 l1) as Htotp.
  apply oel_tot; auto.
    apply tc_X2p_partial_order; auto.
assert (exists r, rel_incl r (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ machine E r (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed. 
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; auto.
    intros e1 e2 H12; auto.
assert (forall w, writes E w -> tc (OEL.LE (tc (X2p E X delay))) (d (DW w)) (f (DW w))) as Hdf.
  intros e He. 
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi;
    apply trc_step; apply Hi; apply trc_step; apply X2p_df; auto.
    destruct He; simpl; auto.
    destruct Hwfl as [? [? [Hw ?]]]; apply Hw; auto.

assert (~ delay E X (evt_of_devt (DW w1)) (evt_of_devt (DR (w1,r1)))) as Hnd.
  simpl; apply ghb_ndelay; auto.
    case_eq A.inter; case_eq A.intra; intros intra inter;
    unfold rrB in Hrrw; unfold AWmm.ghb; rewrite intra in * |- *; rewrite inter in * |- *;
    destruct Hrrw as [? Hgrf].
      inversion Hgrf as [Hinter | Hintra]; [right;left | left]; split; auto.
      left; split; auto.
      left; split; auto.
      inversion Hgrf.
assert (Included Label
    (Union Label (dom (OEL.LE (tc (X2p E X delay))))
       (ran (OEL.LE (tc (X2p E X delay))))) (labels_of_es E X)) as Hudr. 
  apply oel_udr; auto. apply tc_X2p_partial_order; auto. 
assert (writes E w1) as Hww1.
  split; auto.
    apply ABasic.dom_rf_is_write with E X r1; auto.
      destruct_valid Hv; auto.
assert (udr (B s1) w1) as Hbw1.
  right; auto.
assert (~(exists l' : Label,
     tc (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1))) l' /\
     tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/
  (f (DR (w1,r1))) = f (DR (w1, r1))) as Hlst.
  auto.

assert (exists r, rel_incl r (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) /\ 
        (~(exists l' : Label,
          tc (OEL.LE (tc (X2p E X delay))) (last r) l' /\
          tc (OEL.LE (tc (X2p E X delay))) l' (f (DR (w1, r1)))) \/ last r = f (DR (w1,r1))) /\
        machine E r (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1)) as Hed'. 
  exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split]; auto.
    intros e1 e2 H12; auto.
    left; apply no_interv_last; auto; intros x y Hxy; auto.

generalize (in_buff_implies_f_after2 E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (f (DR (w1,r1))) n s0 s1 (f (DR (w1,r1))) w1 Hudr Hww1 Hdiff Hlr1 Hdiff Hlr1 Htp Hacp Htotp Hdf Hlst Hed' Hbw1);
intro Hpr.
generalize (X2p_ghb E X delay (DW w1) (DR (w1,r1)) Hnd); intro Hrp.
assert (tc (OEL.LE (tc (X2p E X delay))) (f (DW w1)) (f (DW w1))) as Hc.
  apply trc_ind with (f (DR (w1,r1))); auto.
    generalize (oel_incl (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hi;
    apply trc_step; apply Hi; apply trc_step; auto.
generalize (oel_ac (X2p E X delay) (labels_of_es E X) (tc_X2p_partial_order E X delay Hwf Hv Hwfd)); intro Hac.
generalize (Hac (f (DW w1)) Hc); intro Ht; inversion Ht.
Qed.

Lemma freads_dont_block :
  forall E X delay n s0 s1 w1 r1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X (f (DR (w1,r1))) -> 
  wf_mem E (M s1) -> wf_buff E (B s1) -> wf_queue E (Q s1) ->
  machine E (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))) (labels_of_es E X) (f (DR (w1,r1))) n (Some s0) (Some s1) ->
  ~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DR (w1,r1))) = None).
Proof.
intros E X delay n s0 s1 w1 r1 Hwf Hv Hwfd Hwfl Hwfrfl Hlfwr Hwfm Hwfb Hwfq Hm Hs; simpl in Hs;
unfold flush_resolved_read_step in Hs.
assert (rf X w1 r1) as Hrf1.
  rewrite <- (rf_p2X_X2p E X delay Hwf Hv Hwfd Hwfl); auto.
  split; auto. 
    destruct Hwfl as [? [[? Hr_f] ?]]; apply Hr_f; auto.

  assert (labels_of_es E X (f (DW w1))) as Hlw1.
    apply wfl_w with E; auto; split.
      apply ABasic.dom_rf_in_events with X r1; auto.
        destruct_valid Hv; split; auto.
      apply ABasic.dom_rf_is_write with E X r1; auto.
        destruct_valid Hv; auto.

case_eq (bin (ran (Q s1)) r1); intro Hq; rewrite Hq in Hs.
Focus 2. 
rewrite in_order_implies_queue_state with E X  delay n s0 s1 w1 r1 in Hq; auto.
  inversion Hq.

  clear Hq; case_eq (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w1))); intro Hdp; rewrite Hdp in Hs.
  Focus 2. 
  apply contrad_dp with E X delay n s0 s1 w1 r1; auto.

    clear Hdp; case_eq (bemp (rrestrict (rel_union (B s1) (Q s1)) (fun e => rel_union (A.ppo E) (A.abc E (p2X E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X) )) e r1))); intro Hppo; rewrite Hppo in Hs.
    Focus 2. 
    apply contrad_ppo with E X delay n s0 s1 (DR (w1,r1)); auto.
      apply label_implies_evt with X (f (DR (w1,r1))); auto.
        destruct_valid Hv; split; auto.
      exists (pred (OEL.LE (tc (X2p E X delay))) (f (DR (w1,r1)))); split; [|split]; auto. 
        intros e1 e2 H12; auto.
        left; apply no_interv_last; auto; intros x y Hxy; auto.

      clear Hppo; case_eq (bin (Intersection _ (ran (B s1)) (rrB r1)) w1); intro Hgrf; rewrite Hgrf in Hs.
        rewrite grf_contrad with E X delay n s0 s1 w1 r1 in Hgrf; auto; inversion Hgrf.

        clear Hgrf; destruct (eqEv_dec w1 (M s1 (loc r1))) as [Heqm | Hneqm].
          case_eq (bemp (rrestrict (B s1) (fun w => po_iico E w r1 /\ loc r1 = loc w))); [intros Hemp; rewrite Hemp in Hs; inversion Hs | intros Hnemp; rewrite Hnemp in Hs].

              rewrite unifr_contrad'' with E X delay n s0 s1 w1 r1 in Hnemp; auto; inversion Hnemp.

       case_eq (bin (fun w => (B s1 w (waft E (B s1) r1)) /\ loc r1 = loc w) w1); intro Haft; rewrite Haft in Hs. 
        Focus 2. 
        rewrite uniwsrf_contrad with E X delay n s0 s1 w1 r1 in Haft; auto; inversion Haft.
 
        clear Haft; 
          case_eq (bin (fun w => (w = wbef E (B s1) r1 \/ B s1 (wbef E (B s1) r1) w) /\ loc r1 = loc w) w1); intro Hbef; rewrite Hbef in Hs.
            inversion Hs.
            rewrite unifr_contrad with E X delay n s0 s1 w1 r1 in Hbef; auto; inversion Hbef.
Qed.

Lemma valid_implies_not_stuck :
  forall E X delay l1,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  wf_labels E (labels_of_es E X) -> wf_rf_labels E (labels_of_es E X) ->
  labels_of_es E X l1 ->
  ~((exists n, exists s0, exists s1,
   wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1) /\
   machine E (pred (OEL.LE (tc (X2p E X delay))) l1) (labels_of_es E X) l1 n (Some s0) (Some s1) /\
   step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) l1 = None)).
Proof.
intros E X delay l1 Hwf Hv Hwfd Hwfl Hwfrfl Hlb1 [n [s0 [s1 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]].

  generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
  generalize (oel_incl (X2p E X delay) (labels_of_es E X) Hpart); intro Hitc.

case_eq l1; intros de Hde.

case_eq de; intros c Hc. 
  destruct c as [w r]; subst;
  generalize Hs; fold (~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (d (DR (w,r))) = None));
  apply dreads_dont_block with (d (DR (w,r))) n s0; auto.

  subst; generalize Hs; fold (~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (d (DW c)) = None)); 
  apply dwrites_dont_block with (d (DW c)) n s0; auto.

case_eq de; intros c Hc.
  destruct c as [w r]; subst; 
  generalize Hs; fold (~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DR (w,r))) = None));
  apply freads_dont_block with n s0; auto.

    generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intros [Hudr ?].

  subst; generalize Hs; fold (~(step E (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) (Some s1) (f (DW c)) = None));
  apply fwrites_dont_block with n s0; auto.
    destruct Hwfl as [[? Hwfw] ?]; generalize (Hwfw c Hlb1); auto.
Qed.

Definition mns E p (L:set Label) :=
  forall l, L l -> ( p l (last p) \/ last p = l) ->
  ~((exists n, exists s0, exists s1,
   wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1) /\
   machine E (pred p l) L l n (Some s0) (Some s1) /\
   step E p L (Some s1) l = None)).

Lemma vexec_forms_path :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  mns E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X).
Proof.
intros E X delay Hwf Hv Hwfd l1 Hlb1 Hllb1.
  apply valid_implies_not_stuck; auto.
  apply wfl_es; auto.
  apply wfl_rf_es; auto.
Qed.

Lemma mp :
  forall E p l L n s0 s1,
  Included _ (Union _ (dom p) (ran p)) L ->
  (forall d, rel_incl d p ->
     (forall x1 x2, x1 <> x2 -> (L x1 /\ udr d x1) -> (L x2 /\ udr d x2) ->
      d x1 x2 \/ d x2 x1)) -> 
  (exists d, d = (pred p l) /\ 
     machine E d L l n (Some s0) (Some s1)) ->
  machine E (pred p l) L l n (Some s0) (Some s1).
Proof.
intros E p l L n s0 s2 Hudr Htot Hm;
generalize p l s2 Hudr Htot Hm; clear p l s2 Hudr Htot Hm;
induction n; intros p l s2 Hudr Htot [d [Hi  Hm]].

  inversion Hm.
     
    subst; auto.

  assert (rel_incl d (pred p l)) as Hincl.
     rewrite Hi; intros x y Hxy; auto.          

  destruct Hm as [l' [s1 [Hl' [Hm [Hs [? ?]]]]]];
  fold (machine E (pred d l') L l' n (Some s0) (Some s1)) in Hm.
    exists l'; exists s1; split; [|split; [|split; [|split]]]; auto.  
      subst; auto.  
      rewrite <- Hi; auto.
    
  assert (exists d : Rln Label,
    d = (pred (pred p l) l') /\
    machine E d L l' n (Some s0) (Some s1)) as Hex'.
    exists (pred d l'); split; auto.
      rewrite Hi; auto.
 
  assert (forall d : Rln Label,
  rel_incl d (pred p l) ->
  forall x1 x2 : Label,
  x1 <> x2 -> L x1 /\ udr d x1 -> L x2 /\ udr d x2 -> d x1 x2 \/ d x2 x1) as Htot'.
    intros d0 Hi0 x1 x2 Hdiff H1 H2; apply (Htot d0); auto.
      apply rel_incl_trans with (pred p l); auto; apply pred_incl; auto.

  assert (Included Label (Union Label (dom (pred p l)) (ran (pred p l))) L) as Hudr'.
    apply incl_trans with (Union _ (dom p) (ran p)); auto.
    intros ? [x [y [Hxy ?]] | x [y [Hyx ?]]].
      left; exists y; auto. right; exists y; auto.

  generalize (IHn (pred p l) l' s1 Hudr' Htot' Hex'); intro Hm'; auto.

      apply step_monot with d; auto.
        
Qed.

Lemma last_is_last :
  forall r l, last r = l -> ~(exists l', r l l').
Proof.
unfold last; intros r l Hl.
generalize (LPick.pick_in (fun l : Label => ran r l /\ ~ (exists l' : Label, r l l')));
rewrite Hl; intros [? ?]; auto.
Qed.

Lemma mns_pred :
  forall E p (L:set Label) l,
  
  Included _ (Union _ (dom p) (ran p)) L ->
  trans p ->
  L l ->
  
  (forall d, rel_incl d p ->
     (forall x1 x2, x1 <> x2 -> (L x1 /\ udr d x1) -> (L x2 /\ udr d x2) ->
      d x1 x2 \/ d x2 x1)) -> 
  
  mns E p L ->
  mns E (pred p l) L.
Proof.
intros E p L l  Hudrp Htp Hl Htotp   Hmns l' Hl' Hll'.
intros [n [s0 [s1 [Hwfm [Hwfb [Hwfq [Hm Hs]]]]]]].
  
  assert (rel_incl (pred p l) p) as Hincl.
    intros x y [Hxy ?]; auto.

apply (Hmns l' Hl').

  destruct (eqLabel_dec l (last p)) as [Heql | Hneql].
  rewrite <- Heql; left.
    inversion Hll' as [Hpred | Hlast]; clear Hll'.
    destruct Hpred as [? ?].
      apply Htp with (last (pred p l)); auto.
      generalize (last_in_ran (pred p l) l' Hlast); intros [? [? ?]]; auto.

  assert (p (last (pred p l)) (last p)) as Hpp.
      assert (last (pred p l) = last (pred p l)) as Ht.
        auto.
      generalize (last_in_ran (pred p l) (last (pred p l)) Ht); intros [x [? ?]]; auto;
      apply Htp with l; auto.
      assert (p l (last p) \/ p (last p) l) as Hor.
        assert (rel_incl p p) as Hi.
          intros e1 e2 H12; auto.
        assert (L l /\ udr p l) as Hll.
          split; auto; right; exists (last (pred p l)); auto.
        assert (L (last p) /\ udr p (last p)) as Hllp.
          split; auto.
            apply Hudrp; right; apply last_in_ran; auto.
            right; apply last_in_ran; auto.
        generalize (Htotp p Hi l (last p) Hneql Hll Hllp); auto.

      inversion Hor as [|Hc]; clear Hor; auto.
        assert (last p = last p) as Htlp.
          auto.
        generalize (last_is_last p (last p) Htlp); intro Hn.
        assert (exists l' : Label, p (last p) l') as Hy.
          exists l; auto.
        contradiction.

  left; inversion Hll' as [Hpred | Hlast]; clear Hll'.
    destruct Hpred as [? ?].
      apply Htp with (last (pred p l)); auto.
      rewrite <- Hlast; auto.

  exists n; exists s0; exists s1; 
    split; [|split; [|split; [|split]]]; auto.

    apply mp; auto.
      exists (pred (pred p l) l'); split; auto.

      apply pred_pred; auto.  
      inversion Hll' as [Hpred | Hlast]; clear Hll'.
        destruct Hpred as [? ?].
          apply Htp with (last (pred p l)); auto.
          generalize (last_in_ran (pred p l) l' Hlast); intros [? [? ?]]; auto.

  apply step_monot with (pred p l); auto.
Qed.

Lemma some_none :
  forall E p L s1 l,
  step E p L (Some s1) l <> None ->
  (exists s2 : State, step E p L (Some s1) l = Some s2).
Proof.
intros E p L s1 l Hn.
case_eq (step E p L (Some s1) l). 
  intros s Hss.
  exists s; auto.

  intros Hns; contradiction.
Qed.

Lemma wfm_fwd :
  forall E p L s1 s2 l,
  wf_labels E L -> wf_rf_labels E L ->
  L l ->
  wf_mem E (M s1) ->
  step E p L (Some s1) l = Some s2 ->
  wf_mem E (M s2).
Proof.
intros E p L s1 s2 l Hwfl Hwfrfl Hil Hwfm Hs.
case_eq l; intros de Hl; rewrite Hl in Hs;
case_eq de.
  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.
  intros w Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.

  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; unfold flush_resolved_read_step in Hs.
    destruct (bin (ran (Q s1)) r); [|inversion Hs]; auto.
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); [|inversion Hs]; auto.
    destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs]; auto.
    destruct (eqEv_dec w (M s1 (loc r))).
      destruct (bemp
           (rrestrict (B s1)
              (fun w : Event => po_iico E w r /\ loc r = loc w))); inversion Hs; simpl; auto.
      destruct (bin (Intersection Event (ran (B s1)) (rrB r)) w); [inversion Hs|].
      destruct (bin (fun w : Event => B s1 w (waft E (B s1) r) /\ loc r = loc w) w); [|inversion Hs].
      destruct (bin
           (fun w : Event =>
            (w = wbef E (B s1) r \/ B s1 (wbef E (B s1) r) w) /\
            loc r = loc w) w); inversion Hs; simpl; auto.

  intros w Hde; rewrite Hde in Hs; simpl in Hs; unfold flush_write_step in Hs.
    destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs].
    destruct (bemp
           (rrestrict (Q s1)
              (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
    destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w))) w); inversion Hs; simpl.
    unfold wf_mem.
      intros a; generalize (Hwfm a); intros [? ?]; split;
        unfold upd_mem; destruct (eqLoc_dec a (loc w)); auto.
          destruct Hwfl as [[? Hf] ?]; auto; rewrite Hde in Hl; 
          rewrite Hl in Hil; apply Hf; auto.
Qed.
    
Lemma updb_trans :
  forall (B:Rln Event) w,
  (forall w1 w2, B w1 w2 -> loc w1 = loc w2) ->  
  trans B ->
  trans (upd_buff B w).
Proof.
intros B w Hrrl Htr; unfold upd_buff; case_eq (bin (udr B) w); [intros Hi | intros Hni]; 
intros e1 e2 e3 H12 H23.

  apply Htr with e2; auto.
  
  inversion H12 as [Hb12 | [Hor12 H2]]; inversion H23 as [Hb23 | [Hor23 H3]].
    left; apply Htr with e2; auto.
    
      right; split; auto.
         split; auto.
          left; exists e2; auto.
            destruct Hor23 as [? Hl].
              rewrite <- Hl; apply Hrrl; auto.

      assert (bin (udr B) w = true) as Ht.
        apply In_bin; left; exists e3; subst; auto.
      rewrite Ht in Hni; inversion Hni.

    right; split; auto.
Qed.

Lemma tc_updb' :
  forall B e w x y,
  loc e <> loc w ->
  tc (rrestrict (upd_buff B w) (fun w0 : Event => loc w0 = loc e)) x y ->
  tc (rrestrict B (fun w0 : Event => loc w0 = loc e)) x y.
Proof.
intros B e w x y Hl Hxy.

induction Hxy as [x y [Hs [Hlx Hly]] |].
  unfold upd_buff in Hs; destruct (bin (udr B) w).
    apply trc_step; split; auto.
    inversion Hs as [? | [Hor Hw]].
      apply trc_step; split; auto.
      unfold Ensembles.In in Hly; rewrite Hw in Hly; assert False as Ht.
        apply Hl; auto.
      inversion Ht.

  apply trc_ind with z; auto.
Qed.

Lemma wfb_upd :
  forall E L s1 w,
  wf_labels E L -> wf_rf_labels E L ->
  
  L (d (DW w)) ->
  wf_buff E (B s1) ->
  wf_buff E (upd_buff (B s1) w).
Proof.
intros E L s1 w Hwfl Hwfrfl  Hl [Hudr [Htot [Hac [Htrans Hrrl]]]]; split; [|split; [|split; [|split]]].

  intros ? [e [e' Hd] | e [e' Hr]].
  unfold upd_buff in Hd.
    destruct (bin (udr (B s1)) w).
      apply Hudr; auto; left; exists e'; auto.
      inversion Hd as [Hb |  [[Hb ?] ?]].
        apply Hudr; auto; left; exists e'; auto.
        apply Hudr; auto; right; auto.

  unfold upd_buff in Hr.
    destruct (bin (udr (B s1)) w).
      apply Hudr; auto; right; exists e'; auto.
      inversion Hr as [Hb |  [[Hb ?] ?]].
        apply Hudr; auto; right; exists e'; auto.
          destruct Hwfl as [[Hd ?] ?]; apply Hd; subst; auto.
        

  intros w1 w2 Hdiff Hr1 Hr2 Heql; apply upd_wfb; auto.
  intros e.
    destruct (eqLoc_dec (loc e) (loc w)) as [Heq | Hneq].
      rewrite Heq; apply acb_persistent; auto. 

      destruct Hwfl as [[Hd ?] ?].
      generalize (Hd w Hl); intros [Hew ?].
        intro Heqw; rewrite Heqw in Hew; generalize (dB_not_in_E E dB Hew); intro Hc. 
        assert False as Ht.
          apply Hc; auto.
        inversion Ht.      
 
    intros x Hx; generalize (tc_updb' (B s1) e w x x Hneq Hx); intro Hc.
      apply (Hac e x Hc).

  apply updb_trans; auto.

  intros w1 w2; unfold upd_buff; case_eq (bin (udr (B s1)) w); intros Hi.
    intros Hb; apply Hrrl; auto.
    intros [Hb | [Hor Hw]].
      apply Hrrl; auto.
        
          destruct Hor as [? Heql].
          subst; auto.
Qed.

Lemma delb_trans :
  forall (B:Rln Event) w,
  trans B ->
  trans (del_buff B w).
Proof.
intros B w Htr; unfold del_buff; intros e1 e2 e3 [H12 [? ?]] [H23 [? ?]];
split; [|split]; auto; apply Htr with e2; auto.  
Qed.

Lemma tc_delb :
  forall B w e x y,
  tc ((rrestrict (del_buff B w) (fun w0 : Event => loc w0 = loc e))) x y ->
  tc ((rrestrict B (fun w0 : Event => loc w0 = loc e))) x y.
Proof.
intros B w e x y Hxy.
induction Hxy as [x y [[Hs ?] ?]|].

  apply trc_step; split; auto.

  apply trc_ind with z; auto.
Qed.

Lemma delb_ac :
  forall B w,
  (forall e, acyclic (rrestrict B (fun w : Event => loc w = loc e))) ->
  (forall e, acyclic (rrestrict (del_buff B w) (fun w0 : Event => loc w0 = loc e))).
Proof.
intros B w Hac e x Hx.
generalize (tc_delb B w e x x Hx); intro Hc.
apply (Hac e x Hc); auto.
Qed.

Lemma wfb_delb :
  forall E L s1 w,
  wf_labels E L -> wf_rf_labels E L ->
  L (f (DW w)) ->
  wf_buff E (B s1) ->
  wf_buff E (del_buff (B s1) w).
Proof.
intros E L s1 w Hwfl Hwfrfl Hl [Hudr [Htot [Hac [Htrans Hrrl]]]]; split; [|split; [|split; [|split]]].

  intros ? [e [e' [Hd ?]] | e [e' [Hr ?]]]; apply Hudr.
    left; exists e'; auto. right; exists e'; auto.

  intros w1 w2 Hd [x1 [Hw1 [? ?]]] [x2 [Hw2 [? ?]]] Heql.
  assert (ran (B s1) w1) as Hr1.
    exists x1; auto.
  assert (ran (B s1) w2) as Hr2.
    exists x2; auto.
  generalize (Htot w1 w2 Hd Hr1 Hr2 Heql); intros [H12 | H21]; [left | right];
  split; auto.

  apply delb_ac; auto.

  apply delb_trans; auto.

  intros w1 w2 [H12 ?]; apply Hrrl; auto.
Qed.

Lemma wfb_fwd :
  forall E p L s1 s2 l,
  wf_labels E L ->
  wf_rf_labels E L ->
  L l ->
  wf_buff E (B s1) ->
  step E p L (Some s1) l = Some s2 ->
  wf_buff E (B s2).
Proof.
intros E p L s1 s2 l Hwfl Hwfrfl Hll Hwfb Hs.
case_eq l; intros de Hl; rewrite Hl in Hs;
case_eq de.
  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.
  intros w Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.
  apply wfb_upd with L; auto.
    rewrite Hde in Hl; rewrite Hl in Hll; auto.

  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; unfold flush_resolved_read_step in Hs.
    destruct (bin (ran (Q s1)) r); [|inversion Hs]; auto.
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); [|inversion Hs]; auto.
    destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs]; auto.
    destruct (eqEv_dec w (M s1 (loc r))).
      destruct (bemp
           (rrestrict (B s1)
              (fun w : Event => po_iico E w r /\ loc r = loc w))); inversion Hs; simpl; auto.
      destruct (bin (Intersection Event (ran (B s1)) (rrB r)) w); [inversion Hs|].
      destruct (bin (fun w : Event => B s1 w (waft E (B s1) r) /\ loc r = loc w) w); [|inversion Hs].
      destruct (bin
           (fun w : Event =>
            (w = wbef E (B s1) r \/ B s1 (wbef E (B s1) r) w) /\
            loc r = loc w) w); inversion Hs; simpl; auto.

  intros w Hde; rewrite Hde in Hs; simpl in Hs; unfold flush_write_step in Hs.
    destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs].
    destruct (bemp
           (rrestrict (Q s1)
              (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
    destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w))) w); inversion Hs; simpl.
    apply wfb_delb with L; auto.
      rewrite Hde in Hl; rewrite Hl in Hll; auto.
Qed.

Lemma wfq_fwd :
  forall E p L s1 s2 l,
  wf_labels E L -> wf_rf_labels E L ->
  L l ->
  wf_queue E (Q s1) ->
  step E p L (Some s1) l = Some s2 ->
  wf_queue E (Q s2).
Proof.
intros E p L s1 s2 l Hwfl Hwfrfl Hil Hwfq Hs.
case_eq l; intros de Hl; rewrite Hl in Hs;
case_eq de.

  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.
  intros e [e' [Hs1 | [? Hr]]]. 
    apply Hwfq; exists e'; auto.
    
    destruct Hwfl as [? [[Hd ?] ?]].
    rewrite Hr; rewrite Hde in Hl; rewrite Hl in Hil; generalize (Hd w r Hil);
    intros [? [? [a [? [[v Har] ?]]]]]; split; auto; exists a; exists v; auto.

  intros w Hde; rewrite Hde in Hs; simpl in Hs; inversion Hs; simpl; auto.

  intros [w r] Hde; rewrite Hde in Hs; simpl in Hs; simpl; auto.
  unfold flush_resolved_read_step in Hs.
    destruct (bin (ran (Q s1)) r); [|inversion Hs]; auto.
    destruct (bemp (rrestrict (Q s1) (fun e : Event => dp.dp E e w))); [|inversion Hs]; auto.
    destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e r))); [|inversion Hs]; auto.
    destruct (eqEv_dec w (M s1 (loc r))) as [Hm | Hnm].
      destruct (bemp
           (rrestrict (B s1)
              (fun w : Event => po_iico E w r /\ loc r = loc w))); inversion Hs; simpl; auto.

      intros e [e' [He'e ?]].  
        apply Hwfq; exists e'; auto.

      destruct (bin (Intersection Event (ran (B s1)) (rrB r)) w); [inversion Hs|].
      destruct (bin (fun w : Event => B s1 w (waft E (B s1) r) /\ loc r = loc w) w); [|inversion Hs].
      destruct (bin
           (fun w : Event =>
            (w = wbef E (B s1) r \/ B s1 (wbef E (B s1) r) w) /\
            loc r = loc w) w); inversion Hs; simpl; auto.

        intros e [e' [He'e ?]].  
          apply Hwfq; exists e'; auto.

  intros w Hde; rewrite Hde in Hs; simpl in Hs; unfold flush_write_step in Hs.
  destruct (bemp
           (rrestrict (rel_union (B s1) (Q s1))
              (fun e : Event => rel_union (A.ppo E) (A.abc E (p2X E p L)) e w))); [|inversion Hs].
  destruct (bemp
           (rrestrict (Q s1)
              (fun r' : Event => loc r' = loc w /\ po_iico E r' w))); [|inversion Hs].
  destruct (eqEv_dec
           (pick_last (rrestrict (B s1) (fun e : Event => loc e = loc w))) w); inversion Hs; auto.
Qed.

Inductive length (p:Rln Label) (L:set Label) lst n : Prop :=
  | Zero : n = 0 -> False -> length p L lst n 
  | Succ : forall m l, n = S m -> L l -> 
    last p = l  ->
    
    (forall s1 s2 s, Some s2 = Some s -> forall a, p (f (DW (M s1 a))) (f (DW (M s a)))) ->
    length (pred p l) L l m -> length p L lst n.

Lemma progress :
  forall E p (L:set Label) lst,
  wf_labels E L -> wf_rf_labels E L ->
  Included _ (Union Label (dom p) (ran p)) L ->
  trans p ->
  
  
  (forall d, rel_incl  d p ->
     (forall x1 x2, x1 <> x2 -> (L x1 /\ udr d x1) -> (L x2 /\ udr d x2) ->
      d x1 x2 \/ d x2 x1)) ->
  (exists n, length p L lst n) ->
  
  mns E p L -> 
  ((exists n, exists s0, exists s1, 
    wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1) /\
    machine E p L lst n (Some s0) (Some s1))).
Proof.
intros E p L lst Hwfl Hwfrfl Hudrp Htp  Htotp [n Hn] .

induction Hn as [p L l n Heqn Hf | p L lx n m l Heqn Hl Hdp Hps] .

  inversion Hf.

  intros Hmns .

  assert (mns E (pred p l) L) as Hmnspred.
    apply mns_pred; auto.

  assert (Included Label (Union Label (dom (pred p l)) (ran (pred p l))) L) as Hudrpp.
    intros ? [x [y [Hx ?]] | x [y [Hx ?]]]; apply Hudrp; [left | right]; exists y; auto.

  assert (forall d : Rln Label,
        rel_incl d (pred p l) ->
        forall x1 x2 : Label,
        x1 <> x2 ->
        L x1 /\ udr d x1 -> L x2 /\ udr d x2 -> 
        d x1 x2 \/ d x2 x1) as Htotpp.
    intros d Hincl x1 x2 Hdiff [Hl1 ?] [Hl2 ?]; apply Htotp; auto.
      intros x y Hxy; apply Hincl; auto.


  generalize (IHHn Hwfl Hwfrfl Hudrpp (td_tpred p l Htp)  Htotpp  Hmnspred); 
  intros [n0 [s0 [s1 [Hwfm [Hwfb [Hwfq Hm]]]]]]; clear IHHn.
  generalize (Hmns l Hl); intro Hno.
  assert (forall n s0 s1, 
            wf_mem E (M s1) ->
            wf_buff E (B s1) ->
            wf_queue E (Q s1) ->
            machine E (pred p l) L l n (Some s0) (Some s1) ->
            ~(step E p L (Some s1) l = None)) as Hfall.
    intros n' s s' Hwfm' Hwfb' Hwfq' Hm' Hsn.
    assert (p l (last p) \/ last p = l) as Hlast.
      destruct Hdp; auto.
    apply (Hno Hlast); exists n'; exists s; exists s'; split; [|split; [|split]]; auto.
  clear Hno.
  generalize (Hfall n0 s0 s1 Hwfm Hwfb Hwfq Hm); intro Hnnone.
    exists (S n0); exists s0.
    
    generalize (some_none E p L s1 l Hnnone); intros [s2 Hs2].
    exists s2; split; [|split; [|split]].

      apply wfm_fwd with p L s1 l; auto. 
      apply wfb_fwd with p L s1 l; auto. 
      apply wfq_fwd with p L s1 l; auto.
      exists l; exists s1; split; [|split; [|split; [|split]]]; auto.
      apply Hps; auto.
Qed.

Definition wf_l_lp p (L:set Label) :=
  forall l, L l ->
  (p l (last p) \/ last p = l).  

Lemma mns_progress_mns:
  forall E p (L:set Label),
  wf_l_lp p L ->
  (forall l, L l -> 
   (exists n, exists s0, exists s1, 
    wf_mem E (M s1) /\ wf_buff E (B s1) /\ wf_queue E (Q s1) /\
    machine E (pred p l) L l n (Some s0) (Some s1))) ->
  mns E p L ->
  machine_not_stuck E p L.
Proof.
intros E p L Hwflp Hprogress Hmns l Hl;
generalize (Hwflp l Hl); intro Hlast;
generalize (Hmns l Hl Hlast); intro Hn.

assert (exists n : nat,
         exists s0 : State,
           exists s1 : State,
               wf_mem E (M s1) /\
               wf_buff E (B s1) /\
               wf_queue E (Q s1) /\
               machine E (pred p l) L l n (Some s0) (Some s1) /\
               (exists s2, step E p L (Some s1) l = Some s2)) as Hex.
  generalize (Hprogress l Hl); intros [n [s0 [s1 [Hwfm [Hwfb [Hwfq Hm]]]]]];
  exists n; exists s0; exists s1; split; [|split; [|split; [|split]]]; auto.
  generalize (excluded_middle (step E p L (Some s1) l = None)); intros [Hnone | Hsome]; auto.
    assert False as Ht.
      apply Hn; exists n; exists s0; exists s1; split; [|split; [|split; [|split]]]; auto.
    inversion Ht.

    generalize (excluded_middle (exists s2, step E p L (Some s1) l = Some s2)); 
      intros [He | Hne]; auto.
    
    apply (some_none E p L s1 l Hsome).

destruct Hex as [n [s0 [s1 [Hwfm [Hwfb [Hwfq [Hm [s2 Hs12]]]]]]]];
exists n; exists s0; exists s1; exists s2; split; [|split; [|split; [|split]]]; auto.
Qed.

Lemma mns_mns :
  forall E p (L:set Label),
  wf_labels E L -> wf_rf_labels E L -> wf_l_lp p L ->
  Included Label (Union Label (dom p) (ran p)) L ->
  trans p ->
  (forall d0 : Rln Label,
   rel_incl d0 p ->
   forall x1 x2 : Label,
   x1 <> x2 -> L x1 /\ udr d0 x1 -> L x2 /\ udr d0 x2 -> 
   d0 x1 x2 \/ d0 x2 x1) ->
  (forall l, L l -> exists n, length (pred p l) L l n) ->
  mns E p L ->
  machine_not_stuck E p L.
Proof.
intros E p L Hwfl Hwfrfl Hwflp Hudrp Htp  Htotp Hfp Hllp Hmns; apply mns_progress_mns; auto.
intros l Hl; apply progress; auto.
  intros ? [x [y [Hx ?]]|x [y [Hx ?]]]; apply Hudrp;
    [left|right]; exists y; auto. 
  apply td_tpred; auto.

    intros d Hi; apply Htotp.
      intros e1 e2 H12; generalize (Hi e1 e2 H12); intros [? ?]; auto.

apply mns_pred; auto.
Qed.

Definition pred_tot E X delay:=
(forall d0 : Rln Label,
rel_incl d0 (OEL.LE (tc (X2p E X delay))) ->
forall x1 x2 : Label,
x1 <> x2 ->
labels_of_es E X x1 /\ udr d0 x1 ->
labels_of_es E X x2 /\ udr d0 x2 -> d0 x1 x2 \/ d0 x2 x1).

Definition fpref E X delay :=
(forall l : Label, labels_of_es E X l ->
  (exists n : nat,
    length (pred (OEL.LE (tc (X2p E X delay))) l) (labels_of_es E X) l n)) /\
wf_l_lp (OEL.LE (tc (X2p E X delay))) (labels_of_es E X) /\
pred_tot E X delay.

Theorem from_ax_to_mach :
  forall E X delay,
  well_formed_event_structure E ->
  AWmm.valid_execution E X ->
  wf_delay delay ->
  fpref E X delay ->
  machine_not_stuck E (OEL.LE (tc (X2p E X delay)))(labels_of_es E X) .
Proof.
intros E X delay Hwf Hv Hwfd [? [? ?]].
generalize (tc_X2p_partial_order E X delay Hwf Hv Hwfd); intro Hpart.
apply mns_mns; auto.
  apply wfl_es; auto.
  apply wfl_rf_es; auto.
  apply oel_udr; auto.
  apply oel_trans with (labels_of_es E X); auto.

intros l1 Hlb1 ?.
  apply valid_implies_not_stuck; auto.
  apply wfl_es; auto.
  apply wfl_rf_es; auto.
Qed.

End Completeness.
